@@ -40,12 +40,14 @@
 #include "address-book.h"
 #include "address-book-ldap.h"
 #include "information.h"
+#include "libbalsa-conf.h"
+#include "i18n.h"
 
 static const int DEBUG_LDAP = 0;
 /* don't search when prefix has length shorter than LDAP_MIN_LEN */
 static const unsigned LDAP_MIN_LEN=2;
 /* Which parameters do we want back? */
-static char* attrs[] = {
+static char* book_attrs[] = {
     "cn",        /* maps to displayed name */
     "mail",      /* maps to itself         */
     "sn",        /* maps to last name      */
@@ -54,6 +56,14 @@ static char* attrs[] = {
     "uid",       /* maps to nick name      */
     NULL
 };
+
+static char* complete_attrs[] = {
+    "cn",        /* maps to displayed name */
+    "mail",      /* maps to itself         */
+    "sn",        /* maps to last name      */
+    "givenname", /* maps to first name     */
+    NULL
+};
 /* End of FIXME */
 
 static LibBalsaAddressBookClass *parent_class = NULL;
@@ -160,10 +170,11 @@ libbalsa_address_book_ldap_init(LibBalsaAddressBookLdap * ab)
     ab->host = NULL;
     ab->base_dn = NULL;
     ab->bind_dn = NULL;
+    ab->priv_book_dn = NULL;
     ab->passwd  = NULL;
     ab->enable_tls = FALSE;
     ab->directory = NULL;
-    LIBBALSA_ADDRESS_BOOK(ab)->is_expensive = FALSE;
+    LIBBALSA_ADDRESS_BOOK(ab)->is_expensive = TRUE;
 }
 
 static void
@@ -178,6 +189,7 @@ libbalsa_address_book_ldap_finalize(GObject * object)
     g_free(addr_ldap->host);    addr_ldap->host = NULL;
     g_free(addr_ldap->base_dn); addr_ldap->base_dn = NULL;
     g_free(addr_ldap->bind_dn); addr_ldap->bind_dn = NULL;
+    g_free(addr_ldap->priv_book_dn); addr_ldap->priv_book_dn = NULL;
     g_free(addr_ldap->passwd);  addr_ldap->passwd  = NULL;
 
     G_OBJECT_CLASS(parent_class)->finalize(object);
@@ -186,7 +198,8 @@ libbalsa_address_book_ldap_finalize(GObject * object)
 LibBalsaAddressBook *
 libbalsa_address_book_ldap_new(const gchar *name, const gchar *host,
                                const gchar *base_dn, const gchar *bind_dn,
-                               const gchar *passwd, gboolean enable_tls)
+                               const gchar *passwd, const gchar *priv_book_dn,
+                               gboolean enable_tls)
 {
     LibBalsaAddressBookLdap *ldap;
     LibBalsaAddressBook *ab;
@@ -201,6 +214,7 @@ libbalsa_address_book_ldap_new(const gchar *name, const gchar *host,
     ldap->host = g_strdup(host);
     ldap->base_dn = g_strdup(base_dn);
     ldap->bind_dn = g_strdup(bind_dn);
+    ldap->priv_book_dn = g_strdup(priv_book_dn ? priv_book_dn : bind_dn);
     ldap->passwd = g_strdup(passwd);
     ldap->enable_tls = enable_tls;
 
@@ -235,7 +249,7 @@ libbalsa_address_book_ldap_open_connection(LibBalsaAddressBookLdap * ab)
 
     g_return_val_if_fail(ab->host != NULL, FALSE);
 
-    ab->directory = ldap_init(ab->host, LDAP_PORT);
+    ldap_initialize(&ab->directory, ab->host);
     if (ab->directory == NULL) { /* very unlikely... */
         libbalsa_address_book_set_status(lbab, g_strdup("Host not found"));
 	return LDAP_SERVER_DOWN;
@@ -247,7 +261,7 @@ libbalsa_address_book_ldap_open_connection(LibBalsaAddressBookLdap * ab)
     if(!v3_enabled) ldap_perror(ab->directory, "ldap_set_option");
     if(v3_enabled && ab->enable_tls) {
 #ifdef HAVE_LDAP_TLS
-        /* turn TLS on */
+        /* turn TLS on  but what if we have SSL already on? */
         result = ldap_start_tls_s(ab->directory, NULL, NULL);
         if(result != LDAP_SUCCESS) {
             ldap_unbind(ab->directory);
@@ -263,7 +277,6 @@ libbalsa_address_book_ldap_open_connection(LibBalsaAddressBookLdap * ab)
 #endif /* HAVE_LDAP_TLS */
     }
 
-    printf("Binding as: %s\n", ab->bind_dn ? ab->bind_dn : "anonymous");
     result = ldap_simple_bind_s(ab->directory, 
                                 ab->bind_dn,
                                 ab->passwd);
@@ -324,7 +337,7 @@ libbalsa_address_book_ldap_load(LibBalsaAddressBook * ab,
             : g_strdup("(objectClass=inetOrgPerson)");
         msgid = ldap_search(ldap_ab->directory, ldap_ab->base_dn,
                             LDAP_SCOPE_SUBTREE, 
-                            ldap_filter, attrs, 0);
+                            ldap_filter, book_attrs, 0);
         if (msgid == -1) {
             libbalsa_address_book_ldap_close_connection(ldap_ab);
             continue; /* try again */
@@ -422,6 +435,48 @@ libbalsa_address_book_ldap_get_address(LibBalsaAddressBook * ab,
     return address;
 }
 
+static InternetAddress*
+lbabl_get_internet_address(LDAP *dir, LDAPMessage * e)
+{
+    InternetAddress *ia;
+    BerElement *ber = NULL;
+    char *attr, **vals;
+    int i;
+    gchar *email = NULL, *sn = NULL, *cn = NULL, *first = NULL;
+
+    for (attr = ldap_first_attribute(dir, e, &ber);
+	 attr != NULL; 
+         attr = ldap_next_attribute(dir, e, ber)) {
+	/*
+	 * For each attribute, get the attribute name and values.
+	 */
+	if ((vals = ldap_get_values(dir, e, attr)) != NULL) {
+	    for (i = 0; vals[i] != NULL; i++) {
+		if ((g_ascii_strcasecmp(attr, "sn") == 0) && (!sn))
+		    sn = g_strdup(vals[i]);
+		if ((g_ascii_strcasecmp(attr, "cn") == 0) && (!cn))
+		    cn = g_strdup(vals[i]);
+		if ((g_ascii_strcasecmp(attr, "givenName") == 0) && (!first))
+		    first = g_strdup(vals[i]);
+		if ((g_ascii_strcasecmp(attr, "mail") == 0) && (!email))
+		    email = g_strdup(vals[i]);
+	    }
+	    ldap_value_free(vals);
+	}
+    }
+    /*
+     * Record will have e-mail (searched)
+     */
+    if(email == NULL) email = g_strdup("none");
+    g_return_val_if_fail(email != NULL, NULL);
+
+    if(!cn)
+        cn = create_name(first, sn);
+    ia = internet_address_new_name(cn, email);
+    g_free(email); g_free(sn); g_free(cn); g_free(first);
+
+    return ia;
+}
 /*
  * create_name()
  *
@@ -476,7 +531,7 @@ libbalsa_address_book_ldap_add_address(LibBalsaAddressBook *ab,
 
     dn = g_strdup_printf("mail=%s,%s",
                          (char*)address->address_list->data,
-                         ldap_ab->bind_dn);
+                         ldap_ab->priv_book_dn);
     mods[0] = &modarr[0];
     modarr[0].mod_op = LDAP_MOD_ADD;
     modarr[0].mod_type = "objectClass";
@@ -553,7 +608,7 @@ libbalsa_address_book_ldap_remove_address(LibBalsaAddressBook *ab,
 
     dn = g_strdup_printf("mail=%s,%s",
                          (char*)address->address_list->data,
-                         ldap_ab->bind_dn);
+                         ldap_ab->priv_book_dn);
     cnt = 0;
     do {
         rc = ldap_delete_s(ldap_ab->directory, dn);
@@ -620,7 +675,7 @@ libbalsa_address_book_ldap_modify_address(LibBalsaAddressBook *ab,
 
     dn = g_strdup_printf("mail=%s,%s",
                          (char*)address->address_list->data,
-                         ldap_ab->bind_dn);
+                         ldap_ab->priv_book_dn);
     cnt = 0;
 
     if(!STREQ(address->full_name,newval->full_name)) {
@@ -697,11 +752,13 @@ libbalsa_address_book_ldap_save_config(LibBalsaAddressBook * ab,
 
     ldap = LIBBALSA_ADDRESS_BOOK_LDAP(ab);
 
-    gnome_config_set_string("Host", ldap->host);
-    if(ldap->base_dn) gnome_config_set_string("BaseDN", ldap->base_dn);
-    if(ldap->bind_dn) gnome_config_private_set_string("BindDN", ldap->bind_dn);
-    if(ldap->passwd)  gnome_config_private_set_string("Passwd", ldap->passwd);
-    gnome_config_set_bool("EnableTLS", ldap->enable_tls);
+    libbalsa_conf_set_string("Host", ldap->host);
+    if(ldap->base_dn) libbalsa_conf_set_string("BaseDN", ldap->base_dn);
+    if(ldap->bind_dn) libbalsa_conf_private_set_string("BindDN", ldap->bind_dn);
+    if(ldap->passwd)  libbalsa_conf_private_set_string("Passwd", ldap->passwd);
+    if(ldap->priv_book_dn)
+        libbalsa_conf_set_string("BookDN", ldap->priv_book_dn);
+    libbalsa_conf_set_bool("EnableTLS", ldap->enable_tls);
     if (LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->save_config)
 	LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->save_config(ab, prefix);
 }
@@ -716,24 +773,31 @@ libbalsa_address_book_ldap_load_config(LibBalsaAddressBook * ab,
 
     ldap = LIBBALSA_ADDRESS_BOOK_LDAP(ab);
 
-    ldap->host = gnome_config_get_string("Host");
-    ldap->base_dn = gnome_config_get_string("BaseDN");
+    ldap->host = libbalsa_conf_get_string("Host");
+    ldap->base_dn = libbalsa_conf_get_string("BaseDN");
     if(ldap->base_dn && *ldap->base_dn == 0) { 
 	g_free(ldap->base_dn); ldap->base_dn = NULL; 
     }
 
-    ldap->bind_dn = gnome_config_private_get_string("BindDN");
+    ldap->bind_dn = libbalsa_conf_private_get_string("BindDN");
     if(ldap->bind_dn && *ldap->bind_dn == 0) { 
 	g_free(ldap->bind_dn); ldap->bind_dn = NULL; 
     }
-    ldap->passwd = gnome_config_private_get_string("Passwd");
+    ldap->passwd = libbalsa_conf_private_get_string("Passwd");
     if(ldap->passwd && *ldap->passwd == 0) { 
 	g_free(ldap->passwd); ldap->passwd = NULL; 
     }
-    ldap->enable_tls = gnome_config_get_bool("EnableTLS");
+    ldap->priv_book_dn = libbalsa_conf_get_string("BookDN");
+    if(ldap->priv_book_dn && *ldap->priv_book_dn == 0) { 
+	g_free(ldap->priv_book_dn); ldap->priv_book_dn = NULL; 
+    }
+    ldap->enable_tls = libbalsa_conf_get_bool("EnableTLS");
 
     if (LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->load_config)
 	LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->load_config(ab, prefix);
+
+    if (ab->is_expensive < 0)
+        ab->is_expensive = TRUE;
 }
 
 
@@ -809,7 +873,7 @@ libbalsa_address_book_ldap_alias_complete(LibBalsaAddressBook * ab,
 {
     static struct timeval timeout = { 15, 0 }; /* 15 sec timeout */
     LibBalsaAddressBookLdap *ldap_ab;
-    LibBalsaAddress *addr;
+    InternetAddress *addr;
     GList *res = NULL;
     gchar* filter;
     gchar* ldap;
@@ -840,7 +904,7 @@ libbalsa_address_book_ldap_alias_complete(LibBalsaAddressBook * ab,
     g_free(ldap);
     result = NULL;
     rc = ldap_search_st(ldap_ab->directory, ldap_ab->base_dn,
-			LDAP_SCOPE_SUBTREE, filter, attrs, 0, 
+			LDAP_SCOPE_SUBTREE, filter, complete_attrs, 0, 
 			&timeout, &result);
     if(DEBUG_LDAP)
         g_print("Sent LDAP request: %s (basedn=%s) res=0x%x\n", 
@@ -850,9 +914,9 @@ libbalsa_address_book_ldap_alias_complete(LibBalsaAddressBook * ab,
     case LDAP_SUCCESS:
 	for(e = ldap_first_entry(ldap_ab->directory, result);
 	    e != NULL; e = ldap_next_entry(ldap_ab->directory, e)) {
-	    addr = libbalsa_address_book_ldap_get_address(ab, e);
+	    addr = lbabl_get_internet_address(ldap_ab->directory, e);
 	    if(new_prefix && !*new_prefix) 
-		*new_prefix = libbalsa_address_to_gchar(addr, 0);
+		*new_prefix = internet_address_to_string(addr, FALSE);
 	    res = g_list_prepend(res, addr);
 	}
     case LDAP_SIZELIMIT_EXCEEDED:
@@ -878,6 +942,7 @@ libbalsa_address_book_ldap_alias_complete(LibBalsaAddressBook * ab,
     if(result) ldap_msgfree(result);
 
     if(res) res = g_list_reverse(res);
+
     return res;
 }
 #endif				/*LDAP_ENABLED */
