@@ -23,11 +23,11 @@
 #include "config.h"
 
 #include <string.h>
-#include <libgnome/libgnome.h>
 #include <gmime/gmime.h>
 
 #include "address.h"
 #include "misc.h"
+#include "i18n.h"
 
 static GObjectClass *parent_class;
 
@@ -78,7 +78,6 @@ libbalsa_address_init(LibBalsaAddress * addr)
     addr->nick_name = NULL;
     addr->full_name = NULL;
     addr->first_name = NULL;
-    addr->middle_name = NULL;
     addr->last_name = NULL;
     addr->organization = NULL;
     addr->address_list = NULL;
@@ -96,7 +95,6 @@ libbalsa_address_finalize(GObject * object)
     g_free(addr->nick_name);    addr->nick_name = NULL;
     g_free(addr->full_name);    addr->full_name = NULL;
     g_free(addr->first_name);   addr->first_name = NULL;
-    g_free(addr->middle_name);  addr->middle_name = NULL;
     g_free(addr->last_name);    addr->last_name = NULL;
     g_free(addr->organization); addr->organization = NULL;
 
@@ -113,220 +111,65 @@ libbalsa_address_new(void)
     return g_object_new(LIBBALSA_TYPE_ADDRESS, NULL);
 }
 
-
-/* returns only first address on the list; ignores remaining ones */
-LibBalsaAddress *
-libbalsa_address_new_from_string(const gchar * str)
-{
-    LibBalsaAddress* addr;
-    InternetAddressList *list;
-    gchar *tmp = g_strdup(str);
-    
-    libbalsa_utf8_sanitize(&tmp, FALSE, NULL);
-    list = internet_address_parse_string(tmp);
-    g_free(tmp);
-    if (!list)
-	return NULL;
-
-    addr = libbalsa_address_new();
-    addr->full_name = g_strdup(list->address->name);
-    addr->address_list = g_list_append(addr->address_list, 
-				       g_strdup(list->address->value.addr));
-    internet_address_list_destroy(list);
-    return addr;
-}
-
 void
-libbalsa_address_set_copy(LibBalsaAddress *dest, LibBalsaAddress *src)
+libbalsa_address_set_copy(LibBalsaAddress * dest, LibBalsaAddress * src)
 {
-    GList *src_al;
-    if(dest == src) /* safety check */
+    GList *src_al, *dst_al;
+
+    if (dest == src)            /* safety check */
         return;
-    g_free(dest->nick_name); dest->nick_name = g_strdup(src->nick_name);
-    g_free(dest->full_name); dest->full_name = g_strdup(src->full_name);
-    g_free(dest->middle_name); dest->middle_name = g_strdup(src->middle_name);
-    g_free(dest->last_name); dest->last_name = g_strdup(src->last_name);
+
+    g_free(dest->nick_name);
+    dest->nick_name = g_strdup(src->nick_name);
+    g_free(dest->full_name);
+    dest->full_name = g_strdup(src->full_name);
+    g_free(dest->first_name);
+    dest->first_name = g_strdup(src->first_name);
+    g_free(dest->last_name);
+    dest->last_name = g_strdup(src->last_name);
     g_free(dest->organization);
     dest->organization = g_strdup(src->organization);
-    g_list_foreach(dest->address_list, (GFunc)g_free, NULL);
+    g_list_foreach(dest->address_list, (GFunc) g_free, NULL);
     g_list_free(dest->address_list);
 
-    dest->address_list = NULL;
-    for(src_al = src->address_list; src_al; src_al = src_al->next)
-        dest->address_list = 
-            g_list_prepend(dest->address_list, g_strdup(src_al->data));
-
-    dest->address_list = g_list_reverse(dest->address_list);
+    dst_al = NULL;
+    for (src_al = src->address_list; src_al; src_al = src_al->next)
+        dst_al = g_list_prepend(dst_al, g_strdup(src_al->data));
+    dest->address_list = g_list_reverse(dst_al);
 }
 
-GList*
-libbalsa_address_new_list_from_string(const gchar * str)
-{
-    LibBalsaAddress* addr;
-    InternetAddressList *list, *l;
-    GList* lst = NULL;
-
-    gchar *tmp = g_strdup(str);
-    
-    libbalsa_utf8_sanitize(&tmp, FALSE, NULL);
-    l = list = internet_address_parse_string(tmp);
-    g_free(tmp);
-    if (!list)
-	return NULL;
-
-    while(list) {
-	addr = libbalsa_address_new();
-	addr->full_name = g_strdup(list->address->name);
-	addr->address_list = g_list_append(addr->address_list, 
-					   g_strdup(list->address->value.addr));
-	lst = g_list_append(lst, addr);
-	list = list->next;
-    }
-    internet_address_list_destroy(l);
-    return lst;
-}
-
-static gboolean
-needs_quotes(const gchar *str)
-{
-    gboolean quoted = FALSE;
-
-    while (*str) {
-        if (*str == '\\') {
-            if (*++str)
-                ++str;
-        } else {
-            if (*str == '"')
-                quoted = !quoted;
-            else if (!quoted
-                /* RFC 2822 specials, less '"': */
-                 && strchr("()<>[]:;@\\,.", *str))
-                return TRUE;
-            ++str;
-        }
-    }
-    return FALSE;
-}
-
-static gchar*
-rfc2822_mailbox(const gchar *full_name, gchar *address)
+static gchar *
+rfc2822_mailbox(const gchar * full_name, const gchar * address)
 {
+    InternetAddress *ia;
     gchar *new_str;
 
-    if(full_name && *full_name) {
-        gchar *dequote = g_new(char, strlen(full_name) + 1);
-        const gchar *p = full_name;
-        gchar *q = dequote;
-    
-        do {
-            if (*p == '\\') {
-                *q++ = *p++;
-                if (*p)
-                    *q++ = *p++;
-            } else if (*p == '"')
-                ++p;
-            else
-                *q++ = *p++;
-        } while (*p);
-        *q = '\0';
+    ia = internet_address_new_name(full_name, address);
+    new_str = internet_address_to_string(ia, FALSE);
+    internet_address_unref(ia);
 
-        if (needs_quotes(dequote))
-	    new_str = g_strdup_printf("\042%s\042 <%s>", dequote, address);
-        else
-            new_str = g_strdup_printf("%s <%s>", dequote, address);
-        g_free(dequote);
-    } else
-	new_str = g_strdup(address);
     return new_str;
 }
 
-
-#if 1
 static gchar*
 rfc2822_group(const gchar *full_name, GList *addr_list)
 {
-    GString *str = g_string_new("");
-    GList *addr_entry;
+    InternetAddress *ia;
     gchar *res;
 
-    if(full_name) { 
-	if(needs_quotes(full_name))
-	    g_string_printf(str, "\042%s\042: ", full_name);
-	else
-	    g_string_printf(str, "%s: ", full_name);
-    }
+    ia = internet_address_new_group(full_name);
+    for (; addr_list; addr_list = addr_list->next) {
+	InternetAddress *member;
 
-    if(addr_list) {
-	g_string_append(str, (gchar*)addr_list->data);
-        
-	for(addr_entry=g_list_next(addr_list); addr_entry; 
-	    addr_entry=g_list_next(addr_entry)) {
-	    g_string_append_printf(str, ", %s", (gchar*)addr_entry->data);
-	}
+	member = internet_address_new_name(NULL, addr_list->data);
+	internet_address_add_member(ia, member);
+	internet_address_unref(member);
     }
-    if(full_name)
-	g_string_append(str, ";");
-    
-    res=str->str;
-    g_string_free(str, FALSE);
-    
+    res = internet_address_to_string(ia, FALSE);
+    internet_address_unref(ia);
+
     return res;
 }
-#endif
-
-#if 0
-static gchar*
-rfc2822_list(GList *list)
-{
-    gchar *retc = NULL; 
-    GString *str;
-    GList *addr_entry;
-    
-    g_return_val_if_fail(list!=NULL, NULL);
-
-    str=g_string_new((gchar *)list->data);
-
-    for(addr_entry=g_list_next(list); addr_entry; 
-	addr_entry=g_list_next(addr_entry)) {
-	g_string_append_printf(str, ", %s", (gchar *)addr_entry->data);
-    }
-    retc=str->str;
-    g_string_free(str, FALSE);
-
-    return retc;
-}
-#endif
-
-/* private version */
-gchar *
-libbalsa_address_to_gchar_p(LibBalsaAddress * address, gint n);
-gchar *
-libbalsa_address_to_gchar_p(LibBalsaAddress * address, gint n)
-{
-    gchar *retc = NULL;
-
-    g_return_val_if_fail(LIBBALSA_IS_ADDRESS(address), NULL);
-
-    /* FIXME: for n==-1, we should be returning nice rfc822_group but
-     * the entry widgets have not proper support for group string format
-     * so we drop this idea for a while. */
-    if(!address->address_list)
-        return NULL;
-    if(n==-1) {
-        if(address->address_list->next)
-            retc = rfc2822_group(address->full_name, address->address_list);
-        else
-            retc = rfc2822_mailbox(address->full_name,
-                                   address->address_list->data);
-    } else {
-	GList *nth_address = g_list_nth(address->address_list, n);
-	g_return_val_if_fail(nth_address != NULL, NULL);
-
-	retc = rfc2822_mailbox(address->full_name, nth_address->data);
-    }
-
-    return retc;
-}
 
 /* 
    Get a string version of this address.
@@ -337,85 +180,93 @@ libbalsa_address_to_gchar_p(LibBalsaAddress * address, gint n)
 gchar *
 libbalsa_address_to_gchar(LibBalsaAddress * address, gint n)
 {
-    return libbalsa_address_to_gchar_p(address, n);
-}
-
-const gchar *
-libbalsa_address_get_name(const LibBalsaAddress * addr)
-{
-    g_return_val_if_fail(LIBBALSA_IS_ADDRESS(addr), NULL);
-
-    return addr->full_name ? addr->full_name :
-	(addr->address_list ? addr->address_list->data : NULL);
-}
-
-#if ENABLE_ESMTP
-
-/* XXX - added by Brian Stafford <brian@stafford.uklinux.net> */
-
-/* libESMTP works with the RFC 821 mailbox and the RFC 822 phrase and 
-   mailbox as seperate entities.  Because of this it is useful to add
-   these extra methods. */
-
-/* Extract the RFC 822 phrase from the address.  Almost the same
-   as libbalsa_address_get_name() except returns NULL if no phrase. */
-const gchar *
-libbalsa_address_get_phrase(LibBalsaAddress * address)
-{
-    g_return_val_if_fail(LIBBALSA_IS_ADDRESS(address), NULL);
-
-    return address->full_name;
-}
-
-#endif
-/* Extract the nth RFC 821/RFC 822 mailbox from the address. */
-const gchar *
-libbalsa_address_get_mailbox(LibBalsaAddress * address, gint n)
-{
-    GList *nth_address;
+    gchar *retc = NULL;
 
     g_return_val_if_fail(LIBBALSA_IS_ADDRESS(address), NULL);
 
-    nth_address = g_list_nth(address->address_list, n);
-    if(nth_address == NULL) return NULL;
-    return (const gchar*)nth_address->data;
+    if(!address->address_list)
+        return NULL;
+    if(n==-1) {
+        if(address->address_list->next)
+            retc = rfc2822_group(address->full_name, address->address_list);
+        else
+            retc = rfc2822_mailbox(address->full_name,
+                                   address->address_list->data);
+    } else {
+	const gchar *mailbox = g_list_nth_data(address->address_list, n);
+	g_return_val_if_fail(mailbox != NULL, NULL);
+
+	retc = rfc2822_mailbox(address->full_name, mailbox);
+    }
+
+    return retc;
 }
 
+/* Helper */
+static const gchar *
+lba_get_name_or_mailbox(const InternetAddressList * address_list,
+                        gboolean get_name, gboolean in_group)
+{
+    const gchar *retval = NULL;
+
+    for (; address_list; address_list = address_list->next) {
+        InternetAddress *ia = address_list->address;
+
+        if (get_name && ia->name && *ia->name)
+            return ia->name;
+
+        if (ia->type == INTERNET_ADDRESS_NAME)
+            retval = ia->value.addr;
+        else if (ia->type == INTERNET_ADDRESS_GROUP) {
+            if (in_group)
+                g_message("Ignoring nested group address");
+            else
+                retval = lba_get_name_or_mailbox(ia->value.members,
+			get_name, TRUE);
+        }
+        if (retval)
+            break;
+    }
+
+    return retval;
+}
+
+/* Get either a name or a mailbox from an InternetAddressList. */
+const gchar *
+libbalsa_address_get_name_from_list(const InternetAddressList *
+                                    address_list)
+{
+    return lba_get_name_or_mailbox(address_list, TRUE, FALSE);
+}
+
+/* Get a mailbox from an InternetAddressList. */
+const gchar *
+libbalsa_address_get_mailbox_from_list(const InternetAddressList *
+                                       address_list)
+{
+    return lba_get_name_or_mailbox(address_list, FALSE, FALSE);
+}
 
 /* =================================================================== */
 /*                                UI PART                              */
 /* =================================================================== */
 
-/** libbalsa_address_get_edit_widget() returns an widget adapted
-    for a LibBalsaAddress edition, with initial values set if address
-    is provided. The edit entries are set in entries array 
-    and enumerated with LibBalsaAddressField constants
+/** libbalsa_address_set_edit_entries() initializes the GtkEntry widgets
+    in entries with values from address
 */
-GtkWidget*
-libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
-                                 GCallback changed_cb, gpointer changed_data)
+void
+libbalsa_address_set_edit_entries(LibBalsaAddress * address,
+                                  GtkWidget **entries)
 {
-    const static gchar *labels[NUM_FIELDS] = {
-	N_("_Displayed Name:"),
-	N_("_First Name:"),
-#if !defined(ENABLE_TOUCH_UI)
-	N_("_Middle Name:"),
-#endif
-	N_("_Last Name:"),
-	N_("_Nickname:"),
-	N_("O_rganization:"),
-	N_("_Email Address:")
-    };
-
-    GtkWidget *table, *label;
     gchar *new_name = NULL;
     gchar *new_email = NULL;
     gchar *new_organization = NULL;
     gchar *first_name = NULL;
-    gchar *middle_name = NULL;
     gchar *last_name = NULL;
-    gchar *carrier = NULL;
-    gint cnt, cnt2;
+    gchar *nick_name = NULL;
+    gint cnt;
+    GtkListStore *store;
+    GtkTreeIter iter;
 
     new_email = g_strdup(address
                          && address->address_list
@@ -436,7 +287,7 @@ libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
 	/* make sure address->personal is not all whitespace */
 	new_name = g_strstrip(g_strdup(address->full_name));
 
-	/* guess the first name, middle name and last name */
+	/* guess the first name and last name */
 	if (*new_name != '\0') {
 	    names = g_strsplit(new_name, " ", 0);
 
@@ -456,6 +307,7 @@ libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
                 else
                     last_name = g_strdup(names[cnt - 1]);
             }
+#if 0
 	    /* get middle name */
 	    middle_name = g_strdup("");
 
@@ -472,29 +324,146 @@ libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
 			g_free(carrier);
 		    }
 		}
-
+#endif
 	    g_strfreev(names);
 	}
     }
 
     if (first_name == NULL)
 	first_name = g_strdup("");
-    if (middle_name == NULL)
-	middle_name = g_strdup("");
     if (last_name == NULL)
 	last_name = g_strdup("");
+    if (!address || address->nick_name == NULL)
+	nick_name = g_strdup("");
+    else
+	nick_name = g_strdup(address->nick_name);
 
+    /* Full name must be set after first and last names. */
+    gtk_entry_set_text(GTK_ENTRY(entries[FIRST_NAME]), first_name);
+    gtk_entry_set_text(GTK_ENTRY(entries[LAST_NAME]), last_name);
+    gtk_entry_set_text(GTK_ENTRY(entries[FULL_NAME]), new_name);
+    gtk_entry_set_text(GTK_ENTRY(entries[NICK_NAME]), nick_name);
+    gtk_entry_set_text(GTK_ENTRY(entries[ORGANIZATION]), new_organization);
+
+    store = GTK_LIST_STORE(gtk_tree_view_get_model
+                           (GTK_TREE_VIEW(entries[EMAIL_ADDRESS])));
+    gtk_list_store_clear(store);
+    if (address) {
+        GList *list;
+
+        for (list = address->address_list; list; list = list->next) {
+            gtk_list_store_append(store, &iter);
+            gtk_list_store_set(store, &iter, 0, list->data, -1);
+        }
+    } else {
+        gtk_list_store_append(store, &iter);
+        gtk_list_store_set(store, &iter, 0, "", -1);
+    }
+
+    gtk_editable_select_region(GTK_EDITABLE(entries[FULL_NAME]), 0, -1);
+
+    for (cnt = FULL_NAME + 1; cnt < NUM_FIELDS; cnt++)
+        if (GTK_IS_EDITABLE(entries[cnt]))
+            gtk_editable_set_position(GTK_EDITABLE(entries[cnt]), 0);
+
+    g_free(new_name);
+    g_free(first_name);
+    g_free(last_name);
+    g_free(new_email);
+    g_free(new_organization);
+    gtk_widget_grab_focus(entries[FULL_NAME]);
+}
+
+/** libbalsa_address_get_edit_widget() returns an widget adapted
+    for a LibBalsaAddress edition, with initial values set if address
+    is provided. The edit entries are set in entries array 
+    and enumerated with LibBalsaAddressField constants
+*/
+static void
+lba_entry_changed(GtkEntry * entry, GtkEntry ** entries)
+{
+    gchar *full_name =
+        g_strconcat(gtk_entry_get_text(entries[FIRST_NAME]), " ",
+                    gtk_entry_get_text(entries[LAST_NAME]), NULL);
+    gtk_entry_set_text(entries[FULL_NAME], full_name);
+    g_free(full_name);
+}
+
+static void
+lba_cell_edited(GtkCellRendererText * cell, const gchar * path_string,
+                const gchar * new_text, GtkListStore * store)
+{
+    GtkTreeIter iter;
+
+    if (gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(store),
+                                            &iter, path_string))
+        gtk_list_store_set(store, &iter, 0, new_text, -1);
+}
+
+static GtkWidget *
+lba_address_list_widget(GCallback changed_cb, gpointer changed_data)
+{
+    GtkListStore *store;
+    GtkWidget *tree_view;
+    GtkCellRenderer *renderer;
+    GtkTreeViewColumn *column;
+
+    store = gtk_list_store_new(1, G_TYPE_STRING);
+    tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
+    g_object_unref(store);
+    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), FALSE);
+
+    renderer = gtk_cell_renderer_text_new();
+    g_object_set(renderer, "editable", TRUE, NULL);
+    g_signal_connect(renderer, "edited", G_CALLBACK(lba_cell_edited),
+                     store);
+    if (changed_cb)
+        g_signal_connect_swapped(renderer, "edited",
+                                 changed_cb, changed_data);
+
+    column = gtk_tree_view_column_new_with_attributes(NULL, renderer,
+                                                      "text", 0, NULL);
+    gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), column);
+
+    return tree_view;
+}
+
+GtkWidget*
+libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
+                                 GCallback changed_cb, gpointer changed_data)
+{
+    const static gchar *labels[NUM_FIELDS] = {
+	N_("_Displayed Name:"),
+	N_("_First Name:"),
+	N_("_Last Name:"),
+	N_("_Nickname:"),
+	N_("O_rganization:"),
+        N_("_Email Address:")
+    };
+
+    GtkWidget *table, *label;
+    gint cnt;
 
     table = gtk_table_new(NUM_FIELDS, 2, FALSE);
-    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
+#define HIG_PADDING 6
+    gtk_table_set_row_spacings(GTK_TABLE(table), HIG_PADDING);
+    gtk_table_set_col_spacings(GTK_TABLE(table), HIG_PADDING);
+    gtk_container_set_border_width(GTK_CONTAINER(table), HIG_PADDING);
 
     for (cnt = 0; cnt < NUM_FIELDS; cnt++) {
+        if (!labels[cnt])
+            continue;
 	label = gtk_label_new_with_mnemonic(_(labels[cnt]));
-	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
-        entries[cnt] = gtk_entry_new();
-        if(changed_cb)
-            g_signal_connect(G_OBJECT(entries[cnt]), "changed",
-                             changed_cb, changed_data);
+	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
+        if (cnt == EMAIL_ADDRESS)
+            entries[cnt] = lba_address_list_widget(changed_cb,
+                                                   changed_data);
+        else {
+            entries[cnt] = gtk_entry_new();
+            if (changed_cb)
+                g_signal_connect_swapped(entries[cnt], "changed",
+                                         changed_cb, changed_data);
+        }
 	gtk_label_set_mnemonic_widget(GTK_LABEL(label), entries[cnt]);
 
 	gtk_table_attach(GTK_TABLE(table), label, 0, 1, cnt + 1, cnt + 2,
@@ -506,32 +475,18 @@ libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
 			 cnt + 2, GTK_FILL | GTK_EXPAND,
 			 GTK_FILL | GTK_EXPAND, 2, 2);
     }
+    g_signal_connect(entries[FIRST_NAME], "changed",
+                     G_CALLBACK(lba_entry_changed), entries);
+    g_signal_connect(entries[LAST_NAME], "changed",
+                     G_CALLBACK(lba_entry_changed), entries);
 
-    gtk_entry_set_text(GTK_ENTRY(entries[FULL_NAME]), new_name);
-    gtk_entry_set_text(GTK_ENTRY(entries[FIRST_NAME]), first_name);
-#if !defined(ENABLE_TOUCH_UI)
-    gtk_entry_set_text(GTK_ENTRY(entries[MIDDLE_NAME]), middle_name);
-#endif
-    gtk_entry_set_text(GTK_ENTRY(entries[LAST_NAME]), last_name);
-    gtk_entry_set_text(GTK_ENTRY(entries[EMAIL_ADDRESS]), new_email);
-    gtk_entry_set_text(GTK_ENTRY(entries[ORGANIZATION]), new_organization);
+    libbalsa_address_set_edit_entries(address, entries);
 
-    gtk_editable_select_region(GTK_EDITABLE(entries[FULL_NAME]), 0, -1);
-
-    for (cnt = FULL_NAME + 1; cnt < NUM_FIELDS; cnt++)
-        gtk_editable_set_position(GTK_EDITABLE(entries[cnt]), 0);
-
-    g_free(new_name);
-    g_free(first_name);
-    g_free(middle_name);
-    g_free(last_name);
-    g_free(new_email);
-    g_free(new_organization);
     return table;
 }
 
-LibBalsaAddress*
-libbalsa_address_new_from_edit_entries(GtkWidget **entries)
+LibBalsaAddress *
+libbalsa_address_new_from_edit_entries(GtkWidget ** entries)
 {
 #define SET_FIELD(f,e)\
   do{ (f) = g_strstrip(gtk_editable_get_chars(GTK_EDITABLE(e), 0, -1));\
@@ -539,7 +494,12 @@ libbalsa_address_new_from_edit_entries(GtkWidget **entries)
  else { while( (p=strchr(address->full_name,';'))) *p = ','; }  } while(0)
 
     LibBalsaAddress *address;
-    char *p, *addr;
+    char *p;
+    GList *list = NULL;
+    GtkTreeModel *model;
+    gboolean valid;
+    GtkTreeIter iter;
+
     /* FIXME: This problem should be solved in the VCard
        implementation in libbalsa: semicolons mess up how GnomeCard
        processes the fields, so disallow them and replace them
@@ -547,16 +507,25 @@ libbalsa_address_new_from_edit_entries(GtkWidget **entries)
 
     address = libbalsa_address_new();
     SET_FIELD(address->full_name,   entries[FULL_NAME]);
+    if (!address->full_name) {
+        g_object_unref(address);
+        return NULL;
+    }
     SET_FIELD(address->first_name,  entries[FIRST_NAME]);
-#if !defined(ENABLE_TOUCH_UI)
-    SET_FIELD(address->middle_name, entries[MIDDLE_NAME]);
-#endif
     SET_FIELD(address->last_name,   entries[LAST_NAME]);
     SET_FIELD(address->nick_name,   entries[NICK_NAME]);
     SET_FIELD(address->organization,entries[ORGANIZATION]);
-    SET_FIELD(addr,                 entries[EMAIL_ADDRESS]);
 
-    if(addr)
-        address->address_list = g_list_append(address->address_list,addr);
+    model = gtk_tree_view_get_model(GTK_TREE_VIEW(entries[EMAIL_ADDRESS]));
+    for (valid = gtk_tree_model_get_iter_first(model, &iter); valid;
+         valid = gtk_tree_model_iter_next(model, &iter)) {
+        gchar *email;
+
+        gtk_tree_model_get(model, &iter, 0, &email, -1);
+        if (email && *email)
+            list = g_list_prepend(list, email);
+    }
+    address->address_list = g_list_reverse(list);
+
     return address;
 }
