@@ -35,7 +35,6 @@ static gchar*
 coalesce_seq_range(int lo, int hi, CoalesceFunc incl, void *data)
 {
   GString * res = g_string_sized_new(16);
-  char buf[10], *str;
   enum { BEGIN, LASTOUT, LASTIN, RANGE } mode = BEGIN;
   int seq;
   unsigned prev =0, num = 0;
@@ -44,11 +43,11 @@ coalesce_seq_range(int lo, int hi, CoalesceFunc incl, void *data)
     if(seq<=hi && (num=incl(seq, data)) != 0) {
       switch(mode) {
       case BEGIN: 
-        sprintf(buf, "%u", num); g_string_append(res, buf); 
+        g_string_append_printf(res, "%u", num);
         mode = LASTIN; break;
       case RANGE:
         if(num!=prev+1) {
-          sprintf(buf, ":%u,%u", prev, num); g_string_append(res, buf); 
+          g_string_append_printf(res, ":%u,%u", prev, num);
           mode = LASTIN;
         }
         break;
@@ -58,7 +57,7 @@ coalesce_seq_range(int lo, int hi, CoalesceFunc incl, void *data)
           break;
         } /* else fall through */
       case LASTOUT: 
-        sprintf(buf, ",%u", num); g_string_append(res, buf); 
+        g_string_append_printf(res, ",%u", num);
         mode = LASTIN; break;
       }
     } else {
@@ -67,21 +66,14 @@ coalesce_seq_range(int lo, int hi, CoalesceFunc incl, void *data)
       case LASTOUT: break;
       case LASTIN: mode = LASTOUT; break;
       case RANGE: 
-        sprintf(buf, ":%u", prev); g_string_append(res, buf); 
+        g_string_append_printf(res, ":%u", prev);
         mode = LASTOUT;
         break;
       }
     }
     prev = num;
   }
-  if(mode == BEGIN) {
-    str = NULL;
-    g_string_free(res, TRUE);
-  } else {
-    str = res->str;
-    g_string_free(res, FALSE);
-  }
-  return str;
+  return g_string_free(res, mode == BEGIN);
 }
 
 static unsigned
@@ -400,9 +392,36 @@ imap_mbox_lsub(ImapMboxHandle *handle, const char* what)
 
 
 /* 6.3.10 STATUS Command */
-/* FIXME: implement */
-
 
+ImapResponse
+imap_mbox_status(ImapMboxHandle *r, const char*what,
+                 struct ImapStatusResult *res)
+{
+  const char *item_arr[ELEMENTS(imap_status_item_names)+1];
+  ImapResponse rc = IMR_OK;
+  unsigned i, ipos;
+  
+  for(ipos = i= 0; res[i].item != IMSTAT_NONE; i++) {
+    /* repeated items? */
+    g_return_val_if_fail(i<ELEMENTS(imap_status_item_names), IMR_BAD);
+    /* invalid item? */
+    g_return_val_if_fail(res[i].item>=IMSTAT_MESSAGES &&
+                         res[i].item<=IMSTAT_UNSEEN, IMR_BAD);
+    item_arr[ipos++] = imap_status_item_names[res[i].item];
+  }
+  item_arr[ipos] = NULL;
+  if(ipos>0) {
+    gchar *mbx7 = imap_utf8_to_mailbox(what);
+    gchar *items = g_strjoinv(" ", (gchar**)&item_arr[0]);
+    gchar *cmd = g_strdup_printf("STATUS \"%s\" (%s)", mbx7, items);
+    g_hash_table_insert(r->status_resps, (gpointer)what, res);
+    rc = imap_cmd_exec(r, cmd);
+    g_hash_table_remove(r->status_resps, what);
+    g_free(mbx7); g_free(cmd);
+    g_free(items);
+  }
+  return IMR_OK; 
+}
 /* 6.3.11 APPEND Command */
 static gchar*
 enum_flag_to_str(ImapMsgFlags flg)
@@ -436,6 +455,7 @@ imap_mbox_append(ImapMboxHandle *handle, const char *mbox,
   size_t s, delta;
   int c;
 
+  imap_handle_idle_disable(handle);
   if(flags) {
     gchar *str = enum_flag_to_str(flags);
     cmd = g_strdup_printf("APPEND \"%s\" (%s) {%u%s}", mbx7, str,
@@ -478,6 +498,7 @@ imap_mbox_append(ImapMboxHandle *handle, const char *mbox,
     rc = imap_cmd_step (handle, cmdno);
   } while (rc == IMR_UNTAGGED);
 
+  imap_handle_idle_enable(handle, 30);
   return rc;
   }
 }
@@ -701,6 +722,7 @@ imap_assure_needed_flags(ImapMboxHandle *h, ImapMsgFlag needed_flags)
       case IMSGF_RECENT:   flg = "RECENT"; break;
       default: g_free(seqno); continue;
       }
+      if(!cmd) imap_handle_idle_disable(h);
       cmd = g_strdup_printf("SEARCH %s %s", seqno, flg);
       g_free(seqno);
       flag[issued_cmd] = fnd.flag;
@@ -718,6 +740,7 @@ imap_assure_needed_flags(ImapMboxHandle *h, ImapMsgFlag needed_flags)
         rc = imap_cmd_step(h, cmdno[i]);
       } while (rc == IMR_UNTAGGED);
     }
+    imap_handle_idle_enable(h, 30);
   }
   h->search_cb = cb; h->search_arg = arg;
   if(rc == IMR_OK) {
@@ -868,9 +891,10 @@ imap_mbox_handle_fetch_set(ImapMboxHandle* handle,
 }
 
 static void
-write_nstring(const char *str, int len, void *fl)
+write_nstring(unsigned seqno, const char *str, int len, void *fl)
 {
-  fwrite(str, 1, len, (FILE*)fl);
+  if (fwrite(str, 1, len, (FILE*)fl) != (size_t) len)
+    perror("write_nstring");
 }
 
 ImapResponse
@@ -878,11 +902,18 @@ imap_mbox_handle_fetch_rfc822(ImapMboxHandle* handle, unsigned seqno,
                               FILE *fl)
 {
   char cmd[40];
+  ImapFetchBodyCb cb = handle->body_cb;
+  void          *arg = handle->body_arg;
+  ImapResponse rc;
+
   IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
   handle->body_cb  = write_nstring;
   handle->body_arg = fl;
   sprintf(cmd, "FETCH %u RFC822", seqno);
-  return imap_cmd_exec(handle, cmd);
+  rc = imap_cmd_exec(handle, cmd);
+  handle->body_cb  = cb;
+  handle->body_arg = arg;
+  return rc;
 }
 
 ImapResponse
@@ -890,38 +921,63 @@ imap_mbox_handle_fetch_rfc822_uid(ImapMboxHandle* handle, unsigned uid,
                               FILE *fl)
 {
   char cmd[40];
+  ImapFetchBodyCb cb = handle->body_cb;
+  void          *arg = handle->body_arg;
+  ImapResponse rc;
+
   IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
   handle->body_cb  = write_nstring;
   handle->body_arg = fl;
   sprintf(cmd, "UID FETCH %u RFC822", uid);
-  return imap_cmd_exec(handle, cmd);
+  rc = imap_cmd_exec(handle, cmd);
+  handle->body_cb  = cb;
+  handle->body_arg = arg;
+  return rc;
 }
 
 ImapResponse
 imap_mbox_handle_fetch_body(ImapMboxHandle* handle, 
                             unsigned seqno, const char *section,
+                            ImapFetchBodyOptions options,
                             ImapFetchBodyCb body_cb, void *arg)
 {
-  char cmd[80];
+  char cmd[160];
   ImapMessage *msg;
+  ImapFetchBodyCb fcb = handle->body_cb;
+  void          *farg = handle->body_arg;
+  ImapResponse rc;
+
   IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
   handle->body_cb  = body_cb;
   handle->body_arg = arg;
-  /* We should use .MIME only for non-message parts but it tends to
-     work also for messages. Let's be clear: this is a stop gap
-     solution. The proper way to fix it is to provide extra arguments
-     to this function so that it clearly knows what is is supposed to
-     do instead of guessing. */
   msg = imap_mbox_handle_get_msg(handle, seqno);
-  if(strcmp(section, "1") == 0 && msg &&
-     msg->body && msg->body->media_basic != IMBMEDIA_MESSAGE_RFC822 &&
-     msg->body->media_basic != IMBMEDIA_MULTIPART)
+  if(options == IMFB_NONE)
     snprintf(cmd, sizeof(cmd), "FETCH %u BODY[%s]",
              seqno, section);
-  else
-    snprintf(cmd, sizeof(cmd), "FETCH %u (BODY[%s.MIME] BODY[%s])",
-             seqno, section, section);
-  return imap_cmd_exec(handle, cmd);
+  else {
+    char prefix[160];
+    if(options == IMFB_HEADER) {
+      /* We have to strip last section part and replace it with HEADER */
+      unsigned sz;
+      char *last_dot = strrchr(section, '.');
+      strncpy(prefix, section, sizeof(prefix));
+
+      if(last_dot) {
+        sz = last_dot-section+1;
+        if(sz>sizeof(prefix)-1) sz = sizeof(prefix)-1;
+      } else sz = 0;
+      strncpy(prefix + sz, "HEADER", sizeof(prefix)-sz-1);
+      prefix[sizeof(prefix)-1] = '\0';
+    } else
+      snprintf(prefix, sizeof(prefix), "%s.MIME", section);
+
+    snprintf(cmd, sizeof(cmd), "FETCH %u (BODY[%s] BODY[%s])",
+             seqno, prefix, section);
+  }
+  rc = imap_cmd_exec(handle, cmd);
+  handle->body_cb  = fcb;
+  handle->body_arg = farg;
+  return rc;
 }
 
 /* 6.4.6 STORE Command */
@@ -1064,17 +1120,21 @@ imap_mbox_thread(ImapMboxHandle *h, const char *how, ImapSearchKey *filter)
 
     cmdno = imap_make_tag(tag);
     
+    imap_handle_idle_disable(h);
     sio_printf(h->sio, "%s THREAD %s UTF-8 ", tag, how);
     if(!filter)
       sio_write(h->sio, "ALL", 3);
-    else
-      imap_write_key(h, filter, cmdno, can_do_literals);
+    else {
+      if( (rc = imap_write_key(h, filter, cmdno, can_do_literals)) != IMR_OK)
+        return rc;
+    }
 
     sio_write(h->sio, "\r\n", 2);
     imap_handle_flush(h);
     do
       rc = imap_cmd_step(h, cmdno);
     while(rc == IMR_UNTAGGED);
+    imap_handle_idle_enable(h, 30);
     return rc;
   } else
     return IMR_NO;
@@ -1109,7 +1169,7 @@ sort_code_to_string(ImapSortKey key)
 
 ImapResponse
 imap_mbox_sort_msgno(ImapMboxHandle *handle, ImapSortKey key,
-                     int ascending, int *msgno, unsigned cnt)
+                     int ascending, unsigned int *msgno, unsigned cnt)
 {
   ImapResponse rc;
   const char *keystr;
@@ -1137,9 +1197,21 @@ imap_mbox_sort_msgno(ImapMboxHandle *handle, ImapSortKey key,
   rc = imap_cmd_exec(handle, cmd1);
   g_free(cmd1);
 
-  if(rc == IMR_OK)
-    for(i=0; i<cnt; i++)
-      msgno[i] = handle->mbox_view.arr[i];
+  if(rc == IMR_OK) {
+  /* one version of dovecot (which?) returned insufficient number of
+   * numbers, we need to work around it. */
+    if(handle->mbox_view.entries == cnt)
+      for(i=0; i<cnt; i++)
+        msgno[i] = handle->mbox_view.arr[i];
+    else {
+      for(i=0; i<cnt; i++)
+        msgno[i] = i + 1;
+      imap_mbox_handle_set_msg(handle,
+                               "bug in implementation of SORT command on "
+                               "IMAP server exposed.");
+      rc = IMR_NO;
+    }
+  }
 
   return rc;
 }
@@ -1164,7 +1236,7 @@ imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
       handle->mbox_view.entries = 0; /* FIXME: I do not like this! 
                                       * we should not be doing such 
                                       * low level manipulations here */
-      rc = imap_search_exec(handle, filter, append_no, NULL);
+      rc = imap_search_exec(handle, FALSE, filter, append_no, NULL);
     } else {
       if(handle->thread_root)
         g_node_destroy(handle->thread_root);
@@ -1191,6 +1263,7 @@ imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
     
     cmdno =  imap_make_tag(tag);
     keystr = sort_code_to_string(key);
+    imap_handle_idle_disable(handle);
     sio_printf(handle->sio, "%s SORT (%s%s) UTF-8 ", tag,
                ascending ? "" : "REVERSE ", keystr);
     
@@ -1199,9 +1272,12 @@ imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
                                     * low level manipulations here */
     if(!filter)
       sio_write(handle->sio, "ALL", 3);
-    else
-      imap_write_key(handle, filter, cmdno, can_do_literals);
+    else {
+      if( (rc=imap_write_key(handle, filter, cmdno, can_do_literals)) !=IMR_OK)
+        return rc;
+    }
     sio_write(handle->sio, "\r\n", 2);
+    imap_handle_idle_enable(handle, 30);
     imap_handle_flush(handle);
     do
       rc = imap_cmd_step(handle, cmdno);
@@ -1230,6 +1306,52 @@ ImapResponse
 imap_mbox_filter_msgnos(ImapMboxHandle *handle, ImapSearchKey *filter,
 			GHashTable *msgnos)
 {
-  return imap_search_exec(handle, filter, (ImapSearchCb)make_msgno_table,
-                          msgnos);
+  return imap_search_exec(handle, FALSE, filter,
+                          (ImapSearchCb)make_msgno_table, msgnos);
+}
+
+/* imap_mbox_complete_msgids:
+   finds which msgnos are missing from the supplied table and
+   completes them if needed.
+*/
+static unsigned
+need_msgid(unsigned seqno, GPtrArray* msgids)
+{
+  return g_ptr_array_index(msgids, seqno-1) ? 0 : seqno;
+}
+
+static void
+msgid_cb(unsigned seqno, const char *buf, int buflen, void* arg)
+{
+  GPtrArray *arr = (GPtrArray*)arg;
+  g_return_if_fail(seqno>=1 && seqno<=arr->len);
+  g_ptr_array_index(arr, seqno-1) = g_strdup(buf);
+}
+
+ImapResponse
+imap_mbox_complete_msgids(ImapMboxHandle *h,
+			  GPtrArray *msgids,
+			  unsigned first_seqno_to_fetch)
+{
+  gchar *seq, *cmd;
+  ImapResponse rc = IMR_OK;
+  ImapFetchBodyCb cb;
+  void *arg;
+
+  IMAP_REQUIRED_STATE1(h, IMHS_SELECTED, IMR_BAD);
+  seq = coalesce_seq_range(first_seqno_to_fetch, msgids->len,
+			   (CoalesceFunc)need_msgid, msgids);
+  if(seq) {
+    cmd = g_strdup_printf("FETCH %s BODY.PEEK[HEADER.FIELDS (message-id)]",
+			  seq);
+    g_free(seq);
+
+    cb = h->body_cb;   h->body_cb = msgid_cb;
+    arg = h->body_arg; h->body_arg = msgids;
+    rc = imap_cmd_exec(h, cmd);
+    h->body_cb  = cb;
+    h->body_arg = arg;
+    g_free(cmd);
+  }
+  return rc;
 }
