@@ -18,8 +18,9 @@
  */
 #include "config.h"
 
-#define _POSIX_SOURCE 1
-#define _BSD_SOURCE   1
+#define _POSIX_C_SOURCE 199506L
+#define _XOPEN_SOURCE 500
+
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netdb.h>
@@ -142,9 +143,13 @@ imap_mbox_handle_init(ImapMboxHandle *handle)
 #endif
   handle->tls_mode = IMAP_TLS_ENABLED;
   handle->cmd_queue = g_hash_table_new(NULL, NULL);
+  handle->status_resps = g_hash_table_new_full(g_str_hash, g_str_equal,
+                                               NULL, NULL);
 
   handle->info_cb  = NULL;
   handle->info_arg = NULL;
+  handle->enable_anonymous = 0;
+  handle->enable_binary    = 0;
   mbox_view_init(&handle->mbox_view);
 }
 
@@ -168,8 +173,8 @@ imap_mbox_handle_class_init(ImapMboxHandleClass * klass)
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET(ImapMboxHandleClass, list_response),
                  NULL, NULL,
-                 libimap_VOID__INT_STRING_POINTER, G_TYPE_NONE, 3,
-                 G_TYPE_INT, G_TYPE_STRING, G_TYPE_POINTER);
+                 libimap_VOID__INT_INT_POINTER, G_TYPE_NONE, 3,
+                 G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);
 
   imap_mbox_handle_signals[LSUB_RESPONSE] = 
     g_signal_new("lsub-response",
@@ -177,8 +182,8 @@ imap_mbox_handle_class_init(ImapMboxHandleClass * klass)
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET(ImapMboxHandleClass, lsub_response),
                  NULL, NULL,
-                 libimap_VOID__INT_STRING_POINTER, G_TYPE_NONE, 3,
-                 G_TYPE_INT, G_TYPE_STRING, G_TYPE_POINTER);
+                 libimap_VOID__INT_INT_POINTER, G_TYPE_NONE, 3,
+                 G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);
 
   imap_mbox_handle_signals[EXPUNGE_NOTIFY] = 
     g_signal_new("expunge-notify",
@@ -207,6 +212,16 @@ imap_mbox_handle_new(void)
   return handle;
 }
 
+void
+imap_handle_set_option(ImapMboxHandle *h, ImapOption opt, gboolean state)
+{
+  switch(opt) {
+  case IMAP_OPT_ANONYMOUS: h->enable_anonymous = !!state; break;
+  case IMAP_OPT_BINARY:    h->enable_binary    = !!state; break;
+  default: g_warning("imap_set_option: invalid option\n");
+  }
+}
+
 void
 imap_handle_set_infocb(ImapMboxHandle* h, ImapInfoCb cb, void *arg)
 {
@@ -243,6 +258,117 @@ imap_handle_set_timeout(ImapMboxHandle *h, int milliseconds)
     sio_set_timeout(h->sio, milliseconds);
 }
 
+/* imap_handle_idle_enable: enables calling IDLE command after seconds
+   of inactivity. IDLE support consists of three subroutines:
+
+1. imap_handle_idle_{enable,disable}() switch to and from the IDLE
+   mode.  switching to the mode is done by registering an idle
+   callback idle_start() with 30 seconds delay time.
+
+2. idle start() sends the IDLE command and registers idle_process() to
+   be notified whenever data is available on the specified descriptor.
+
+3. idle_process() processes the data sent from the server. */
+static gboolean
+idle_process(GIOChannel *source, GIOCondition condition, gpointer data)
+{
+  ImapMboxHandle *h = (ImapMboxHandle*)data;
+  ImapResponse rc;
+  int retval;
+  g_return_val_if_fail(h, FALSE);
+  if(h->state == IMHS_DISCONNECTED)
+    return FALSE;
+  while( (retval = sio_poll(h->sio, TRUE, FALSE, TRUE)) != -1 &&
+         (retval & SIO_READ) != 0) {
+    if ( (rc=imap_cmd_step(h, 0)) == IMR_UNKNOWN ||
+         rc == IMR_SEVERED || rc == IMR_BYE || rc == IMR_PROTOCOL ||
+         rc  == IMR_BAD) {
+      printf("idle_process() got unexpected response %i!\n"
+             "Last message was: \"%s\" - shutting down connection.\n",
+             rc, h->last_msg);
+      imap_handle_disconnect(h);
+      return FALSE;
+    }
+  }
+  return TRUE;
+}
+
+static gboolean
+idle_start(gpointer data)
+{
+  ImapMboxHandle *h = (ImapMboxHandle*)data;
+  ImapResponse rc;
+  int c;
+
+  /* The test below can probably be weaker since it is ok for the
+     channel to get disconnected before IDLE gets activated */
+  IMAP_REQUIRED_STATE3(h, IMHS_CONNECTED, IMHS_AUTHENTICATED,
+                       IMHS_SELECTED, FALSE);
+
+  sio_write(h->sio, "0 IDLE\r\n", 8); sio_flush(h->sio);
+  do { /* FIXME: we could at this stage just as well watch for
+          response instead of polling. */
+    rc = imap_cmd_step (h, 0);
+  } while (rc == IMR_UNTAGGED);
+  if(rc != IMR_RESPOND) {
+    g_message("Expected IMR_RESPOND but got %d\n", rc);
+    return FALSE;
+  }
+  while( (c=sio_getc(h->sio)) != -1 && c != '\n');
+  if(!h->iochannel) {
+    h->iochannel = g_io_channel_unix_new(h->sd);
+    g_io_channel_set_encoding(h->iochannel, NULL, NULL);
+  }
+  h->idle_watch_id = g_io_add_watch(h->iochannel, G_IO_IN, idle_process, h);
+  h->idle_enable_id = 0;
+  return FALSE;
+}
+
+gboolean
+imap_handle_idle_enable(ImapMboxHandle *h, int seconds)
+{
+  if( !imap_mbox_handle_can_do(h, IMCAP_IDLE))
+    return FALSE;
+  if(h->idle_watch_id) {
+    fprintf(stderr, "idle already enabled\n");
+    return FALSE;
+  }
+  if(!h->idle_enable_id)
+    h->idle_enable_id = g_timeout_add(seconds*1000, idle_start, h);
+  return TRUE;
+}
+
+gboolean
+imap_handle_idle_disable(ImapMboxHandle *h)
+{
+  if(h->idle_enable_id) {
+    g_source_remove(h->idle_enable_id);
+    h->idle_enable_id = 0;
+  }
+  if(h->idle_watch_id) {
+    g_source_remove(h->idle_watch_id);
+    h->idle_watch_id = 0;
+    if(h->sio) { /* we might have been disconnected before */
+      sio_write(h->sio,"DONE\r\n",6); sio_flush(h->sio);
+    }
+  }
+  return TRUE;
+}
+
+void
+imap_handle_disconnect(ImapMboxHandle *h)
+{
+  imap_handle_idle_disable(h);
+  if(h->sio) {
+    sio_detach(h->sio); h->sio = NULL;
+  }
+  if(h->iochannel) {
+    g_io_channel_unref(h->iochannel); h->iochannel = NULL;
+  }
+  close(h->sd);
+  h->state = IMHS_DISCONNECTED;
+}
+
 int imap_mbox_is_disconnected (ImapMboxHandle *h)
 { return IMAP_MBOX_IS_DISCONNECTED(h); }
 int imap_mbox_is_connected    (ImapMboxHandle *h)
@@ -259,8 +385,10 @@ imap_mbox_handle_connect(ImapMboxHandle* ret, const char *host, int over_ssl)
 
   g_return_val_if_fail(imap_mbox_is_disconnected(ret), IMAP_CONNECT_FAILED);
 #if !defined(USE_TLS)
-  if(over_ssl)
+  if(over_ssl) {
+    imap_mbox_handle_set_msg(ret,"SSL requested but SSL support not compiled");
     return IMAP_UNSECURE;
+  }
 #else
   ret->over_ssl = over_ssl;
 #endif
@@ -374,10 +502,7 @@ imap_socket_open(const char* host, const char *def_port)
   for (cur = res; cur != NULL; cur = cur->ai_next) {
     fd = socket (cur->ai_family, cur->ai_socktype, cur->ai_protocol);
     if (fd >= 0) {
-      int sa_size = sizeof (struct sockaddr_in);
-      if (cur->ai_addr->sa_family == AF_INET6)
-        sa_size = sizeof (struct sockaddr_in6);
-      if ((rc=connect(fd, cur->ai_addr, sa_size)) == 0) {
+      if ((rc=connect(fd, cur->ai_addr, cur->ai_addrlen)) == 0) {
 	break;
       } else {
 	close (fd);
@@ -423,13 +548,16 @@ imap_mbox_connect(ImapMboxHandle* handle)
 #ifdef USE_TLS
   if(handle->over_ssl) {
     SSL *ssl = imap_create_ssl();
-    if(!ssl) return IMAP_UNSECURE;
+    if(!ssl) {
+      imap_mbox_handle_set_msg(handle,"SSL context could not be created");
+      return IMAP_UNSECURE;
+    }
     if(imap_setup_ssl(handle->sio, handle->host, ssl,
                       handle->user_cb, handle->user_arg)) 
       handle->using_tls = 1;
     else {
-      sio_detach(handle->sio); handle->sio = NULL; close(handle->sd);
-      handle->state = IMHS_DISCONNECTED;
+      imap_mbox_handle_set_msg(handle,"SSL negotiation failed");
+      imap_handle_disconnect(handle);
       return IMAP_UNSECURE;
     }
   }
@@ -438,11 +566,10 @@ imap_mbox_connect(ImapMboxHandle* handle)
     sio_set_monitorcb(handle->sio, handle->monitor_cb, handle->monitor_arg);
 
   handle->state = IMHS_CONNECTED;
-  if (imap_cmd_step(handle, 0) != IMR_UNTAGGED) {
-    g_warning("imap_mbox_connect:unexpected initial response\n");
-    sio_detach(handle->sio);
-    close(handle->sd);
-    handle->state = IMHS_DISCONNECTED;
+  if ( (resp=imap_cmd_step(handle, 0)) != IMR_UNTAGGED) {
+    g_message("imap_mbox_connect:unexpected initial response(%d):\n%s\n",
+	      resp, handle->last_msg);
+    imap_handle_disconnect(handle);
     return IMAP_PROTOCOL_ERROR;
   }
   handle->can_fetch_body = 
@@ -452,17 +579,20 @@ imap_mbox_connect(ImapMboxHandle* handle)
     resp = IMR_OK; /* secured already with SSL */
   else if(handle->tls_mode != IMAP_TLS_DISABLED &&
           imap_mbox_handle_can_do(handle, IMCAP_STARTTLS)) {
-    if( imap_handle_starttls(handle) != IMR_OK)
+    if( imap_handle_starttls(handle) != IMR_OK) {
+      imap_mbox_handle_set_msg(handle,"TLS negotiation failed");
       return IMAP_UNSECURE; /* TLS negotiation error */
+    }
     resp = IMR_OK; /* secured with TLS */
   } else
     resp = IMR_NO; /* not over SSL and TLS unavailable */
 #else
   resp = IMR_NO;
 #endif
-  if(handle->tls_mode == IMAP_TLS_REQUIRED && resp != IMR_OK)
+  if(handle->tls_mode == IMAP_TLS_REQUIRED && resp != IMR_OK) {
+    imap_mbox_handle_set_msg(handle,"TLS required but not available");
     return IMAP_UNSECURE;
-    
+  }
   return IMAP_SUCCESS;
 }
 
@@ -607,14 +737,13 @@ imap_mbox_handle_finalize(GObject* gobject)
     handle->doing_logout = TRUE;
     imap_cmd_exec(handle, "LOGOUT");
   }
-  handle->state = IMHS_DISCONNECTED;
-  if(handle->sio) { sio_detach(handle->sio); handle->sio = NULL; }
-  close(handle->sd);
+  imap_handle_disconnect(handle);
   g_free(handle->host);    handle->host   = NULL;
   g_free(handle->mbox);    handle->mbox   = NULL;
   g_free(handle->last_msg);handle->last_msg = NULL;
 
   g_hash_table_destroy(handle->cmd_queue); handle->cmd_queue = NULL;
+  g_hash_table_destroy(handle->status_resps); handle->status_resps = NULL;
 
   mbox_view_dispose(&handle->mbox_view);
   imap_mbox_resize_cache(handle, 0);
@@ -1020,7 +1149,6 @@ imap_body_append_child(ImapBody* body, ImapBody* child)
 void
 imap_body_set_id(ImapBody *body, char *id)
 {
-  if(id) printf("part ID='%s'\n", id);
   body->content_id = id;
 }
 /* ================ END OF BODY STRUCTURE FUNCTIONS ==================== */
@@ -1215,9 +1343,8 @@ imap_cmd_step(ImapMboxHandle* handle, unsigned lastcmd)
   }
 #endif
   if( imap_cmd_get_tag(handle->sio, tag, sizeof(tag))<0) {
-    printf("connection severed.\n");
-    close(handle->sd);
-    handle->state = IMHS_DISCONNECTED;
+    printf("imap connection severed.\n");
+    imap_handle_disconnect(handle);
     return IMR_SEVERED;
   }
   /* handle untagged messages. The caller still gets its shot afterwards. */
@@ -1247,9 +1374,7 @@ imap_cmd_step(ImapMboxHandle* handle, unsigned lastcmd)
   /* our command tags are hexadecimal numbers */
   if(sscanf(tag, "%x", &cmdno) != 1) {
     printf("scanning '%s' for tag number failed. Cannot recover.\n", tag);
-    /* sio_detach(handle->sio); leads to a crash */
-    close(handle->sd);
-    handle->state = IMHS_DISCONNECTED;
+    imap_handle_disconnect(handle);
     return IMR_BAD;
   }
 
@@ -1257,9 +1382,10 @@ imap_cmd_step(ImapMboxHandle* handle, unsigned lastcmd)
   if(cmdno == lastcmd)
     return rc;
   else {
-    g_hash_table_insert(handle->cmd_queue,
-                        GUINT_TO_POINTER(cmdno),
-                        GINT_TO_POINTER(rc));
+    if(cmdno)
+      g_hash_table_insert(handle->cmd_queue,
+                          GUINT_TO_POINTER(cmdno),
+                          GINT_TO_POINTER(rc));
     return IMR_UNTAGGED;
   }
 }
@@ -1280,6 +1406,7 @@ imap_cmd_exec(ImapMboxHandle* handle, const char* cmd)
     return IMR_SEVERED;
 
   /* create sequence for command */
+  imap_handle_idle_disable(handle);
   if (imap_cmd_start(handle, cmd, &cmdno)<0)
     return IMR_SEVERED;  /* irrecoverable connection error. */
 
@@ -1288,8 +1415,7 @@ imap_cmd_exec(ImapMboxHandle* handle, const char* cmd)
     rc = imap_cmd_step (handle, cmdno);
   } while (rc == IMR_UNTAGGED);
 
-  if (rc != IMR_OK) /* we should be silent here */
-    printf("cmd '%s' failed, rc=%d.\n", cmd, rc);
+  imap_handle_idle_enable(handle, 30);
 
   return rc;
 }
@@ -1298,6 +1424,7 @@ int
 imap_handle_write(ImapMboxHandle *conn, const char *buf, size_t len)
 {
   g_return_val_if_fail(conn, -1);
+  g_return_val_if_fail(conn->sio, -1);
 
   sio_write(conn->sio, buf, len); /* why it is void? */
   return 0;
@@ -1307,6 +1434,7 @@ void
 imap_handle_flush(ImapMboxHandle *handle)
 {
   g_return_if_fail(handle);
+  g_return_if_fail(handle->sio);
   sio_flush(handle->sio);
 }
 
@@ -1315,12 +1443,11 @@ imap_mbox_gets(ImapMboxHandle *h, char* buf, size_t sz)
 {
   char* rc;
   g_return_val_if_fail(h, NULL);
+  g_return_val_if_fail(h->sio, NULL);
 
   rc = sio_gets(h->sio, buf, sz);
-  if(rc == NULL) {
-    sio_detach(h->sio);
-    h->state = IMHS_DISCONNECTED;
-  }
+  if(rc == NULL)
+      imap_handle_disconnect(h);
   return rc;
 }
 
@@ -1455,10 +1582,11 @@ ir_capability_data(ImapMboxHandle *handle)
 {
   /* ordered identically as ImapCapability constants */
   static const char* capabilities[] = {
-    "IMAP4", "IMAP4rev1", "STATUS", "ACL", "NAMESPACE", "AUTH=CRAM-MD5", 
-    "AUTH=GSSAPI", "AUTH=ANONYMOUS", "STARTTLS", "LOGINDISABLED", "SORT",
-    "THREAD=ORDEREDSUBJECT", "THREAD=REFERENCES", "UNSELECT", "SCAN",
-    "CHILDREN", "LITERAL+", "SASL-IR"
+    "IMAP4", "IMAP4rev1", "STATUS", "ACL", "NAMESPACE",
+    "AUTH=ANONYMOUS", "AUTH=CRAM-MD5", "AUTH=GSSAPI", "AUTH=PLAIN",
+    "STARTTLS", "LOGINDISABLED", "SORT",
+    "THREAD=ORDEREDSUBJECT", "THREAD=REFERENCES",
+    "UNSELECT", "SCAN", "CHILDREN", "LITERAL+", "IDLE", "SASL-IR"
   };
   unsigned x;
   int c;
@@ -1491,6 +1619,7 @@ ir_resp_text_code(ImapMboxHandle *h)
   int c = imap_get_atom(h->sio, buf, sizeof(buf));
   ImapResponse rc = IMR_OK;
 
+
   for(o=0; o<ELEMENTS(resp_text_code); o++)
     if(g_ascii_strcasecmp(buf, resp_text_code[o]) == 0) break;
 
@@ -1508,7 +1637,7 @@ ir_resp_text_code(ImapMboxHandle *h)
   case 10:imap_get_atom(h->sio, buf, sizeof(buf)); h->unseen =atoi(buf); break;
   default: while( (c=sio_getc(h->sio)) != EOF && c != ']') ; break;
   }
-  return rc;
+  return c != EOF ? rc : IMR_SEVERED;
 }
 static ImapResponse
 ir_ok(ImapMboxHandle *h)
@@ -1629,7 +1758,7 @@ ir_list_lsub(ImapMboxHandle *h, ImapHandleSignal signal)
     "Marked", "Unmarked", "Noselect", "Noinferiors",
     "HasChildren", "HasNoChildren"
   };
-  int flags = 0;
+  ImapMboxFlags flags = 0;
   char buf[LONG_STRING], *s, *mbx;
   int c, delim;
   ImapResponse rc;
@@ -1666,7 +1795,7 @@ ir_list_lsub(ImapMboxHandle *h, ImapHandleSignal signal)
   mbx = imap_mailbox_to_utf8(s);
   rc = ir_check_crlf(h, c);
   g_signal_emit(G_OBJECT(h), imap_mbox_handle_signals[signal],
-                0, delim, &flags, mbx);
+                0, delim, flags, mbx);
   g_free(s);
   g_free(mbx);
   return rc;
@@ -1684,12 +1813,46 @@ ir_lsub(ImapMboxHandle *h)
   return ir_list_lsub(h, LSUB_RESPONSE);
 }
 
+/* 7.2.4 STATUS Response */
+const char* imap_status_item_names[5] = {
+  "MESSAGES", "RECENT", "UIDNEXT", "UIDVALIDITY", "UNSEEN" };
 static ImapResponse
 ir_status(ImapMboxHandle *h)
 {
-  /* FIXME: implement! */
-  int c; while( (c=sio_getc(h->sio))!=EOF && c != 0x0a);
-  return IMR_OK;
+  int c;
+  char *name;
+  struct ImapStatusResult *resp;
+
+  name = imap_get_astring(h->sio, &c);
+  resp = g_hash_table_lookup(h->status_resps, name);
+  if(c                != ' ') {g_free(name); return IMR_PROTOCOL;}
+  if(sio_getc(h->sio) != '(') {g_free(name); return IMR_PROTOCOL;}
+  do {
+    char item[13], count[13]; /* longest than UIDVALIDITY */
+    c = imap_get_atom(h->sio, item, sizeof(item));
+    if(c == ')') break;
+    if(c != ' ') {g_free(name); return IMR_PROTOCOL;}
+    c = imap_get_atom(h->sio, count, sizeof(count));
+    /* FIXME: process the response */
+    if(resp) {
+      unsigned idx, i;
+      for(idx=0; idx<ELEMENTS(imap_status_item_names); idx++)
+        if(g_ascii_strcasecmp(item, imap_status_item_names[idx]) == 0)
+          break;
+      for(i= 0; resp[i].item != IMSTAT_NONE; i++) {
+        if(resp[i].item == idx) {
+          if (sscanf(count, "%u", &resp[i].result) != 1) {
+            g_free(name);
+            return IMR_PROTOCOL;
+          }
+          break;
+        }
+      }
+    }
+  } while(c == ' ');
+  g_free(name);
+  /* g_return_val_if-fail(c == ')', IMR_BAD) */
+  return ir_check_crlf(h, sio_getc(h->sio));
 }
 
 static ImapResponse
@@ -1779,14 +1942,20 @@ flags_tasklet(ImapMboxHandle *h, void *data)
     h->flags_cb(1, &seqno, h->flags_arg);
 }
 
+#define CREATE_IMSG_IF_NEEDED(h,seqno) \
+  if((h)->msg_cache[seqno-1] == NULL) \
+     (h)->msg_cache[(seqno)-1] = imap_message_new();
+
 static ImapResponse
 ir_msg_att_flags(ImapMboxHandle *h, int c, unsigned seqno)
 {
   unsigned i;
-  ImapMessage *msg = h->msg_cache[seqno-1];
+  ImapMessage *msg;
   ImapFlagCache *flags;
 
   if(sio_getc(h->sio) != '(') return IMR_PROTOCOL;
+  CREATE_IMSG_IF_NEEDED(h, seqno);
+  msg = h->msg_cache[seqno-1];
   msg->flags = 0;
 
   do {
@@ -2019,9 +2188,11 @@ ir_envelope(struct siobuf *sio, ImapEnvelope *env)
 static ImapResponse
 ir_msg_att_envelope(ImapMboxHandle *h, int c, unsigned seqno)
 {
-  ImapMessage *msg = h->msg_cache[seqno-1];
+  ImapMessage *msg;
   ImapEnvelope *env;
 
+  CREATE_IMSG_IF_NEEDED(h, seqno);
+  msg = h->msg_cache[seqno-1];
   if(msg->envelope) env = NULL;
   else {
     msg->envelope = env = imap_envelope_new();
@@ -2038,8 +2209,8 @@ static ImapResponse
 ir_msg_att_rfc822(ImapMboxHandle *h, int c, unsigned seqno)
 {
   gchar *str = imap_get_nstring(h->sio);
-  if(h->body_cb)
-    h->body_cb(str, strlen(str), h->body_arg);
+  if(str && h->body_cb)
+    h->body_cb(seqno, str, strlen(str), h->body_arg);
   g_free(str);
   return IMR_OK;
 }
@@ -2058,11 +2229,15 @@ static ImapResponse
 ir_msg_att_rfc822_size(ImapMboxHandle *h, int c, unsigned seqno)
 {
   char buf[12];
-  ImapMessage *msg = h->msg_cache[seqno-1];
-  
+  ImapMessage *msg;
+
   c = imap_get_atom(h->sio, buf, sizeof(buf));
 
   if(c!= -1) sio_ungetc(h->sio);
+
+  CREATE_IMSG_IF_NEEDED(h, seqno);
+  msg = h->msg_cache[seqno-1];
+  
   msg->rfc822size = atoi(buf);  
   return IMR_OK;
 }
@@ -2160,9 +2335,9 @@ ir_body_fld_enc(struct siobuf* sio, ImapBody *body)
   gchar* str = imap_get_string(sio);
   ImapBodyEncoding enc;
 
-  if(!str) return IMR_PROTOCOL;
-
-  if     (g_ascii_strcasecmp(str, "7BIT")==0)             enc = IMBENC_7BIT;
+  /* if(!str) return IMR_PROTOCOL; required - but forgive this error */
+  if     (str == NULL) enc = IMBENC_OTHER;
+  else if(g_ascii_strcasecmp(str, "7BIT")==0)             enc = IMBENC_7BIT;
   else if(g_ascii_strcasecmp(str, "8BIT")==0)             enc = IMBENC_8BIT;
   else if(g_ascii_strcasecmp(str, "BINARY")==0)           enc = IMBENC_BINARY;
   else if(g_ascii_strcasecmp(str, "BASE64")==0)           enc = IMBENC_BASE64;
@@ -2344,7 +2519,7 @@ ir_body_extension (struct siobuf *sio, ImapBody * body)
     }
   else
     /* nstring */
-    imap_get_nstring (sio);
+    g_free(imap_get_nstring(sio));
 
   return IMR_OK;
 }
@@ -2676,45 +2851,58 @@ ir_body (struct siobuf *sio, int c, ImapBody * body,
 
 /* read [section] and following string. FIXME: other kinds of body. */ 
 static ImapResponse
-ir_body_section(struct siobuf *sio, ImapFetchBodyCb body_cb, void *arg)
+ir_body_section(struct siobuf *sio, unsigned seqno,
+		ImapFetchBodyCb body_cb, void *arg)
 {
   char buf[80], *str;
   int c = imap_get_atom(sio, buf, sizeof(buf));
   if(c != ']') { puts("] expected"); return IMR_PROTOCOL; }
   if(sio_getc(sio) != ' ') { puts("space expected"); return IMR_PROTOCOL;}
   str = imap_get_nstring(sio);
-  if(body_cb)
-    body_cb(str, strlen(str), arg);
-  g_free(str);
+  if(str) {
+    if(body_cb)
+      body_cb(seqno, str, strlen(str), arg);
+    g_free(str);
+  }
   return IMR_OK;
 }
 
 static ImapResponse
-ir_body_header_fields(struct siobuf* sio, ImapMessage *msg)
+ir_body_header_fields(ImapMboxHandle *h, unsigned seqno)
 {
+  ImapMessage *msg;
   char *tmp;
   int c;
 
-  if(sio_getc(sio) != '(') return IMR_PROTOCOL;
+  if(sio_getc(h->sio) != '(') return IMR_PROTOCOL;
 
-  while ((tmp = imap_get_astring(sio, &c))) {
+  while ((tmp = imap_get_astring(h->sio, &c))) {
       /* nothing (yet?) */;
     g_free(tmp);
     if (c == ')')
       break;
   }
   if(c != ')') return IMR_PROTOCOL;
-  if(sio_getc(sio) != ']') return IMR_PROTOCOL;
-  if(sio_getc(sio) != ' ') return IMR_PROTOCOL;
-  g_free(msg->fetched_header_fields);
-  msg->fetched_header_fields = imap_get_nstring(sio);
+  if(sio_getc(h->sio) != ']') return IMR_PROTOCOL;
+  if(sio_getc(h->sio) != ' ') return IMR_PROTOCOL;
+
+  tmp = imap_get_nstring(h->sio);
+  if(h->body_cb) {
+    if(tmp) h->body_cb(seqno, tmp, strlen(tmp), h->body_arg);
+    g_free(tmp);
+  } else {
+    CREATE_IMSG_IF_NEEDED(h, seqno);
+    msg = h->msg_cache[seqno-1];
+    g_free(msg->fetched_header_fields);
+    msg->fetched_header_fields = tmp;
+  }
   return IMR_OK;
 }
 
 static ImapResponse
 ir_msg_att_body(ImapMboxHandle *h, int c, unsigned seqno)
 {
-  ImapMessage *msg = h->msg_cache[seqno-1];
+  ImapMessage *msg;
   ImapResponse rc;
   char buf[19];	/* Just large enough to hold "HEADER.FIELDS.NOT". */
 
@@ -2723,24 +2911,26 @@ ir_msg_att_body(ImapMboxHandle *h, int c, unsigned seqno)
     c = sio_getc (h->sio);
     sio_ungetc (h->sio);
     if(isdigit (c)) {
-      rc = ir_body_section(h->sio, h->body_cb, h->body_arg);
+      rc = ir_body_section(h->sio, seqno, h->body_cb, h->body_arg);
       break;
     }
     c = imap_get_atom(h->sio, buf, sizeof buf);
     if (c == ']' &&
         g_ascii_strcasecmp(buf, "HEADER") == 0) {
       sio_ungetc (h->sio); /* put the ']' back */
-      rc = ir_body_section(h->sio, h->body_cb, h->body_arg);
+      rc = ir_body_section(h->sio, seqno, h->body_cb, h->body_arg);
     } else {
       if (c == ' ' && 
           (g_ascii_strcasecmp(buf, "HEADER.FIELDS") == 0 ||
            g_ascii_strcasecmp(buf, "HEADER.FIELDS.NOT") == 0))
-        rc = ir_body_header_fields(h->sio, msg);
+        rc = ir_body_header_fields(h, seqno);
       else
         rc = IMR_PROTOCOL;
     }
     break;
   case ' ':
+    CREATE_IMSG_IF_NEEDED(h, seqno);
+    msg = h->msg_cache[seqno-1];
     rc = ir_body(h->sio, sio_getc(h->sio),
                  msg->body ? NULL : (msg->body = imap_body_new()), 
 		 IMB_NON_EXTENSIBLE);
@@ -2753,11 +2943,13 @@ ir_msg_att_body(ImapMboxHandle *h, int c, unsigned seqno)
 static ImapResponse
 ir_msg_att_bodystructure(ImapMboxHandle *h, int c, unsigned seqno)
 {
-  ImapMessage *msg = h->msg_cache[seqno-1];
+  ImapMessage *msg;
   ImapResponse rc;
 
   switch(c) {
   case ' ':
+    CREATE_IMSG_IF_NEEDED(h, seqno);
+    msg = h->msg_cache[seqno-1];
     rc = ir_body(h->sio, sio_getc(h->sio),
                  msg->body ? NULL : (msg->body = imap_body_new()), 
 		 IMB_EXTENSIBLE);
@@ -2771,11 +2963,11 @@ static ImapResponse
 ir_msg_att_uid(ImapMboxHandle *h, int c, unsigned seqno)
 {
   char buf[12];
-  ImapMessage *msg = h->msg_cache[seqno-1];
   c = imap_get_atom(h->sio, buf, sizeof(buf));
 
   if(c!= -1) sio_ungetc(h->sio);
-  msg->uid = atoi(buf);
+  CREATE_IMSG_IF_NEEDED(h, seqno);
+  h->msg_cache[seqno-1]->uid = atoi(buf);
   return IMR_OK;
 }
 
@@ -2803,9 +2995,8 @@ ir_fetch_seq(ImapMboxHandle *h, unsigned seqno)
   int c;
   ImapResponse rc;
 
+  if(seqno<1 || seqno > h->exists) return IMR_PROTOCOL;
   if(sio_getc(h->sio) != '(') return IMR_PROTOCOL;
-  if(h->msg_cache[seqno-1] == NULL)
-    h->msg_cache[seqno-1] = imap_message_new();
   do {
     for(i=0; (c = sio_getc(h->sio)) != -1; i++) {
       c = toupper(c);
@@ -2873,6 +3064,7 @@ ir_thread_sub(ImapMboxHandle *h, GNode *parent, int last)
   ImapResponse rc = IMR_OK;
 
   c = imap_get_atom(h->sio, buf, sizeof(buf));
+
   seqno = atoi(buf);
   if(seqno == 0 && c == '(') {
       while (c == '(') {
@@ -2881,6 +3073,7 @@ ir_thread_sub(ImapMboxHandle *h, GNode *parent, int last)
 	      return rc;
 	  }
 	  c=sio_getc(h->sio);
+          if(c<0) return IMR_SEVERED;
       }
       return rc;
   }
