@@ -33,18 +33,17 @@
 #include <stdarg.h>
 #include <unistd.h>
 
-#include <libgnome/libgnome.h>
-
 #if ENABLE_LDAP
 #include <ldap.h>
 #endif
 
-#include "address-book-gpe.h"
 #include "libbalsa.h"
 #include "misc.h"
+#include "i18n.h"
 
 #ifdef BALSA_USE_THREADS
 static pthread_t main_thread_id;
+static pthread_t libbalsa_threads_id;
 #endif
 
 #define POP_SERVER "pop"
@@ -149,15 +148,14 @@ libbalsa_guess_email_address(void)
     /* Q: Find this location with configure? or at run-time? */
     static const gchar* MAILNAME_FILE = "/etc/mailname";
     char hostbuf[512];
+    FILE *mailname_in = NULL;
 
     gchar* preset, *domain;
     if(g_getenv("EMAIL") != NULL){                  /* 1. */
         preset = g_strdup(g_getenv("EMAIL"));
-    } else if(access(MAILNAME_FILE, F_OK) == 0){    /* 2. */
-        FILE *mailname_in = fopen(MAILNAME_FILE, "r");
-        fgets(hostbuf, 511, mailname_in);
-        hostbuf[strlen(hostbuf)-1] = '\0';
-        fclose(mailname_in);
+    } else if( (mailname_in = fopen(MAILNAME_FILE, "r")) != NULL
+              && fgets(hostbuf, sizeof(hostbuf)-1, mailname_in)){ /* 2. */
+        hostbuf[sizeof(hostbuf)-1] = '\0';
         preset = g_strconcat(g_get_user_name(), "@", hostbuf, NULL);
         
     }else if((domain = libbalsa_get_domainname())){ /* 3. */
@@ -167,6 +165,8 @@ libbalsa_guess_email_address(void)
         gethostname(hostbuf, 511);
         preset = g_strconcat(g_get_user_name(), "@", hostbuf, NULL);
     }
+    if (mailname_in)
+        fclose(mailname_in);
     return preset;
 }
 
@@ -182,8 +182,8 @@ libbalsa_guess_mail_spool(void)
     gchar *env;
     gchar *spool;
     static const gchar *guesses[] = {
-	"/var/spool/mail/",
 	"/var/mail/",
+	"/var/spool/mail/",
 	"/usr/spool/mail/",
 	"/usr/mail/",
 	NULL
@@ -207,7 +207,7 @@ libbalsa_guess_mail_spool(void)
      * ($HOME/mailbox) exists on
      * some systems, and it's a good enough default if we
      * can't guess it any other way. */
-    return gnome_util_prepend_user_home("mailbox");
+    return g_strconcat(g_get_home_dir(), "/mailbox", NULL);
 }
 
 /* Some more "guess" functions symmetric to libbalsa_guess_mail_spool()... */
@@ -307,14 +307,15 @@ gchar *libbalsa_guess_ldif_file()
     };
 
     for (i = 0; guesses[i] != NULL; i++) {
-	ldif = gnome_util_prepend_user_home(guesses[i]);
-	
+	ldif =  g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, 
+			    guesses[i], NULL);
 	if (g_file_test(ldif, G_FILE_TEST_EXISTS))
 	     return ldif;
 	  
 	g_free(ldif);
     }
-    return  gnome_util_prepend_user_home(guesses[0]); /* *** Or NULL */
+    return g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, 
+			guesses[i], NULL); /* *** Or NULL */
     
 }
 
@@ -342,7 +343,7 @@ libbalsa_assure_balsa_dir(void)
 }
 
 gchar*
-libbalsa_date_to_gchar(const time_t *date, const gchar *date_string)
+libbalsa_date_to_utf8(const time_t *date, const gchar *date_string)
 {
     struct tm footime;
     gchar rettime[128];
@@ -354,7 +355,7 @@ libbalsa_date_to_gchar(const time_t *date, const gchar *date_string)
 
     strftime(rettime, sizeof(rettime), date_string, &footime);
 
-    return g_strdup(rettime);
+    return g_locale_to_utf8(rettime, -1, NULL, NULL, NULL);
 }
 
 LibBalsaMessageStatus
@@ -422,7 +423,7 @@ static void
 x509_fingerprint (char *s, unsigned len, X509 * cert)
 {
     unsigned j, i, n, c;
-    char md[EVP_MAX_MD_SIZE];
+    unsigned char md[EVP_MAX_MD_SIZE];
 
 
     X509_digest(cert, EVP_md5(), md, &n);
@@ -680,37 +681,141 @@ static pthread_cond_t mailbox_cond = PTHREAD_COND_INITIALIZER;
 
 /* Lock/unlock a mailbox; no argument checking--we'll assume the caller
  * took care of that. 
- * We assume that this function is called for helper threads without gdk lock
- * since they usually do not have it but WITH for the main lock - which 
- * usually has it.
  */
+#define LIBBALSA_DEBUG_THREADS FALSE
 void
 libbalsa_lock_mailbox(LibBalsaMailbox * mailbox)
 {
     pthread_t thread_id = pthread_self();
-    gboolean is_main_thread =
-	(thread_id == main_thread_id);
+    gint count = 0;
+
+    if (thread_id == libbalsa_threads_id
+        && thread_id == mailbox->thread_id
+        && mailbox->lock > 0) {
+        /* We already have both locks, so we'll just hold on to both of
+         * them. */
+        ++mailbox->lock;
+#if LIBBALSA_DEBUG_THREADS
+        g_message("Avoided temporary gdk_threads_leave!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+        return;
+    }
+
+    while (thread_id == libbalsa_threads_id) {
+        ++count;
+#if LIBBALSA_DEBUG_THREADS
+        g_message("Temporary gdk_threads_leave!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+        gdk_threads_leave();
+    }
 
     pthread_mutex_lock(&mailbox_lock);
-    while (mailbox->lock && mailbox->thread_id != thread_id) {
-        if(is_main_thread) gdk_threads_leave();
-	pthread_cond_wait(&mailbox_cond, &mailbox_lock);
-        if(is_main_thread) gdk_threads_enter();
-    }
+
+    while (mailbox->lock && mailbox->thread_id != thread_id)
+        pthread_cond_wait(&mailbox_cond, &mailbox_lock);
+
     /* We'll assume that no-one would destroy a mailbox while we've been
      * trying to lock it. If they have, we have larger problems than
      * this reference! */
     mailbox->lock++;
     mailbox->thread_id = thread_id;
+
     pthread_mutex_unlock(&mailbox_lock);
+
+    while (--count >= 0) {
+        gdk_threads_enter();
+#if LIBBALSA_DEBUG_THREADS
+        g_message("...and gdk_threads_enter!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+    }
 }
 
 void
 libbalsa_unlock_mailbox(LibBalsaMailbox * mailbox)
 {
+    pthread_t self;
+
+    self = pthread_self();
+
     pthread_mutex_lock(&mailbox_lock);
-    if(!--mailbox->lock)
+
+    if (mailbox->lock == 0 || self != mailbox->thread_id) {
+	g_warning("Not holding mailbox lock!!!");
+        pthread_mutex_unlock(&mailbox_lock);
+	return;
+    }
+
+    if(--mailbox->lock == 0) {
         pthread_cond_broadcast(&mailbox_cond);
+        mailbox->thread_id = 0;
+    }
+
     pthread_mutex_unlock(&mailbox_lock);
 }
+
+/* Recursive mutex for gdk_threads_{enter,leave}. */
+static pthread_mutex_t libbalsa_threads_mutex;
+static guint libbalsa_threads_lock;
+
+static void
+libbalsa_threads_enter(void)
+{
+    pthread_t self;
+
+    self = pthread_self();
+
+    if (self != libbalsa_threads_id) {
+        pthread_mutex_lock(&libbalsa_threads_mutex);
+        libbalsa_threads_id = self;
+    }
+    ++libbalsa_threads_lock;
+}
+
+static void
+libbalsa_threads_leave(void)
+{
+    pthread_t self;
+
+    self = pthread_self();
+
+    if (libbalsa_threads_lock == 0 || self != libbalsa_threads_id) {
+	g_warning("Not holding gdk lock!!!");
+	return;
+    }
+
+    if (--libbalsa_threads_lock == 0) {
+	if (self != main_thread_id)
+	    gdk_display_flush(gdk_display_get_default());
+        libbalsa_threads_id = 0;
+        pthread_mutex_unlock(&libbalsa_threads_mutex);
+    }
+}
+
+void
+libbalsa_threads_init(void)
+{
+    pthread_mutex_init(&libbalsa_threads_mutex, NULL);
+    gdk_threads_set_lock_functions(G_CALLBACK(libbalsa_threads_enter),
+                                   G_CALLBACK(libbalsa_threads_leave));
+}
+
+void
+libbalsa_threads_destroy(void)
+{
+    pthread_mutex_destroy(&libbalsa_threads_mutex);
+}
+
+gboolean
+libbalsa_threads_has_lock(void)
+{
+    return libbalsa_threads_lock > 0
+        && libbalsa_threads_id == pthread_self();
+}
+
 #endif				/* BALSA_USE_THREADS */
+
+/* Initialized by the front end. */
+void (*libbalsa_progress_set_text) (LibBalsaProgress * progress,
+                                    const gchar * text, guint total);
+void (*libbalsa_progress_set_fraction) (LibBalsaProgress * progress,
+                                        gdouble fraction);
