@@ -30,10 +30,10 @@ typedef struct _LibBalsaMessage LibBalsaMessage;
 typedef struct _LibBalsaMessageHeaders LibBalsaMessageHeaders;
 typedef struct _LibBalsaMessageBody LibBalsaMessageBody;
 typedef struct _LibBalsaServer LibBalsaServer;
+typedef struct _LibBalsaSmtpServer LibBalsaSmtpServer;
 typedef struct _LibBalsaCondition LibBalsaCondition;
 
 
-#include "address.h"
 #include "message.h"
 #include "body.h"
 #include "files.h"
@@ -51,6 +51,9 @@ typedef struct _LibBalsaCondition LibBalsaCondition;
 #if ENABLE_LDAP
 #include "address-book-ldap.h"
 #endif
+#if HAVE_SQLITE
+#include "address-book-gpe.h"
+#endif
 
 #include "mailbox.h"
 #include "mailbox_local.h"
@@ -77,11 +80,16 @@ enum {
 enum {
     LIBBALSA_SCANNER_ERROR_IMAP,
     LIBBALSA_MAILBOX_APPEND_ERROR,
-    LIBBALSA_MAILBOX_AUTH_ERROR,
+    LIBBALSA_MAILBOX_AUTH_ERROR, /* retryable */
+    LIBBALSA_MAILBOX_AUTH_CANCELLED, /* do not try again */
     LIBBALSA_MAILBOX_COPY_ERROR,
+    LIBBALSA_MAILBOX_RENAME_ERROR,
+    LIBBALSA_MAILBOX_CREATE_ERROR,
+    LIBBALSA_MAILBOX_DELETE_ERROR,
     LIBBALSA_MAILBOX_NETWORK_ERROR,
     LIBBALSA_MAILBOX_OPEN_ERROR,
-    LIBBALSA_MAILBOX_TOOMANYOPEN_ERROR
+    LIBBALSA_MAILBOX_TOOMANYOPEN_ERROR,
+    LIBBALSA_MAILBOX_ACCESS_ERROR
 };
 
 
@@ -115,7 +123,7 @@ gchar *libbalsa_guess_ldif_file(void);
 gboolean libbalsa_ldap_exists(const gchar *server);
 
 void libbalsa_assure_balsa_dir(void);
-gchar* libbalsa_date_to_gchar(const time_t *date, const gchar *date_string);
+gchar* libbalsa_date_to_utf8(const time_t *date, const gchar *date_string);
 LibBalsaMessageStatus libbalsa_get_icon_from_flags(LibBalsaMessageFlag flags);
 
 #ifdef USE_TLS
@@ -128,11 +136,31 @@ void libbalsa_certs_destroy(void);
 #include <pthread.h>
 pthread_t libbalsa_get_main_thread(void);
 gboolean libbalsa_am_i_subthread(void);
+void libbalsa_threads_init(void);
+void libbalsa_threads_destroy(void);
+gboolean libbalsa_threads_has_lock(void);
 #else
 #define libbalsa_am_i_subthread() FALSE
+#define libbalsa_threads_has_lock() TRUE
 #endif /* BALSA_USE_THREADS */
 void libbalsa_message(const char *fmt, ...);
 gchar * libbalsa_rot(const gchar * pass);
 
+typedef enum {
+    LIBBALSA_PROGRESS_NO = 0,
+    LIBBALSA_PROGRESS_YES
+} LibBalsaProgress;
+#define LIBBALSA_PROGRESS_INIT LIBBALSA_PROGRESS_NO
+/* We will not use the progress bar if the number of increments is less
+ * than LIBBALSA_PROGRESS_MIN_COUNT, and we will not update the fraction
+ * if it has increased by less than LIBBALSA_PROGRESS_MIN_UPDATE. */
+#define LIBBALSA_PROGRESS_MIN_COUNT  400
+#define LIBBALSA_PROGRESS_MIN_UPDATE 0.02
+
+extern void (*libbalsa_progress_set_text) (LibBalsaProgress * progress,
+                                           const gchar * text,
+                                           guint total);
+extern void (*libbalsa_progress_set_fraction) (LibBalsaProgress * progress,
+                                               gdouble fraction);
 
 #endif                          /* __LIBBALSA_H__ */
