@@ -71,6 +71,7 @@ do {\
     }\
 } while (0)
 
+#define LIBBALSA_MAILBOX_UNTHREADED "libbalsa-mailbox-unthreaded"
 
 
 typedef enum {
@@ -86,7 +87,7 @@ typedef enum {
 typedef struct _SortTuple SortTuple;
 /* Sorting */
 struct _SortTuple {
-    gint offset;
+    guint offset;
     GNode *node;
 };
 
@@ -120,6 +121,12 @@ typedef enum {
     LB_MAILBOX_SHOW_TO
 } LibBalsaMailboxShow;
 
+typedef enum {
+    LB_MAILBOX_SUBSCRIBE_NO,
+    LB_MAILBOX_SUBSCRIBE_YES,
+    LB_MAILBOX_SUBSCRIBE_UNSET
+} LibBalsaMailboxSubscribe;
+
 typedef enum {
     LB_FETCH_RFC822_HEADERS = 1<<0, /* prepare all rfc822 headers */
     LB_FETCH_STRUCTURE      = 1<<1  /* prepare message structure */
@@ -153,7 +160,7 @@ typedef struct _LibBalsaMailboxClass LibBalsaMailboxClass;
 
 typedef struct _LibBalsaMailboxView LibBalsaMailboxView;
 struct _LibBalsaMailboxView {
-    LibBalsaAddress *mailing_list_address;
+    InternetAddressList *mailing_list_address;
     gchar *identity_name;
     LibBalsaMailboxThreadingType threading_type;
     /** filter is a frontend-specific code determining used view
@@ -166,14 +173,27 @@ struct _LibBalsaMailboxView {
     LibBalsaMailboxSortType      sort_type;
     LibBalsaMailboxSortFields    sort_field;
     LibBalsaMailboxShow          show;
-    unsigned exposed:1;
-    unsigned open:1;
-    unsigned in_sync:1;		/* view is in sync with config */
-    unsigned frozen:1;		/* don't update view if set    */
+    LibBalsaMailboxSubscribe     subscribe;
+    gboolean exposed;
+    gboolean open;
+    gboolean in_sync;		/* view is in sync with config */
+    gboolean frozen;		/* don't update view if set    */
+    gboolean used;		/* keep track of usage         */
 
 #ifdef HAVE_GPGME
     LibBalsaChkCryptoMode gpg_chk_mode;
 #endif
+
+    /* Display statistics:
+     * - total >= 0                both counts are valid;
+     * - total <  0 && unread == 0 unread is known to be zero;
+     * - total <  0 && unread >  0 unread is known to be > 0,
+     *                             but the count is not valid;
+     * - total <  0 && unread <  0 both are unknown.
+     */
+    int unread;
+    int total;
+    time_t mtime;       /* Mailbox mtime when counts were cached. */
 };
 
 struct _LibBalsaMailbox {
@@ -200,7 +220,8 @@ struct _LibBalsaMailbox {
     GPtrArray *mindex;  /* the basic message index used for index
                          * displaying/columns of GtkTreeModel interface
                          * and NOTHING else. */
-    GNode *msg_tree; /* the possibly filtered tree of messages */
+    GNode *msg_tree; /* the possibly filtered tree of messages;
+                      * gdk lock MUST BE HELD when accessing. */
     LibBalsaCondition *view_filter; /* to choose a subset of messages
                                      * to be displayed, e.g., only
                                      * undeleted. */
@@ -213,9 +234,16 @@ struct _LibBalsaMailbox {
                             */
     /* Associated filters (struct mailbox_filter) */
     GSList * filters;
+    gboolean filters_loaded;
 
     LibBalsaMailboxView *view;
     LibBalsaMailboxState state;
+
+    /* Whether to reassemble a message from its parts. */
+    gboolean no_reassemble;
+
+    /* Whether the tree has been changed since some event. */
+    gboolean msg_tree_changed;
 };
 
 /* Search iter */
@@ -239,7 +267,8 @@ struct _LibBalsaMailboxClass {
     gboolean (*open_mailbox) (LibBalsaMailbox * mailbox, GError **err);
     void (*close_mailbox) (LibBalsaMailbox * mailbox, gboolean expunge);
     LibBalsaMessage *(*get_message) (LibBalsaMailbox * mailbox, guint msgno);
-    void (*prepare_threading)(LibBalsaMailbox *mailbox, guint lo, guint hi);
+    void (*prepare_threading)(LibBalsaMailbox *mailbox, guint * msgnos,
+                              guint len);
     gboolean (*fetch_message_structure)(LibBalsaMailbox *mailbox,
                                         LibBalsaMessage * message,
                                         LibBalsaFetchFlag flags);
@@ -248,9 +277,10 @@ struct _LibBalsaMailboxClass {
     void (*release_message) (LibBalsaMailbox * mailbox,
 			     LibBalsaMessage * message);
     gboolean (*get_message_part) (LibBalsaMessage *message,
-				  LibBalsaMessageBody *part);
+				  LibBalsaMessageBody *part,
+                                  GError **err);
     GMimeStream *(*get_message_stream) (LibBalsaMailbox * mailbox,
-					LibBalsaMessage * message);
+                                        guint msgno);
 
     void (*check) (LibBalsaMailbox * mailbox);
 
@@ -263,8 +293,9 @@ struct _LibBalsaMailboxClass {
     void (*save_config) (LibBalsaMailbox * mailbox, const gchar * prefix);
     void (*load_config) (LibBalsaMailbox * mailbox, const gchar * prefix);
     gboolean (*sync) (LibBalsaMailbox * mailbox, gboolean expunge);
-    int (*add_message) (LibBalsaMailbox * mailbox, LibBalsaMessage * message,
-                        GError **err);
+    gboolean (*add_message) (LibBalsaMailbox * mailbox,
+                             GMimeStream * stream,
+                             LibBalsaMessageFlag flags, GError ** err);
     gboolean (*messages_change_flags) (LibBalsaMailbox * mailbox,
 				       GArray *msgnos,
 				       LibBalsaMessageFlag set,
@@ -285,6 +316,10 @@ struct _LibBalsaMailboxClass {
     void (*sort) (LibBalsaMailbox * mailbox, GArray *sort_array);
     gboolean (*close_backend)(LibBalsaMailbox * mailbox);
     guint (*total_messages)(LibBalsaMailbox * mailbox);
+    GArray *(*duplicate_msgnos) (LibBalsaMailbox * mailbox);
+#if BALSA_USE_THREADS
+    void (*lock_store) (LibBalsaMailbox * mailbox, gboolean lock);
+#endif                          /* BALSA_USE_THREADS */
 };
 
 GType libbalsa_mailbox_get_type(void);
@@ -301,6 +336,7 @@ gboolean libbalsa_mailbox_is_open(LibBalsaMailbox *mailbox);
 void libbalsa_mailbox_close(LibBalsaMailbox * mailbox, gboolean expunge);
 
 void libbalsa_mailbox_check(LibBalsaMailbox * mailbox);
+void libbalsa_mailbox_changed(LibBalsaMailbox * mailbox);
 void libbalsa_mailbox_set_unread_messages_flag(LibBalsaMailbox * mailbox,
 					       gboolean has_unread);
 void libbalsa_mailbox_progress_notify(LibBalsaMailbox * mailbox,
@@ -320,11 +356,11 @@ LibBalsaMessage *libbalsa_mailbox_get_message(LibBalsaMailbox * mailbox,
 					      guint msgno);
 
 /** libbalsa_mailbox_prepare_threading() requests prefetching of information
-    needed for client-size message threading.
-    lo and hi are related to currently set view.
+    needed for client-side message threading.
+    msgnos are related to currently set view.
 */
 void libbalsa_mailbox_prepare_threading(LibBalsaMailbox *mailbox,
-					guint lo, guint hi);
+					guint * msgnos, guint len);
 
 /** libbalsa_mailbox_fetch_message_structure() fetches detailed
     message structure for given message. It can also fetch all RFC822
@@ -347,17 +383,18 @@ void libbalsa_mailbox_release_message(LibBalsaMailbox * mailbox,
 void libbalsa_mailbox_set_msg_headers(LibBalsaMailbox * mailbox,
 				      LibBalsaMessage * message);
 
-/** libbalsa_mailbox_get_message_stream() returns an allocated block containing
-    selected, single part of the message.
+/** libbalsa_mailbox_get_message_part() ensures that a selected, single
+    part of the message is loaded.
 */
 gboolean libbalsa_mailbox_get_message_part(LibBalsaMessage    *message,
-					   LibBalsaMessageBody *part);
+					   LibBalsaMessageBody *part,
+                                           GError **err);
 
 /** libbalsa_mailbox_get_message_stream() returns a message stream associated
     with full RFC822 text of the message.
 */
 GMimeStream *libbalsa_mailbox_get_message_stream(LibBalsaMailbox * mailbox,
-						 LibBalsaMessage * message);
+						 guint msgno);
 
 
 /** libbalsa_mailbox_sync_storage() asks the mailbox to synchronise
@@ -401,11 +438,11 @@ void libbalsa_mailbox_run_filters_on_reception(LibBalsaMailbox * mailbox);
 
 void libbalsa_mailbox_save_config(LibBalsaMailbox * mailbox,
 				  const gchar * prefix);
-void libbalsa_mailbox_load_config(LibBalsaMailbox * mailbox,
-				  const gchar * prefix);
 
-int libbalsa_mailbox_copy_message(LibBalsaMessage *message,
-				  LibBalsaMailbox *dest, GError **err);
+gboolean libbalsa_mailbox_add_message(LibBalsaMailbox * mailbox,
+                                      GMimeStream * stream,
+                                      LibBalsaMessageFlag flags,
+                                      GError ** err);
 gboolean libbalsa_mailbox_close_backend(LibBalsaMailbox * mailbox);
 
 /* Message number-list methods */
@@ -427,6 +464,10 @@ gboolean libbalsa_mailbox_messages_move(LibBalsaMailbox * mailbox,
 GType libbalsa_mailbox_type_from_path(const gchar * filename);
 
 guint libbalsa_mailbox_total_messages(LibBalsaMailbox * mailbox);
+gboolean libbalsa_mailbox_can_move_duplicates(LibBalsaMailbox * mailbox);
+void libbalsa_mailbox_move_duplicates(LibBalsaMailbox * mailbox,
+                                      LibBalsaMailbox * dest,
+                                      GError ** err);
 
 /*
  * Mailbox views-related functions.
@@ -476,6 +517,9 @@ void libbalsa_mailbox_set_sort_field(LibBalsaMailbox * mailbox,
 				     LibBalsaMailboxSortFields sort_field);
 gboolean libbalsa_mailbox_set_show(LibBalsaMailbox * mailbox,
 				   LibBalsaMailboxShow show);
+gboolean libbalsa_mailbox_set_subscribe(LibBalsaMailbox * mailbox,
+                                        LibBalsaMailboxSubscribe
+                                        subscribe);
 void libbalsa_mailbox_set_exposed(LibBalsaMailbox * mailbox,
 				  gboolean exposed);
 void libbalsa_mailbox_set_open(LibBalsaMailbox * mailbox, gboolean open);
@@ -485,9 +529,12 @@ void libbalsa_mailbox_set_frozen(LibBalsaMailbox * mailbox, gboolean frozen);
 gboolean libbalsa_mailbox_set_crypto_mode(LibBalsaMailbox * mailbox,
 					  LibBalsaChkCryptoMode gpg_chk_mode);
 #endif
+void libbalsa_mailbox_set_unread(LibBalsaMailbox * mailbox, gint unread);
+void libbalsa_mailbox_set_total (LibBalsaMailbox * mailbox, gint total);
+void libbalsa_mailbox_set_mtime (LibBalsaMailbox * mailbox, time_t mtime);
 
-LibBalsaAddress *libbalsa_mailbox_get_mailing_list_address(LibBalsaMailbox
-							   * mailbox);
+InternetAddressList
+    *libbalsa_mailbox_get_mailing_list_address(LibBalsaMailbox * mailbox);
 const gchar *libbalsa_mailbox_get_identity_name(LibBalsaMailbox * mailbox);
 LibBalsaMailboxThreadingType
 libbalsa_mailbox_get_threading_type(LibBalsaMailbox * mailbox);
@@ -496,6 +543,8 @@ LibBalsaMailboxSortType libbalsa_mailbox_get_sort_type(LibBalsaMailbox *
 LibBalsaMailboxSortFields libbalsa_mailbox_get_sort_field(LibBalsaMailbox *
 							  mailbox);
 LibBalsaMailboxShow libbalsa_mailbox_get_show(LibBalsaMailbox * mailbox);
+LibBalsaMailboxSubscribe libbalsa_mailbox_get_subscribe(LibBalsaMailbox *
+                                                        mailbox);
 gboolean libbalsa_mailbox_get_exposed(LibBalsaMailbox * mailbox);
 gboolean libbalsa_mailbox_get_open(LibBalsaMailbox * mailbox);
 gint libbalsa_mailbox_get_filter(LibBalsaMailbox * mailbox);
@@ -503,17 +552,23 @@ gboolean libbalsa_mailbox_get_frozen(LibBalsaMailbox * mailbox);
 #ifdef HAVE_GPGME
 LibBalsaChkCryptoMode libbalsa_mailbox_get_crypto_mode(LibBalsaMailbox * mailbox);
 #endif
+gint libbalsa_mailbox_get_unread(LibBalsaMailbox * mailbox);
+gint libbalsa_mailbox_get_total (LibBalsaMailbox * mailbox);
+time_t libbalsa_mailbox_get_mtime(LibBalsaMailbox * mailbox);
 
 /** force update of given msgno */
 void libbalsa_mailbox_msgno_changed(LibBalsaMailbox  *mailbox, guint seqno);
-void libbalsa_mailbox_msgno_inserted(LibBalsaMailbox *mailbox, guint seqno);
+void libbalsa_mailbox_msgno_inserted(LibBalsaMailbox * mailbox,
+                                     guint seqno, GNode * parent,
+                                     GNode ** sibling);
 void libbalsa_mailbox_msgno_removed(LibBalsaMailbox  *mailbox, guint seqno);
 void libbalsa_mailbox_msgno_filt_in(LibBalsaMailbox * mailbox, guint seqno);
 void libbalsa_mailbox_msgno_filt_out(LibBalsaMailbox * mailbox, guint seqno);
 void libbalsa_mailbox_msgno_filt_check(LibBalsaMailbox * mailbox,
 				       guint seqno,
 				       LibBalsaMailboxSearchIter
-				       * search_iter);
+				       * search_iter,
+				       gboolean hold_selected);
 
 /* Search */
 gboolean libbalsa_mailbox_msgno_find(LibBalsaMailbox * mailbox,
@@ -559,6 +614,22 @@ const gchar *libbalsa_mailbox_msgno_get_subject(LibBalsaMailbox * mailbox,
 void libbalsa_mailbox_msgno_update_attach(LibBalsaMailbox * mailbox,
 					  guint msgno,
 					  LibBalsaMessage * message);
+void libbalsa_mailbox_cache_message(LibBalsaMailbox * mailbox, guint msgno,
+                                    LibBalsaMessage * message);
+
+#if BALSA_USE_THREADS
+
+/* Lock and unlock the mail store--currently, a no-op except for mbox.
+ */
+void libbalsa_mailbox_lock_store  (LibBalsaMailbox * mailbox);
+void libbalsa_mailbox_unlock_store(LibBalsaMailbox * mailbox);
+
+#else                           /* BALSA_USE_THREADS */
+
+#define libbalsa_mailbox_lock_store(mailbox)
+#define libbalsa_mailbox_unlock_store(mailbox)
+
+#endif                          /* BALSA_USE_THREADS */
 
 /* columns ids */
 typedef enum {
