@@ -24,16 +24,20 @@
 
 #define _XOPEN_SOURCE          500
 #define _XOPEN_SOURCE_EXTENDED 1
-#include <libgnome/libgnome.h>
 
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include "libbalsa.h"
+#include "libbalsa-conf.h"
 #include "misc.h"
 #include "libbalsa_private.h"
+#include "mime-stream-shared.h"
+#include "i18n.h"
 
 struct message_info {
     char *key;
@@ -42,6 +46,11 @@ struct message_info {
     LibBalsaMessageFlag flags;  	/* May have pseudo-flags */
     LibBalsaMessageFlag orig_flags;     /* Has only real flags */
     LibBalsaMessage *message;
+
+    /* The message's order when parsing; needed for saving the message
+     * tree in a form that will match the msgnos when the mailbox is
+     * reopened. */
+    guint fileno;
 };
 #define REAL_FLAGS(flags) (flags & LIBBALSA_MESSAGE_FLAGS_REAL)
 #define FLAGS_REALLY_DIFFER(orig_flags, flags) \
@@ -54,15 +63,17 @@ static LibBalsaMailboxLocalClass *parent_class = NULL;
 
 static void libbalsa_mailbox_maildir_class_init(LibBalsaMailboxMaildirClass *klass);
 static void libbalsa_mailbox_maildir_init(LibBalsaMailboxMaildir * mailbox);
+
+/* Object class method */
 static void libbalsa_mailbox_maildir_finalize(GObject * object);
+
+/* Mailbox class methods */
 static void libbalsa_mailbox_maildir_load_config(LibBalsaMailbox * mailbox,
 						 const gchar * prefix);
 
 static GMimeStream *libbalsa_mailbox_maildir_get_message_stream(LibBalsaMailbox *
 							   mailbox,
-							   LibBalsaMessage *
-							   message);
-static void libbalsa_mailbox_maildir_remove_files(LibBalsaMailboxLocal *mailbox);
+							   guint msgno);
 
 static gboolean libbalsa_mailbox_maildir_open(LibBalsaMailbox * mailbox,
 					      GError **err);
@@ -72,18 +83,17 @@ static void libbalsa_mailbox_maildir_close_mailbox(LibBalsaMailbox *
 static void libbalsa_mailbox_maildir_check(LibBalsaMailbox * mailbox);
 static gboolean libbalsa_mailbox_maildir_sync(LibBalsaMailbox * mailbox,
                                               gboolean expunge);
-static struct message_info *message_info_from_msgno(
-						  LibBalsaMailbox * mailbox,
-						  guint msgno);
-static LibBalsaMessage *libbalsa_mailbox_maildir_get_message(LibBalsaMailbox * mailbox,
+static LibBalsaMessage *lbm_maildir_get_message(LibBalsaMailbox * mailbox,
 						     guint msgno);
 static gboolean
 libbalsa_mailbox_maildir_fetch_message_structure(LibBalsaMailbox * mailbox,
 						 LibBalsaMessage * message,
 						 LibBalsaFetchFlag flags);
-static int libbalsa_mailbox_maildir_add_message(LibBalsaMailbox * mailbox,
-						LibBalsaMessage * message,
-                                                GError **err);
+static gboolean libbalsa_mailbox_maildir_add_message(LibBalsaMailbox *
+                                                     mailbox,
+                                                     GMimeStream * stream,
+                                                     LibBalsaMessageFlag
+                                                     flags, GError ** err);
 static gboolean
 libbalsa_mailbox_maildir_messages_change_flags(LibBalsaMailbox * mailbox,
                                                GArray * msgnos,
@@ -97,10 +107,22 @@ libbalsa_mailbox_maildir_msgno_has_flags(LibBalsaMailbox * mailbox,
 static guint
 libbalsa_mailbox_maildir_total_messages(LibBalsaMailbox * mailbox);
 
+/* MailboxLocal class methods */
+static gint lbm_maildir_check_files(const gchar * path, gboolean create);
+static void lbm_maildir_set_path(LibBalsaMailboxLocal * local,
+                                 const gchar * path);
+static void lbm_maildir_remove_files(LibBalsaMailboxLocal * local);
+static LibBalsaMessageFlag lbm_maildir_load_message(LibBalsaMailboxLocal *
+                                                    local, guint msgno,
+                                                    LibBalsaMessage **
+                                                    msg);
+static guint lbm_maildir_fileno(LibBalsaMailboxLocal * local, guint msgno);
+
 /* util functions */
+static struct message_info *message_info_from_msgno(LibBalsaMailboxMaildir
+                                                    * mdir, guint msgno);
 static LibBalsaMessageFlag parse_filename(const gchar *subdir,
 					  const gchar *filename);
-static void parse_mailbox(LibBalsaMailbox * mailbox, const gchar *subdir);
 static void free_message_info(struct message_info *msg_info);
 static int libbalsa_mailbox_maildir_open_temp (const gchar *dest_path,
 					  char **name_used);
@@ -158,7 +180,7 @@ libbalsa_mailbox_maildir_class_init(LibBalsaMailboxMaildirClass * klass)
     libbalsa_mailbox_class->sync = libbalsa_mailbox_maildir_sync;
     libbalsa_mailbox_class->close_mailbox =
 	libbalsa_mailbox_maildir_close_mailbox;
-    libbalsa_mailbox_class->get_message = libbalsa_mailbox_maildir_get_message;
+    libbalsa_mailbox_class->get_message = lbm_maildir_get_message;
     libbalsa_mailbox_class->fetch_message_structure =
 	libbalsa_mailbox_maildir_fetch_message_structure;
     libbalsa_mailbox_class->add_message = libbalsa_mailbox_maildir_add_message;
@@ -169,110 +191,110 @@ libbalsa_mailbox_maildir_class_init(LibBalsaMailboxMaildirClass * klass)
     libbalsa_mailbox_class->total_messages =
 	libbalsa_mailbox_maildir_total_messages;
 
-    libbalsa_mailbox_local_class->remove_files = 
-	libbalsa_mailbox_maildir_remove_files;
-    libbalsa_mailbox_local_class->load_message =
-        libbalsa_mailbox_maildir_get_message;
+    libbalsa_mailbox_local_class->check_files  = lbm_maildir_check_files;
+    libbalsa_mailbox_local_class->set_path     = lbm_maildir_set_path;
+    libbalsa_mailbox_local_class->remove_files = lbm_maildir_remove_files;
+    libbalsa_mailbox_local_class->load_message = lbm_maildir_load_message;
+    libbalsa_mailbox_local_class->fileno       = lbm_maildir_fileno;
 }
 
 static void
-libbalsa_mailbox_maildir_init(LibBalsaMailboxMaildir * mailbox)
+libbalsa_mailbox_maildir_init(LibBalsaMailboxMaildir * mdir)
 {
 }
 
-static void lbm_maildir_set_subdirs(LibBalsaMailboxMaildir * mdir,
-                                    const gchar * path);
-
-gint
-libbalsa_mailbox_maildir_create(const gchar * path, gboolean create,
-                                LibBalsaMailboxMaildir * mdir)
+static gint
+lbm_maildir_check_files(const gchar * path, gboolean create)
 {
-    gint exists;
-    GType magic_type;
+    g_return_val_if_fail(path != NULL, -1);
 
-    g_return_val_if_fail( path != NULL, -1);
+    if (access(path, F_OK) == 0) {
+        /* File exists. Check if it is a maildir... */
+        if (libbalsa_mailbox_type_from_path(path) !=
+            LIBBALSA_TYPE_MAILBOX_MAILDIR) {
+            libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                                 _("Mailbox %s does not appear to be a Maildir mailbox."),
+                                 path);
+            return -1;
+        }
+    } else if (create) {
+        char tmp[_POSIX_PATH_MAX];
 
-    exists = access(path, F_OK);
-    if ( exists == 0 ) {
-	/* File exists. Check if it is a maildir... */
-	
-	magic_type = libbalsa_mailbox_type_from_path(path);
-	if ( magic_type != LIBBALSA_TYPE_MAILBOX_MAILDIR ) {
-	    libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-				 _("Mailbox %s does not appear to be a Maildir mailbox."), path);
-	    return(-1);
-	}
-    } else {
-	if(create) {    
-	    char tmp[_POSIX_PATH_MAX];
-	    
-	    if (mkdir (path, S_IRWXU)) {
-		libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-				     _("Could not create a MailDir directory at %s (%s)"), path, strerror(errno) );
-		return (-1);
-	    }
-	    
-	    snprintf (tmp, sizeof (tmp), "%s/cur", path);
-	    if (mkdir (tmp, S_IRWXU)) {
-		libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-				     _("Could not create a MailDir at %s (%s)"), path, strerror(errno) );
-		rmdir (path);
-		return (-1);
-	    }
-	    
-	    snprintf (tmp, sizeof (tmp), "%s/new", path);
-	    if (mkdir (tmp, S_IRWXU)) {
-		libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-				     _("Could not create a MailDir at %s (%s)"), path, strerror(errno) );
-		snprintf (tmp, sizeof (tmp), "%s/cur", path);
-		rmdir (tmp);
-		rmdir (path);
-		return (-1);
-	    }
-	    
-	    snprintf (tmp, sizeof (tmp), "%s/tmp", path);
-	    if (mkdir (tmp, S_IRWXU)) {
-		libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-				     _("Could not create a MailDir at %s (%s)"), path, strerror(errno) );
-		snprintf (tmp, sizeof (tmp), "%s/cur", path);
-		rmdir (tmp);
-		snprintf (tmp, sizeof (tmp), "%s/new", path);
-		rmdir (tmp);
-		rmdir (path);
-		return (-1);
-	    }
-	} else 
-	    return(-1);
-    }
-    lbm_maildir_set_subdirs(mdir, path);
-    return(0);
+        if (mkdir(path, S_IRWXU)) {
+            libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                                 _("Could not create a MailDir directory at %s (%s)"),
+                                 path, strerror(errno));
+            return -1;
+        }
+
+        snprintf(tmp, sizeof(tmp), "%s/cur", path);
+        if (mkdir(tmp, S_IRWXU)) {
+            libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                                 _("Could not create a MailDir at %s (%s)"),
+                                 path, strerror(errno));
+            rmdir(path);
+            return -1;
+        }
+
+        snprintf(tmp, sizeof(tmp), "%s/new", path);
+        if (mkdir(tmp, S_IRWXU)) {
+            libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                                 _("Could not create a MailDir at %s (%s)"),
+                                 path, strerror(errno));
+            snprintf(tmp, sizeof(tmp), "%s/cur", path);
+            rmdir(tmp);
+            rmdir(path);
+            return -1;
+        }
+
+        snprintf(tmp, sizeof(tmp), "%s/tmp", path);
+        if (mkdir(tmp, S_IRWXU)) {
+            libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                                 _("Could not create a MailDir at %s (%s)"),
+                                 path, strerror(errno));
+            snprintf(tmp, sizeof(tmp), "%s/cur", path);
+            rmdir(tmp);
+            snprintf(tmp, sizeof(tmp), "%s/new", path);
+            rmdir(tmp);
+            rmdir(path);
+            return -1;
+        }
+    } else
+        return -1;
+
+    return 0;
 }
 
 static void
 lbm_maildir_set_subdirs(LibBalsaMailboxMaildir * mdir, const gchar * path)
 {
-    mdir->curdir = g_strdup_printf("%s/cur", path);
-    mdir->newdir = g_strdup_printf("%s/new", path);
-    mdir->tmpdir = g_strdup_printf("%s/tmp", path);
+    g_free(mdir->curdir);
+    mdir->curdir = g_build_filename(path, "cur", NULL);
+    g_free(mdir->newdir);
+    mdir->newdir = g_build_filename(path, "new", NULL);
+    g_free(mdir->tmpdir);
+    mdir->tmpdir = g_build_filename(path, "tmp", NULL);
+}
+
+static void
+lbm_maildir_set_path(LibBalsaMailboxLocal * local, const gchar * path)
+{
+    lbm_maildir_set_subdirs(LIBBALSA_MAILBOX_MAILDIR(local), path);
 }
 
 GObject *
 libbalsa_mailbox_maildir_new(const gchar * path, gboolean create)
 {
     LibBalsaMailbox *mailbox;
-    LibBalsaMailboxMaildir *mdir;
 
     mailbox = g_object_new(LIBBALSA_TYPE_MAILBOX_MAILDIR, NULL);
 
     mailbox->is_directory = TRUE;
 
-    LIBBALSA_MAILBOX(mailbox)->url = g_strconcat("file://", path, NULL);
-
-
-    mdir = LIBBALSA_MAILBOX_MAILDIR(mailbox);
-    if(libbalsa_mailbox_maildir_create(path, create, mdir) < 0) {
-	g_object_unref(G_OBJECT(mailbox));
-	return NULL;
+    if (libbalsa_mailbox_local_set_path(LIBBALSA_MAILBOX_LOCAL(mailbox),
+                                        path, create) != 0) {
+        g_object_unref(mailbox);
+        return NULL;
     }
 
     return G_OBJECT(mailbox);
@@ -300,7 +322,7 @@ libbalsa_mailbox_maildir_load_config(LibBalsaMailbox * mailbox,
     LibBalsaMailboxMaildir *mdir = LIBBALSA_MAILBOX_MAILDIR(mailbox);
     gchar *path;
 
-    path = gnome_config_get_string("Path");
+    path = libbalsa_conf_get_string("Path");
     lbm_maildir_set_subdirs(mdir, path);
     g_free(path);
 
@@ -309,13 +331,14 @@ libbalsa_mailbox_maildir_load_config(LibBalsaMailbox * mailbox,
 
 static GMimeStream *
 libbalsa_mailbox_maildir_get_message_stream(LibBalsaMailbox * mailbox,
-					    LibBalsaMessage * message)
+					    guint msgno)
 {
     struct message_info *msg_info;
 
     g_return_val_if_fail(MAILBOX_OPEN(mailbox), NULL);
 
-    msg_info = message_info_from_msgno(mailbox, message->msgno);
+    msg_info =
+        message_info_from_msgno((LibBalsaMailboxMaildir *) mailbox, msgno);
     if (!msg_info)
 	return NULL;
 
@@ -325,7 +348,7 @@ libbalsa_mailbox_maildir_get_message_stream(LibBalsaMailbox * mailbox,
 }
 
 static void
-libbalsa_mailbox_maildir_remove_files(LibBalsaMailboxLocal *mailbox)
+lbm_maildir_remove_files(LibBalsaMailboxLocal *mailbox)
 {
     const gchar* path;
     g_return_if_fail(LIBBALSA_IS_MAILBOX_MAILDIR(mailbox));
@@ -343,6 +366,7 @@ libbalsa_mailbox_maildir_remove_files(LibBalsaMailboxLocal *mailbox)
 			     _("Could not remove %s:\n%s"),
 			     path, strerror(errno));
     }
+    LIBBALSA_MAILBOX_LOCAL_CLASS(parent_class)->remove_files(mailbox);
 }
 
 static LibBalsaMessageFlag parse_filename(const gchar *subdir,
@@ -384,7 +408,8 @@ static LibBalsaMessageFlag parse_filename(const gchar *subdir,
     return flags;
 }
 
-static void parse_mailbox(LibBalsaMailbox * mailbox, const gchar *subdir)
+static void lbm_maildir_parse(LibBalsaMailboxMaildir * mdir,
+                              const gchar *subdir, guint * fileno)
 {
     gchar *path;
     GDir *dir;
@@ -394,10 +419,10 @@ static void parse_mailbox(LibBalsaMailbox * mailbox, const gchar *subdir)
     gchar *key;
     gchar *p;
     LibBalsaMessageFlag flags;
-    LibBalsaMailboxMaildir *mdir = LIBBALSA_MAILBOX_MAILDIR(mailbox);
+    LibBalsaMailbox *mailbox = (LibBalsaMailbox *) mdir;
 
-    path = g_strdup_printf("%s/%s", libbalsa_mailbox_local_get_path(mailbox),
-			   subdir);
+    path = g_build_filename(libbalsa_mailbox_local_get_path(mailbox),
+			    subdir, NULL);
     dir = g_dir_open(path, 0, NULL);
     g_free(path);
     if (dir == NULL)
@@ -420,9 +445,14 @@ static void parse_mailbox(LibBalsaMailbox * mailbox, const gchar *subdir)
 	msg_info = g_hash_table_lookup(messages_info, key);
 	if (msg_info) {
 	    g_free(key);
+	    g_free(msg_info->filename);
+	    msg_info->filename = g_strdup(filename);
 	    if (FLAGS_REALLY_DIFFER(msg_info->orig_flags, flags)) {
-		g_warning("Message flags for \"%s\" changed\n",
+#undef DEBUG /* #define DEBUG TRUE */
+#ifdef DEBUG
+		g_message("Message flags for \"%s\" changed\n",
                           msg_info->key);
+#endif
 		msg_info->orig_flags = flags;
 	    }
 	} else {
@@ -434,17 +464,29 @@ static void parse_mailbox(LibBalsaMailbox * mailbox, const gchar *subdir)
 	    msg_info->key=key;
 	    msg_info->filename=g_strdup(filename);
 	    msg_info->flags = msg_info->orig_flags = flags;
+	    msg_info->fileno = 0;
 	}
 	msg_info->subdir = subdir;
+        if (!msg_info->fileno)
+            /* First time we saw this key. */
+	    msg_info->fileno = ++*fileno;
     }
     g_dir_close(dir);
 }
 
 static void
-parse_mailbox_subdirs(LibBalsaMailbox * mailbox)
+lbm_maildir_parse_subdirs(LibBalsaMailboxMaildir * mdir)
 {
-    parse_mailbox(mailbox, "cur");
-    parse_mailbox(mailbox, "new");
+    guint msgno, fileno = 0;
+
+    for (msgno = mdir->msgno_2_msg_info->len; msgno > 0; --msgno) {
+        struct message_info *msg_info =
+            message_info_from_msgno(mdir, msgno);
+        msg_info->fileno = 0;
+    }
+
+    lbm_maildir_parse(mdir, "cur", &fileno);
+    lbm_maildir_parse(mdir, "new", &fileno);
 }
 
 static gboolean
@@ -469,15 +511,8 @@ libbalsa_mailbox_maildir_open(LibBalsaMailbox * mailbox, GError **err)
 				  NULL, (GDestroyNotify)free_message_info);
     mdir->msgno_2_msg_info = g_ptr_array_new();
 
-    mdir->mtime = st.st_mtime;
-    mdir->mtime_cur = 0;
-    mdir->mtime_new = 0;
-    if (stat(mdir->curdir, &st) != -1) {
-	mdir->mtime_cur = st.st_mtime;
-    }
-    if (stat(mdir->newdir, &st) != -1) {
-	mdir->mtime_new = st.st_mtime;
-    }
+    if (stat(mdir->tmpdir, &st) != -1)
+	libbalsa_mailbox_set_mtime(mailbox, st.st_mtime);
 
     mailbox->readonly = 
 	!(access(mdir->curdir, W_OK) == 0 &&
@@ -485,7 +520,7 @@ libbalsa_mailbox_maildir_open(LibBalsaMailbox * mailbox, GError **err)
           access(mdir->tmpdir, W_OK) == 0);
 
     mailbox->unread_messages = 0;
-    parse_mailbox_subdirs(mailbox);
+    lbm_maildir_parse_subdirs(mdir);
     LIBBALSA_MAILBOX_LOCAL(mailbox)->sync_time = 0;
     LIBBALSA_MAILBOX_LOCAL(mailbox)->sync_cnt  = 1;
 #ifdef DEBUG
@@ -524,12 +559,12 @@ lbm_maildir_check(const gchar * subdir)
 static void
 libbalsa_mailbox_maildir_check(LibBalsaMailbox * mailbox)
 {
-    struct stat st, st_cur, st_new;
-    int modified = 0;
+    struct stat st;
     LibBalsaMailboxMaildir *mdir;
     guint renumber, msgno;
     struct message_info *msg_info;
     const gchar *path;
+    time_t mtime;
 
     g_assert(LIBBALSA_IS_MAILBOX_MAILDIR(mailbox));
 
@@ -537,34 +572,16 @@ libbalsa_mailbox_maildir_check(LibBalsaMailbox * mailbox)
 
     if (stat(mdir->tmpdir, &st) == -1)
 	return;
-    if (mdir->mtime == 0)
-	/* First check--just cache the mtime. */
-	mdir->mtime = st.st_mtime;
-    else if (st.st_mtime > mdir->mtime)
-	modified = 1;
 
-    if (stat(mdir->curdir, &st_cur) == -1)
-	return;
-    if (mdir->mtime_cur == 0)
+    if ((mtime = libbalsa_mailbox_get_mtime(mailbox)) == 0) {
 	/* First check--just cache the mtime. */
-	mdir->mtime_cur = st_cur.st_mtime;
-    else if (st_cur.st_mtime > mdir->mtime_cur)
-	modified = 1;
-
-    if (stat(mdir->newdir, &st_new) == -1)
-	return;
-    if (mdir->mtime_new == 0)
-	/* First check--just cache the mtime. */
-	mdir->mtime_new = st_new.st_mtime;
-    else if (st_new.st_mtime > mdir->mtime_new)
-	modified = 1;
-
-    if (!modified)
+	libbalsa_mailbox_set_mtime(mailbox, st.st_mtime);
+        return;
+    }
+    if (st.st_mtime <= mtime)
 	return;
 
-    mdir->mtime = st.st_mtime;
-    mdir->mtime_cur = st_cur.st_mtime;
-    mdir->mtime_new = st_new.st_mtime;
+    libbalsa_mailbox_set_mtime(mailbox, st.st_mtime);
 
     if (!MAILBOX_OPEN(mailbox)) {
 	libbalsa_mailbox_set_unread_messages_flag(mailbox,
@@ -582,7 +599,7 @@ libbalsa_mailbox_maildir_check(LibBalsaMailbox * mailbox)
     for (msgno = 1; msgno <= mdir->msgno_2_msg_info->len; ) {
 	gchar *filename;
 
-        msg_info = message_info_from_msgno(mailbox, msgno);
+        msg_info = message_info_from_msgno(mdir, msgno);
 	filename = g_build_filename(path, msg_info->subdir,
 				    msg_info->filename, NULL);
 	if (access(filename, F_OK) == 0)
@@ -598,13 +615,13 @@ libbalsa_mailbox_maildir_check(LibBalsaMailbox * mailbox)
 	g_free(filename);
     }
     for (msgno = renumber; msgno <= mdir->msgno_2_msg_info->len; msgno++) {
-	msg_info = message_info_from_msgno(mailbox, msgno);
+	msg_info = message_info_from_msgno(mdir, msgno);
 	if (msg_info->message)
 	    msg_info->message->msgno = msgno;
     }
 
     msgno = mdir->msgno_2_msg_info->len;
-    parse_mailbox_subdirs(mailbox);
+    lbm_maildir_parse_subdirs(mdir);
     libbalsa_mailbox_local_load_messages(mailbox, msgno);
 }
 
@@ -629,37 +646,48 @@ libbalsa_mailbox_maildir_close_mailbox(LibBalsaMailbox * mailbox,
                                        gboolean expunge)
 {
     LibBalsaMailboxMaildir *mdir = LIBBALSA_MAILBOX_MAILDIR(mailbox);
-    guint len = 0;
+    guint len;
 
-    if (mdir->msgno_2_msg_info)
-	len = mdir->msgno_2_msg_info->len;
+    g_assert(mdir->msgno_2_msg_info != NULL);
+    g_assert(mdir->messages_info != NULL);
+
+    len = mdir->msgno_2_msg_info->len;
     libbalsa_mailbox_maildir_sync(mailbox, expunge);
+    if (mdir->msgno_2_msg_info->len != len)
+        libbalsa_mailbox_changed(mailbox);
 
-    if (mdir->messages_info) {
-	g_hash_table_destroy(mdir->messages_info);
-	mdir->messages_info = NULL;
-    }
-    if (mdir->msgno_2_msg_info) {
-	if (mdir->msgno_2_msg_info->len != len)
-	    g_signal_emit_by_name(mailbox, "changed");
-	g_ptr_array_free(mdir->msgno_2_msg_info, TRUE);
-	mdir->msgno_2_msg_info = NULL;
-    }
     if (LIBBALSA_MAILBOX_CLASS(parent_class)->close_mailbox)
         LIBBALSA_MAILBOX_CLASS(parent_class)->close_mailbox(mailbox,
                                                             expunge);
+
+    /* Now it's safe to free the message info. */
+    g_hash_table_destroy(mdir->messages_info);
+    mdir->messages_info = NULL;
+
+    g_ptr_array_free(mdir->msgno_2_msg_info, TRUE);
+    mdir->msgno_2_msg_info = NULL;
 }
 
-static int libbalsa_mailbox_maildir_open_temp (const gchar *dest_path,
-					  char **name_used)
+static gchar *
+lbm_mdir_get_key(void)
+{
+    static int counter;
+
+    return g_strdup_printf("%s-%d-%d", g_get_user_name(), getpid(),
+                           counter++);
+}
+
+static int
+libbalsa_mailbox_maildir_open_temp(const gchar *dest_path,
+                                   char **name_used)
 {
     int fd;
-    static int counter;
     gchar *filename;
 
     do {
-	filename = g_strdup_printf("%s/tmp/%s-%d-%d", dest_path,
-	   		      g_get_user_name(), getpid (), counter++);
+	gchar *key = lbm_mdir_get_key();
+	filename = g_build_filename(dest_path, "tmp", key, NULL);
+	g_free(key);
 	fd = open (filename, O_WRONLY | O_EXCL | O_CREAT, 0600);
 	if (fd == -1)
 	{
@@ -679,6 +707,7 @@ maildir_sync_add(struct message_info *msg_info, const gchar * path)
     gchar new_flags[10];
     int len;
     const gchar *subdir;
+    gchar *filename;
     gchar *new;
     gchar *orig;
     gboolean retval = TRUE;
@@ -697,22 +726,39 @@ maildir_sync_add(struct message_info *msg_info, const gchar * path)
     new_flags[len] = '\0';
 
     subdir = msg_info->flags & LIBBALSA_MESSAGE_FLAG_RECENT ? "new" : "cur";
-    new = g_strdup_printf("%s/%s/%s%s%s", path, subdir, msg_info->key,
-			  (len ? ":2," : ""), new_flags);
-    orig = g_strdup_printf("%s/%s/%s", path, msg_info->subdir,
-			   msg_info->filename);
+    filename =
+        g_strconcat(msg_info->key, (len ? ":2," : ""), new_flags, NULL);
+    new = g_build_filename(path, subdir, filename, NULL);
+    orig =
+        g_build_filename(path, msg_info->subdir, msg_info->filename, NULL);
     if (strcmp(orig, new)) {
+        while (g_file_test(new, G_FILE_TEST_EXISTS)) {
+#ifdef DEBUG
+	    g_message("File \"%s\" exists, requesting new key.\n", new);
+#endif
+            g_free(msg_info->key);
+            msg_info->key = lbm_mdir_get_key();
+            g_free(filename);
+            filename =
+                g_strconcat(msg_info->key, (len ? ":2," : ""), new_flags,
+                            NULL);
+            g_free(new);
+            new = g_build_filename(path, subdir, filename, NULL);
+        }
 	if (rename(orig, new) >= 0) /* FIXME: change to safe_rename??? */
 	    msg_info->subdir = subdir;
-	else
+	else {
+#ifdef DEBUG
+            g_message("Rename \"%s\" \"%s\": %s\n", orig, new,
+                      g_strerror(errno));
+#endif
 	    retval = FALSE;
+	}
     }
     g_free(orig);
     g_free(new);
-    new = g_strdup_printf("%s%s%s", msg_info->key, (len ? ":2," : ""),
-			  new_flags);
     g_free(msg_info->filename);
-    msg_info->filename = new;
+    msg_info->filename = filename;
     msg_info->orig_flags = REAL_FLAGS(msg_info->flags);
 
     return retval;
@@ -735,19 +781,20 @@ libbalsa_mailbox_maildir_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     GSList *l;
     guint renumber, msgno;
     struct message_info *msg_info;
+    guint changes = 0;
 
     for (msgno = 1; msgno <= mdir->msgno_2_msg_info->len; msgno++) {
-	msg_info = message_info_from_msgno(mailbox, msgno);
+	msg_info = message_info_from_msgno(mdir, msgno);
 
 	if (expunge && (msg_info->flags & LIBBALSA_MESSAGE_FLAG_DELETED)) {
 	    /* skip this block if the message should only be renamed */
-	    gchar *orig =
-		g_strdup_printf("%s/%s/%s", path, msg_info->subdir, 
-                                msg_info->filename);
+	    gchar *orig = g_build_filename(path, msg_info->subdir,
+                                           msg_info->filename, NULL);
 	    unlink (orig);
 	    g_free(orig);
 	    removed_list =
 		g_slist_prepend(removed_list, GUINT_TO_POINTER(msgno));
+	    ++changes;
 	    continue;
 	}
 
@@ -760,6 +807,7 @@ libbalsa_mailbox_maildir_sync(LibBalsaMailbox * mailbox, gboolean expunge)
             || FLAGS_CHANGED(msg_info)) {
 	    if (!maildir_sync_add(msg_info, path))
 		ok = FALSE;
+	    ++changes;
 	}
     }
 
@@ -773,7 +821,7 @@ libbalsa_mailbox_maildir_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	msgno = GPOINTER_TO_UINT(l->data);
 	if (renumber > msgno)
 	    renumber = msgno;
-	msg_info = message_info_from_msgno(mailbox, msgno);
+	msg_info = message_info_from_msgno(mdir, msgno);
 
 	g_ptr_array_remove(mdir->msgno_2_msg_info, msg_info);
 	libbalsa_mailbox_local_msgno_removed(mailbox, msgno);
@@ -782,28 +830,32 @@ libbalsa_mailbox_maildir_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     }
     g_slist_free(removed_list);
     for (msgno = renumber; msgno <= mdir->msgno_2_msg_info->len; msgno++) {
-	msg_info = message_info_from_msgno(mailbox, msgno);
+	msg_info = message_info_from_msgno(mdir, msgno);
 	if (msg_info->message)
 	    msg_info->message->msgno = msgno;
     }
 
-    /* Record mtime of dirs; we'll just use the current time--someone
-     * else might have changed something since we did, but we'll find
-     * out eventually. */
-    mdir->mtime = mdir->mtime_cur = mdir->mtime_new = time(NULL);
+    if (changes) {              /* Record mtime of dir. */
+        struct stat st;
+
+        /* Reparse, to get the fileno entries right. */
+        lbm_maildir_parse_subdirs(mdir);
+
+        if (stat(mdir->tmpdir, &st) == 0)
+            libbalsa_mailbox_set_mtime(mailbox, st.st_mtime);
+    }
 
     return TRUE;
 }
 
-static struct message_info *message_info_from_msgno(
-						  LibBalsaMailbox * mailbox,
-						  guint msgno)
+static struct message_info *
+message_info_from_msgno(LibBalsaMailboxMaildir * mdir, guint msgno)
 {
-    struct message_info *msg_info = NULL;
+    g_assert(msgno > 0 && msgno <= mdir->msgno_2_msg_info->len);
 
-    msg_info = g_ptr_array_index(LIBBALSA_MAILBOX_MAILDIR(mailbox)->msgno_2_msg_info,
-			      msgno - 1);
-    return msg_info;
+    return (struct message_info *) g_ptr_array_index(mdir->
+                                                     msgno_2_msg_info,
+                                                     msgno - 1);
 }
 
 
@@ -814,7 +866,8 @@ libbalsa_mailbox_maildir_fetch_message_structure(LibBalsaMailbox * mailbox,
 {
     if (!message->mime_msg) {
 	struct message_info *msg_info =
-	    message_info_from_msgno(mailbox, message->msgno);
+            message_info_from_msgno((LibBalsaMailboxMaildir *) mailbox,
+                                    message->msgno);
 	message->mime_msg =
 	    libbalsa_mailbox_local_get_mime_message(mailbox,
 						    msg_info->subdir,
@@ -826,86 +879,102 @@ libbalsa_mailbox_maildir_fetch_message_structure(LibBalsaMailbox * mailbox,
 }
 
 static LibBalsaMessage *
-libbalsa_mailbox_maildir_get_message(LibBalsaMailbox * mailbox,
-                                     guint msgno)
+lbm_maildir_get_message(LibBalsaMailbox * mailbox, guint msgno)
 {
     struct message_info *msg_info =
-	message_info_from_msgno(mailbox, msgno);
+        message_info_from_msgno((LibBalsaMailboxMaildir *) mailbox, msgno);
+    LibBalsaMessage *message;
+
+    message = msg_info->message;
+    if (message)
+        return g_object_ref(message);
+
+    msg_info->message = message = libbalsa_message_new();
+    g_object_add_weak_pointer(G_OBJECT(message),
+                              (gpointer) & msg_info->message);
+
+    message->flags = REAL_FLAGS(msg_info->flags);
+    message->mailbox = mailbox;
+    message->msgno = msgno;
+    libbalsa_message_load_envelope(message);
+
+    return message;
+}
+
+static LibBalsaMessageFlag
+lbm_maildir_load_message(LibBalsaMailboxLocal * local, guint msgno,
+                         LibBalsaMessage ** msg)
+{
+    struct message_info *msg_info =
+        message_info_from_msgno(LIBBALSA_MAILBOX_MAILDIR(local), msgno);
 
     if (msg_info->message)
-	g_object_ref(msg_info->message);
-    else {
-	LibBalsaMessage *message;
+        *msg = g_object_ref(msg_info->message);
 
-	msg_info->message = message = libbalsa_message_new();
-	g_object_add_weak_pointer(G_OBJECT(message),
-				  (gpointer) & msg_info->message);
+    return REAL_FLAGS(msg_info->flags);
+}
 
-	message->flags = REAL_FLAGS(msg_info->flags);
-	message->mailbox = mailbox;
-	message->msgno = msgno;
-	libbalsa_message_load_envelope(message);
-    }
+static guint
+lbm_maildir_fileno(LibBalsaMailboxLocal * local, guint msgno)
+{
+    struct message_info *msg_info;
 
-    return msg_info->message;
+    if (!msgno)
+        return 0;
+
+    msg_info =
+        message_info_from_msgno((LibBalsaMailboxMaildir *) local, msgno);
+
+    return msg_info->fileno;
 }
 
 /* Called with mailbox locked. */
-static int
+static gboolean
 libbalsa_mailbox_maildir_add_message(LibBalsaMailbox * mailbox,
-                                     LibBalsaMessage * message, GError **err)
+                                     GMimeStream * stream,
+                                     LibBalsaMessageFlag flags,
+                                     GError **err)
 {
     const char *path;
     char *tmp;
     int fd;
     GMimeStream *out_stream;
-    GMimeStream *tmp_stream;
     GMimeStream *in_stream;
     GMimeFilter *crlffilter;
     char *new_filename;
     struct message_info *msg_info;
     gint retval;
+    LibBalsaMailboxMaildir *mdir;
+    time_t mtime;
 
     /* open tempfile */
     path = libbalsa_mailbox_local_get_path(mailbox);
     fd = libbalsa_mailbox_maildir_open_temp(path, &tmp);
     if (fd == -1)
-	return -1;
+	return FALSE;
     out_stream = g_mime_stream_fs_new(fd);
 
-    tmp_stream =
-        libbalsa_mailbox_get_message_stream(message->mailbox, message);
-    if (!tmp_stream) {
-        g_object_unref(out_stream);
-        unlink(tmp);
-        g_free(tmp);
-        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
-                    LIBBALSA_MAILBOX_COPY_ERROR,
-                    _("Cannot read message"));
-        return -1;
-    }
-    in_stream = g_mime_stream_filter_new_with_stream(tmp_stream);
-    g_object_unref(tmp_stream);
-
+    in_stream = g_mime_stream_filter_new_with_stream(stream);
     crlffilter =
         g_mime_filter_crlf_new(GMIME_FILTER_CRLF_DECODE,
                                GMIME_FILTER_CRLF_MODE_CRLF_ONLY);
     g_mime_stream_filter_add(GMIME_STREAM_FILTER(in_stream), crlffilter);
     g_object_unref(crlffilter);
  
-    if (g_mime_stream_write_to_stream( in_stream, out_stream) == -1)
-    {
-	g_object_unref(in_stream);
-	g_object_unref(out_stream);
+    libbalsa_mime_stream_shared_lock(stream);
+    retval = g_mime_stream_write_to_stream(in_stream, out_stream);
+    libbalsa_mime_stream_shared_unlock(stream);
+    g_object_unref(in_stream);
+    g_object_unref(out_stream);
+
+    if (retval < 0) {
 	unlink (tmp);
 	g_free(tmp);
         g_set_error(err, LIBBALSA_MAILBOX_ERROR,
                     LIBBALSA_MAILBOX_COPY_ERROR,
                     _("Data copy error"));
-	return -1;
+	return FALSE;
     }
-    g_object_unref(in_stream);
-    g_object_unref(out_stream);
 
     new_filename = strrchr(tmp, '/');
     if (new_filename)
@@ -916,11 +985,18 @@ libbalsa_mailbox_maildir_add_message(LibBalsaMailbox * mailbox,
     msg_info->subdir = "tmp";
     msg_info->key = g_strdup(new_filename);
     msg_info->filename = g_strdup(new_filename);
-    msg_info->flags = message->flags | LIBBALSA_MESSAGE_FLAG_RECENT;
-    retval = maildir_sync_add(msg_info, path) ? 1 : -1;
+    msg_info->flags = flags | LIBBALSA_MESSAGE_FLAG_RECENT;
+    retval = maildir_sync_add(msg_info, path);
     free_message_info(msg_info);
     g_free(tmp);
 
+    mdir = (LibBalsaMailboxMaildir *) mailbox;
+    if ((mtime = libbalsa_mailbox_get_mtime(mailbox)) != 0)
+	/* If we checked or synced the mailbox less than 1 second ago,
+	 * the cached modification time could be the same as the new
+	 * modification time, so we'll invalidate the cached time. */
+	libbalsa_mailbox_set_mtime(mailbox, --mtime);
+
     return retval;
 }
 
@@ -936,7 +1012,8 @@ libbalsa_mailbox_maildir_messages_change_flags(LibBalsaMailbox * mailbox,
     for (i = 0; i < msgnos->len; i++) {
         guint msgno = g_array_index(msgnos, guint, i);
         struct message_info *msg_info =
-            message_info_from_msgno(mailbox, msgno);
+            message_info_from_msgno((LibBalsaMailboxMaildir *) mailbox,
+                                    msgno);
         LibBalsaMessageFlag old_flags = msg_info->flags;
         gboolean was_unread_undeleted, is_unread_undeleted;
 
@@ -951,7 +1028,6 @@ libbalsa_mailbox_maildir_messages_change_flags(LibBalsaMailbox * mailbox,
             msg_info->message->flags = msg_info->flags;
 
         libbalsa_mailbox_index_set_flags(mailbox, msgno, msg_info->flags);
-        libbalsa_mailbox_msgno_changed(mailbox, msgno);
 
         was_unread_undeleted = (old_flags & LIBBALSA_MESSAGE_FLAG_NEW)
             && !(old_flags & LIBBALSA_MESSAGE_FLAG_DELETED);
@@ -978,7 +1054,7 @@ libbalsa_mailbox_maildir_msgno_has_flags(LibBalsaMailbox * mailbox,
                                          LibBalsaMessageFlag unset)
 {
     struct message_info *msg_info =
-        message_info_from_msgno(mailbox, msgno);
+        message_info_from_msgno((LibBalsaMailboxMaildir *) mailbox, msgno);
 
     return (msg_info->flags & set) == set
         && (msg_info->flags & unset) == 0;
