@@ -24,21 +24,36 @@
 
 #define _SVID_SOURCE           1
 #define _XOPEN_SOURCE          500
-#include <stdio.h>
-#include <sys/utsname.h>
 #include <ctype.h>
-#include <string.h>
-#include <errno.h>
 #include <dirent.h>
+#include <errno.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <sys/file.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/utsname.h>
+#if HAVE_COMPFACE
+#include <compface.h>
+#endif                          /* HAVE_COMPFACE */
 
+#ifdef HAVE_GNOME
 #include <libgnomevfs/gnome-vfs.h>
-#include <libgnome/libgnome.h>
+#endif
+
+#if HAVE_GTKSOURCEVIEW
+#include <gtksourceview/gtksourceview.h>
+#include <gtksourceview/gtksourcebuffer.h>
+#include <gtksourceview/gtksourcetag.h>
+#include <gtksourceview/gtksourcetagstyle.h>
+#endif
 
 #include "libbalsa.h"
 #include "libbalsa_private.h"
 #include "misc.h"
 #include "html.h"
+#include "i18n.h"
 
 static const gchar *libbalsa_get_codeset_name(const gchar *txt, 
 					      LibBalsaCodeset Codeset);
@@ -51,6 +66,7 @@ static int getdnsdomainname(char *s, size_t l);
 gchar*
 libbalsa_lookup_mime_type(const gchar * path)
 {
+#ifdef HAVE_GNOME
     GnomeVFSFileInfo* vi = gnome_vfs_file_info_new();
     gchar* uri, *mime_type;
 
@@ -67,7 +83,10 @@ libbalsa_lookup_mime_type(const gchar * path)
     g_free(uri);
     mime_type = g_strdup(gnome_vfs_file_info_get_mime_type(vi));
     gnome_vfs_file_info_unref(vi);
-    return mime_type;
+    return mime_type ? mime_type : g_strdup("application/octet-stream");
+#else
+    return g_strdup("application/octet-stream");
+#endif/* HAVE_GNOME */
 }
 
 gchar *
@@ -135,7 +154,7 @@ getdnsdomainname (char *s, size_t l)
 gchar *
 libbalsa_get_domainname(void)
 {
-    char domainname[SYS_NMLN];
+    char domainname[256]; /* arbitrary length */
     struct utsname utsname;
     gchar *d;
  
@@ -145,7 +164,7 @@ libbalsa_get_domainname(void)
         return g_strdup( d+1 );
     }
  
-    if ( getdnsdomainname(domainname, SYS_NMLN) == 0 ) {
+    if ( getdnsdomainname(domainname, sizeof(domainname)) == 0 ) {
 	return g_strdup(domainname);
     }
     return NULL;
@@ -207,35 +226,6 @@ libbalsa_urldecode(const gchar * str)
     return retval;
 }
 
-/* FIXME: Move to address.c and change name to
- *   libbalsa_address_list_to_string or something */
-gchar *
-libbalsa_make_string_from_list(const GList * the_list)
-{
-    return libbalsa_make_string_from_list_p(the_list);
-}
-/* private to libbalsa: */
-gchar *
-libbalsa_make_string_from_list_p(const GList * the_list)
-{
-    const GList *list;
-    GString *gs = g_string_new(NULL);
-
-    for (list = the_list; list; list = list->next) {
-        LibBalsaAddress *addy = list->data;
-        gchar *str = libbalsa_address_to_gchar_p(addy, -1);
-        if (str) {
-            if (gs->len > 0)
-                g_string_append(gs, ", ");
-            g_string_append(gs, str);
-            g_free(str);
-        }
-    }
-
-    return g_string_free(gs, FALSE);
-}
-
-
 /* readfile allocates enough space for the ending '\0' characeter as well.
    returns the number of read characters.
 */
@@ -423,6 +413,11 @@ libbalsa_wrap_string(gchar * str, int width)
 /* Updated 2003 to implement DelSp=Yes as in
  * http://www.ietf.org/internet-drafts/draft-gellens-format-bis-01.txt
  */
+/* Now documented in RFC 3676:
+ * http://www.ietf.org/rfc/rfc3676.txt
+ * or
+ * http://www.faqs.org/rfcs/rfc3676.html
+ */
 
 #define MAX_WIDTH	997	/* enshrined somewhere */
 #define QUOTE_CHAR	'>'
@@ -606,8 +601,12 @@ dowrap_rfc2646(GList * list, gint width, gboolean to_screen,
              * */
             if (str > start && isspace((int)str[-1]) && str[-1] != ' ')
                 g_string_append_c(result, ' ');
-            if (*str)           /* line separator */
-                g_string_append_c(result, '\n');
+            if (*str) {         /* line separator */
+                if (to_screen || str == start)
+		    g_string_append_c(result, '\n');
+		else		/* DelSP = Yes */
+		    g_string_append(result, " \n");
+	    }
         } while (*str);         /* end of loop over output lines */
 
         g_free(text->str);
@@ -752,22 +751,22 @@ libbalsa_wrap_view(GtkTextView * view, gint length)
 		/* Break at the last break we saw. */
 		gtk_text_iter_set_line_offset(&iter, brk_offset);
 	    else {
+                GtkTextIter start = iter;
 		/* Break at the next line break. */
 		if (offset <= quote_len)
 		    offset = quote_len + 1;
 		while (offset < num_chars
 		       && (is_in_url(&iter, offset, url_tag)
-			   || !log_attrs[offset].is_line_break)) {
-		    gboolean next_is_space =
-			g_unichar_isspace(gtk_text_iter_get_char(&iter));
-		    if (FALSE && in_space && !next_is_space)
-			break;
-		    in_space = next_is_space;
+			   || !log_attrs[offset].is_line_break))
 		    offset++;
-		}
+
 		if (offset >= num_chars)
 		    /* No next line break. */
 		    break;
+
+                /* Trim extra trailing whitespace */
+                gtk_text_iter_forward_char(&start);
+                gtk_text_buffer_delete(buffer, &start, &iter);
 	    }
 
 	    gtk_text_buffer_insert_with_tags(buffer, &iter, "\n", 1,
@@ -792,13 +791,15 @@ get_quote_tag(GtkTextIter * iter)
     GSList *list;
     GSList *tag_list = gtk_text_iter_get_tags(iter);
 
-    for (list = tag_list; list; list = g_slist_next(list)) {
+    for (list = tag_list; list; list = list->next) {
         GtkTextTag *tag = list->data;
         gchar *name;
         g_object_get(G_OBJECT(tag), "name", &name, NULL);
-        if (!strncmp(name, "quote-", 6))
-            quote_tag = tag_list->data;
-        g_free(name);
+        if (name) {
+            if (!strncmp(name, "quote-", 6))
+                quote_tag = tag_list->data;
+            g_free(name);
+        }
     }
     g_slist_free(tag_list);
 
@@ -1202,6 +1203,45 @@ libbalsa_utf8_sanitize(gchar **text, gboolean fallback,
     return FALSE;
 }
 
+/* libbalsa_utf8_strstr() returns TRUE if s2 is a substring of s1.
+ * libbalsa_utf8_strstr is case insensitive
+ * this functions understands utf8 strings (as you might have guessed ;-)
+ */
+gboolean
+libbalsa_utf8_strstr(const gchar *s1, const gchar *s2)
+{
+    const gchar * p,* q;
+
+    /* convention : NULL string is contained in anything */
+    if (!s2) return TRUE;
+    /* s2 is non-NULL, so if s1==NULL we return FALSE :)*/
+    if (!s1) return FALSE;
+    /* OK both are non-NULL now*/
+    /* If s2 is the empty string return TRUE */
+    if (!*s2) return TRUE;
+    while (*s1) {
+	/* We look for the first char of s2*/
+	for (;*s1 &&
+		 g_unichar_toupper(g_utf8_get_char(s2))!=g_unichar_toupper(g_utf8_get_char(s1));
+	     s1 = g_utf8_next_char(s1));
+	if (*s1) {
+	    /* We found the first char let see if this potential match is an actual one */
+	    s1 = g_utf8_next_char(s1);
+	    q = s1;
+	    p = g_utf8_next_char(s2);
+	    while (*q && *p && 
+		   g_unichar_toupper(g_utf8_get_char(p))
+		   ==g_unichar_toupper(g_utf8_get_char(q))) {
+		p = g_utf8_next_char(p);
+		q = g_utf8_next_char(q);
+	    }
+	    /* We have a match if p has reached the end of s2, ie *p==0 */
+	    if (!*p) return TRUE;
+	}
+    }
+    return FALSE;
+}
+
 /* The LibBalsaCodeset enum is not used for anything currently, but this
  * list must be the same length, and should probably be kept consistent: */
 LibBalsaCodesetInfo libbalsa_codeset_info[LIBBALSA_NUM_CODESETS] = {
@@ -1235,7 +1275,7 @@ LibBalsaCodesetInfo libbalsa_codeset_info[LIBBALSA_NUM_CODESETS] = {
      "iso-8859-15"} ,
     {N_("russian"),             /* RUSSIAN              */
      "koi-8r"} ,
-    {N_("ukranian"),            /* UKRAINE              */
+    {N_("ukrainian"),           /* UKRAINE              */
      "koi-8u"} ,
     {N_("japanese"),            /* JAPAN                */
      "iso-2022-jp"} ,
@@ -1273,7 +1313,6 @@ libbalsa_get_codeset_name(const gchar * txt, LibBalsaCodeset Codeset)
     return info->std;
 }
 
-#if NEW_CHARSET_WIDGET
 /* Create a GtkComboBox with the national charsets as options;
  * called when some text is found to be neither US-ASCII nor UTF-8, so
  * the list includes neither of these. */
@@ -1303,21 +1342,19 @@ libbalsa_charset_button_new(void)
 
     return combo_box;
 }
-#endif                          /* NEW_CHARSET_WIDGET */
 
 /* Helper */
 static void
 lb_text_attr(const gchar * text, gboolean * has_esc, gboolean * has_hi_bit,
              gboolean * has_hi_ctrl)
 {
-    const guchar *p;
-
-    for (p = text; *p; p++) {
-        if (*p == 0x1b)
+    for (; *text; text++) {
+	guchar c = *text;
+        if (c == 0x1b)
             *has_esc = TRUE;
-        if (*p >= 0x80) {
+        if (c >= 0x80) {
             *has_hi_bit = TRUE;
-            if (*p <= 0x9f)
+            if (c <= 0x9f)
                 *has_hi_ctrl = TRUE;
         }
     }
@@ -1502,102 +1539,148 @@ get_url_reg(void)
     return url_reg;
 }
 
-void
+static regex_t *
+get_ml_url_reg(void)
+{
+    static regex_t *url_reg = NULL;
+    
+    if (!url_reg) {
+        /* one-time compilation of a constant url_str expression */
+        static const char url_str[] =
+#ifdef HAVE_PCRE
+            "(%[0-9A-F]{2}|[-_.!~*';/?:@&=+$,#\\w]|[ \\t]*[\\r\\n]+[ \\t]*)+>";
+#else
+	    "(%[0-9A-F]{2}|[-_.!~*';/?:@&=+$,#[:alnum:]]|[ \t]*[\r\n]+[ \t]*)+>";
+#endif
+
+	url_reg = g_new(regex_t, 1);
+        if (regcomp(url_reg, url_str, REG_EXTENDED | REG_ICASE) != 0)
+            g_warning("libbalsa_insert_with_url: "
+                      "multiline url regex compilation failed.");
+    }
+    
+    return url_reg;
+}
+
+gboolean
 libbalsa_insert_with_url(GtkTextBuffer * buffer,
                          const char *chars,
+			 const char *all_chars,
                          GtkTextTag * tag,
-                         void (*callback) (GtkTextBuffer *,
-                                           GtkTextIter *,
-                                           const gchar *,
-                                           gpointer),
-                         gpointer callback_data)
+                         LibBalsaUrlInsertInfo *url_info)
 {
     gchar *p, *buf;
+    const gchar *all_p;
     GtkTextIter iter;
 
     buf = p = g_strdup(chars);
+    all_p = all_chars;
     gtk_text_buffer_get_iter_at_mark(buffer, &iter,
                                      gtk_text_buffer_get_insert(buffer));
 
-    if (prescanner(p)) {
-        GtkTextTagTable *table = gtk_text_buffer_get_tag_table(buffer);
-        GtkTextTag *url_tag = gtk_text_tag_table_lookup(table, "url");
-        gint offset = 0;
-        regex_t *url_reg = get_url_reg();
-        regmatch_t url_match;
-        gint match = regexec(url_reg, p, 1, &url_match, 0);
+    /* if there shouldn't be a callback for URL's we don't need to detect
+       them... */
+    if (url_info) {
+	GtkTextTagTable *table = gtk_text_buffer_get_tag_table(buffer);
+	GtkTextTag *url_tag = gtk_text_tag_table_lookup(table, "url");
 
-        while (!match) {
-            gchar *buf;
+	if (url_info->ml_url) {
+	    gchar *url_end = strchr(p, '>');
 
-            if (url_match.rm_so) {
-                buf = g_strndup(p, url_match.rm_so);
-                gtk_text_buffer_insert_with_tags(buffer, &iter,
-                                                 buf, -1, tag, NULL);
-                g_free(buf);
-            }
+	    if (url_end) {
+		url_info->ml_url_buffer =
+		    g_string_append_len(url_info->ml_url_buffer, p, url_end - p);
+		gtk_text_buffer_insert_with_tags(buffer, &iter,
+						 url_info->ml_url_buffer->str, -1,
+						 url_tag, tag, NULL);
+		if (url_info->callback)
+		    url_info->callback(buffer, &iter, url_info->ml_url, url_info->callback_data);
+		g_string_free(url_info->ml_url_buffer, TRUE);
+		g_free(url_info->ml_url);
+		url_info->ml_url_buffer = NULL;
+		url_info->ml_url = NULL;
+		p = url_end;
+	    } else {
+		url_info->ml_url_buffer =
+		    g_string_append(url_info->ml_url_buffer, p);
+		url_info->ml_url_buffer =
+		    g_string_append_c(url_info->ml_url_buffer, '\n');
+		return TRUE;
+	    }
+	}
 
-            buf = g_strndup(p + url_match.rm_so,
-                            url_match.rm_eo - url_match.rm_so);
-            gtk_text_buffer_insert_with_tags(buffer, &iter, buf, -1,
-                                             url_tag, tag, NULL);
+	if (prescanner(p)) {
+	    gint offset = 0;
+	    regex_t *url_reg = get_url_reg();
+	    regmatch_t url_match;
+	    gint match = regexec(url_reg, p, 1, &url_match, 0);
 
-            /* remember the URL and its position within the text */
-            if (callback)
-                callback(buffer, &iter, buf, callback_data);
-            g_free(buf);
+	    while (!match) {
+		gchar *buf;
 
-            p += url_match.rm_eo;
-            offset += url_match.rm_eo;
-            if (prescanner(p))
-                match = regexec(url_reg, p, 1, &url_match, 0);
-            else
-                match = -1;
-        }
+		if (url_match.rm_so) {
+		    /* check if we hit a multi-line URL... (see RFC 1738) */
+		    if (all_p && (p[url_match.rm_so - 1] == '<' ||
+				  (url_match.rm_so > 4 &&
+				   g_ascii_strcasecmp(p + url_match.rm_so - 5, "<URL:") == 0)) &&
+			!strchr(p + url_match.rm_eo, '>')) {
+			regex_t *ml_url_reg = get_ml_url_reg();
+			regmatch_t ml_url_match;
+		    
+			if (!regexec(ml_url_reg,
+				     all_chars + offset + url_match.rm_eo, 1,
+				     &ml_url_match, 0) && ml_url_match.rm_so == 0) {
+			    GString *ml_url = g_string_new("");
+			    const gchar *ml_p = all_chars + offset + url_match.rm_so;
+			    gint ml_cnt =
+				url_match.rm_eo - url_match.rm_so + ml_url_match.rm_eo - 1;
+
+			    for (; ml_cnt; (ml_p++, ml_cnt--))
+				if (*ml_p > ' ')
+				    ml_url = g_string_append_c(ml_url, *ml_p);
+			    url_info->ml_url = ml_url->str;
+			    g_string_free(ml_url, FALSE);
+			}
+		    }
+
+		    buf = g_strndup(p, url_match.rm_so);
+		    gtk_text_buffer_insert_with_tags(buffer, &iter,
+						     buf, -1, tag, NULL);
+		    g_free(buf);
+		}
+
+		if (url_info->ml_url) {
+		    url_info->ml_url_buffer = g_string_new(p + url_match.rm_so);
+		    url_info->ml_url_buffer =
+			g_string_append_c(url_info->ml_url_buffer, '\n');
+		    return TRUE;
+		}
+
+		buf = g_strndup(p + url_match.rm_so,
+				url_match.rm_eo - url_match.rm_so);
+		gtk_text_buffer_insert_with_tags(buffer, &iter, buf, -1,
+						 url_tag, tag, NULL);
+
+		/* remember the URL and its position within the text */
+		if (url_info->callback)
+		    url_info->callback(buffer, &iter, buf, url_info->callback_data);
+		g_free(buf);
+
+		p += url_match.rm_eo;
+		offset += url_match.rm_eo;
+		if (prescanner(p))
+		    match = regexec(url_reg, p, 1, &url_match, 0);
+		else
+		    match = -1;
+	    }
+	}
     }
 
     if (*p)
         gtk_text_buffer_insert_with_tags(buffer, &iter, p, -1, tag, NULL);
     g_free(buf);
-}
 
-/* Helper for "Select All" callbacks: if the currently focused widget
- * supports any concept of "select-all", do it.
- *
- * It would be nice if all such widgets had a "select-all" signal, but
- * they don't; in fact, the only one that does (GtkTreeView) is
- * broken--if we emit it when the tree is not in multiple selection
- * mode, bad stuff happens.
- */
-void
-libbalsa_window_select_all(GtkWindow * window)
-{
-    GtkWidget *focus_widget = gtk_window_get_focus(window);
-
-    if (!focus_widget)
-	return;
-
-    if (GTK_IS_TEXT_VIEW(focus_widget)) {
-        GtkTextBuffer *buffer =
-            gtk_text_view_get_buffer(GTK_TEXT_VIEW(focus_widget));
-        GtkTextIter start, end;
-
-        gtk_text_buffer_get_bounds(buffer, &start, &end);
-        gtk_text_buffer_place_cursor(buffer, &start);
-        gtk_text_buffer_move_mark_by_name(buffer, "selection_bound", &end);
-    } else if (GTK_IS_EDITABLE(focus_widget)) {
-        gtk_editable_select_region(GTK_EDITABLE(focus_widget), 0, -1);
-    } else if (GTK_IS_TREE_VIEW(focus_widget)) {
-        GtkTreeSelection *selection =
-            gtk_tree_view_get_selection(GTK_TREE_VIEW(focus_widget));
-        if (gtk_tree_selection_get_mode(selection) ==
-            GTK_SELECTION_MULTIPLE)
-            gtk_tree_selection_select_all(selection);
-#ifdef    HAVE_GTKHTML
-    } else if (libbalsa_html_can_select(focus_widget)) {
-	libbalsa_html_select_all(focus_widget);
-#endif /* HAVE_GTKHTML */
-    }
+    return FALSE;
 }
 
 void
@@ -1674,12 +1757,12 @@ libbalsa_match_regex(const gchar * line, regex_t * rex, guint * count,
 #define compare_stat(osb, nsb)  ( (osb.st_dev != nsb.st_dev || osb.st_ino != nsb.st_ino || osb.st_rdev != nsb.st_rdev) ? -1 : 0 )
 
 int 
-libbalsa_safe_open (const char *path, int flags)
+libbalsa_safe_open (const char *path, int flags, mode_t mode)
 {
   struct stat osb, nsb;
   int fd;
  
-  if ((fd = open (path, flags, 0600)) < 0)
+  if ((fd = open (path, flags, mode)) < 0)
     return fd;
  
   /* make sure the file is not symlink */
@@ -1778,7 +1861,7 @@ libbalsa_lock_file (const char *path, int fd, int excl, int dot, int timeout)
 #if defined (USE_FCNTL) || defined (USE_FLOCK)
     int count;
     int attempt;
-    struct stat prev_sb;
+    struct stat prev_sb = { 0 };
 #endif
     int r = 0;
 
@@ -1909,7 +1992,7 @@ libbalsa_unlock_file (const char *path, int fd, int dot)
     return 0;
 }
 
-#ifndef HAVE_GLIB22
+#if !GLIB_CHECK_VERSION(2, 2, 0)
 gboolean
 libbalsa_str_has_prefix(const gchar * str, const gchar * prefix)
 {
@@ -1923,4 +2006,378 @@ libbalsa_str_has_prefix(const gchar * str, const gchar * prefix)
 
     return *prefix == '\0';
 }
-#endif				/* HAVE_GLIB22 */
+#endif				/* !GLIB_CHECK_VERSION(2, 2, 0) */
+
+
+/* libbalsa_ia_rfc2821_equal
+   compares two addresses according to rfc2821: local-part@domain is equal,
+   if the local-parts are case sensitive equal, but the domain case-insensitive
+*/
+gboolean
+libbalsa_ia_rfc2821_equal(const InternetAddress * a,
+			  const InternetAddress * b)
+{
+    const gchar *a_atptr, *b_atptr;
+    gint a_atpos, b_atpos;
+
+    if (!a || !b || a->type != INTERNET_ADDRESS_NAME ||
+	b->type != INTERNET_ADDRESS_NAME)
+        return FALSE;
+
+    /* first find the "@" in the two addresses */
+    a_atptr = strchr(a->value.addr, '@');
+    b_atptr = strchr(b->value.addr, '@');
+    if (!a_atptr || !b_atptr)
+        return FALSE;
+    a_atpos = a_atptr - a->value.addr;
+    b_atpos = b_atptr - b->value.addr;
+
+    /* now compare the strings */
+    if (!a_atpos || !b_atpos || a_atpos != b_atpos || 
+        strncmp(a->value.addr, b->value.addr, a_atpos) ||
+        g_ascii_strcasecmp(a_atptr, b_atptr))
+        return FALSE;
+    else
+        return TRUE;
+}
+
+/*
+ * Face and X-Face header support.
+ */
+gchar *
+libbalsa_get_header_from_path(const gchar * header, const gchar * path,
+                              gsize * size, GError ** err)
+{
+    gchar *buf, *content;
+    size_t name_len;
+    gchar *p, *q;
+
+    if (!g_file_get_contents(path, &buf, size, err))
+        return NULL;
+
+    content = buf;
+    name_len = strlen(header);
+    if (g_ascii_strncasecmp(content, header, name_len) == 0)
+        /* Skip header and trailing colon: */
+        content += name_len + 1;
+
+    /* Unfold. */
+    for (p = q = content; *p; p++)
+        if (*p != '\r' && *p != '\n')
+            *q++ = *p;
+    *q = '\0';
+
+    content = g_strdup(content);
+    g_free(buf);
+
+    return content;
+}
+
+GtkWidget *
+libbalsa_get_image_from_face_header(const gchar * content, GError ** err)
+{
+    GMimeStream *stream;
+    GMimeStream *stream_filter;
+    GMimeFilter *filter;
+    GByteArray *array;
+    GtkWidget *image = NULL;
+
+    stream = g_mime_stream_mem_new();
+    stream_filter = g_mime_stream_filter_new_with_stream(stream);
+
+    filter = g_mime_filter_basic_new_type(GMIME_FILTER_BASIC_BASE64_DEC);
+    g_mime_stream_filter_add(GMIME_STREAM_FILTER(stream_filter), filter);
+    g_object_unref(filter);
+
+    g_mime_stream_write_string(stream_filter, content);
+    g_object_unref(stream_filter);
+
+    array = GMIME_STREAM_MEM(stream)->buffer;
+    if (array->len == 0)
+        g_set_error(err, LIBBALSA_IMAGE_ERROR,
+                    LIBBALSA_IMAGE_ERROR_NO_DATA, _("No image data"));
+    else {
+        GdkPixbufLoader *loader =
+            gdk_pixbuf_loader_new_with_type("png", NULL);
+
+        gdk_pixbuf_loader_write(loader, array->data, array->len, err);
+        gdk_pixbuf_loader_close(loader, *err ? NULL : err);
+
+        if (!*err)
+            image = gtk_image_new_from_pixbuf(gdk_pixbuf_loader_get_pixbuf
+                                              (loader));
+        g_object_unref(loader);
+    }
+    g_object_unref(stream);
+
+    return image;
+}
+
+#if HAVE_COMPFACE
+GtkWidget *
+libbalsa_get_image_from_x_face_header(const gchar * content, GError ** err)
+{
+    gchar buf[2048];
+    GdkPixbuf *pixbuf;
+    guchar *pixels;
+    gint lines;
+    const gchar *p;
+    GtkWidget *image = NULL;
+
+    strncpy(buf, content, sizeof buf - 1);
+
+    switch (uncompface(buf)) {
+    case -1:
+        g_set_error(err, LIBBALSA_IMAGE_ERROR, LIBBALSA_IMAGE_ERROR_FORMAT,
+                    _("Invalid input format"));
+        return image;
+    case -2:
+        g_set_error(err, LIBBALSA_IMAGE_ERROR, LIBBALSA_IMAGE_ERROR_BUFFER,
+                    _("Internal buffer overrun"));
+        return image;
+    }
+
+    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, 48, 48);
+    pixels = gdk_pixbuf_get_pixels(pixbuf);
+
+    p = buf;
+    for (lines = 48; lines > 0; --lines) {
+        guint x[3];
+        gint j, k;
+        guchar *q;
+
+        if (sscanf(p, "%x,%x,%x,", &x[0], &x[1], &x[2]) != 3) {
+            g_set_error(err, LIBBALSA_IMAGE_ERROR,
+                        LIBBALSA_IMAGE_ERROR_BAD_DATA,
+                        /* Translators: please do not translate Face. */
+                        _("Bad X-Face data"));
+            g_object_unref(pixbuf);
+            return image;
+        }
+        for (j = 0, q = pixels; j < 3; j++)
+            for (k = 15; k >= 0; --k){
+                guchar c = x[j] & (1 << k) ? 0x00 : 0xff;
+                *q++ = c;       /* red   */
+                *q++ = c;       /* green */
+                *q++ = c;       /* blue  */
+            }
+        p = strchr(p, '\n') + 1;
+        pixels += gdk_pixbuf_get_rowstride(pixbuf);
+    }
+
+    image = gtk_image_new_from_pixbuf(pixbuf);
+    g_object_unref(pixbuf);
+
+    return image;
+}
+#endif                          /* HAVE_COMPFACE */
+
+GQuark
+libbalsa_image_error_quark(void)
+{
+    static GQuark quark = 0;
+    if (quark == 0)
+        quark = g_quark_from_static_string("libbalsa-image-error-quark");
+    return quark;
+}
+
+#if HAVE_GTKSOURCEVIEW
+GtkWidget *
+libbalsa_source_view_new(gboolean highlight_phrases, GdkColor *q_colour)
+{
+    GtkTextTag * text_tag;
+    GtkSourceTagStyle *tag_style;
+    GtkSourceTagTable *tag_table;
+    GSList *tag_list;
+    GtkSourceBuffer *sbuffer;
+    GtkWidget *sview;
+
+    /* create the tag table */
+    tag_list = NULL;
+    tag_table = gtk_source_tag_table_new();
+
+    /* add highlighting for quoted text if requested */
+    if (q_colour) {
+	int k;
+
+	for (k = 1; k <= 9; k++) {
+	    gchar * tag_id;
+	    gchar * pattern;
+
+	    tag_id = g_strdup_printf("Quote-%d", k);
+	    if (k == 1)
+		pattern = g_strdup("^> ?($|[^|>:}#])");
+	    else
+		pattern = g_strdup_printf("^(> ?){%d}($|[^|>:}#])", k);
+	    printf("%d: %s\n", k, pattern);
+	    text_tag = gtk_line_comment_tag_new(tag_id, tag_id, pattern);
+	    g_free(pattern);
+	    g_free(tag_id);
+	    tag_style = gtk_source_tag_style_new();
+	    tag_style->mask = GTK_SOURCE_TAG_STYLE_USE_FOREGROUND;
+	    tag_style->foreground = q_colour[(k - 1) & 1];
+	    gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	    gtk_source_tag_style_free(tag_style);
+	    tag_list = g_slist_prepend(tag_list, text_tag);
+	}
+    }
+
+    /* if requested create the patterns for bold, italic and underline */
+    if (highlight_phrases) {
+	text_tag = gtk_pattern_tag_new("Bold", "Bold",
+				       "(^|[[:space:]])\\*[[:alnum:]][^*\n]*[[:alnum:]]\\*");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->bold = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+
+	text_tag = gtk_pattern_tag_new("Italic", "Italic",
+				       "(^|[[:space:]])/[[:alnum:]][^/\n]*[[:alnum:]]/");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->italic = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+
+	text_tag = gtk_pattern_tag_new("Underline", "Underline",
+				       "(^|[[:space:]])_[[:alnum:]][^_\n]*[[:alnum:]]_");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->underline = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+    }
+
+    /* add tags to the table if present */
+    if (tag_list) {
+	gtk_source_tag_table_add_tags(tag_table, tag_list);
+	g_slist_foreach(tag_list, (GFunc)g_object_unref, NULL);
+	g_slist_free(tag_list);
+    }
+
+    /* create the source buffer */
+    sbuffer = gtk_source_buffer_new(tag_table);
+    g_object_unref(tag_table);
+    gtk_source_buffer_set_highlight(sbuffer, highlight_phrases || q_colour);
+    gtk_source_buffer_set_check_brackets(sbuffer, FALSE);
+
+    /* create & return the source view */
+    sview = gtk_source_view_new_with_buffer(sbuffer);
+    g_object_unref(sbuffer);
+
+    return sview;
+}
+#endif  /* HAVE_GTKSOURCEVIEW */
+
+/*
+ * Utilities for making consistent dialogs.
+ */
+
+#define LB_PADDING 12           /* per HIG */
+
+GtkWidget *
+libbalsa_create_table(guint rows, guint columns)
+{
+    GtkWidget *table;
+
+    table = gtk_table_new(rows, columns, FALSE);
+
+    gtk_table_set_row_spacings(GTK_TABLE(table), LB_PADDING);
+    gtk_table_set_col_spacings(GTK_TABLE(table), LB_PADDING);
+
+    return table;
+}
+
+/* create_label:
+   Create a label and add it to a table in the first column of given row,
+   setting the keyval to found accelerator value, that can be later used 
+   in create_entry.
+*/
+GtkWidget *
+libbalsa_create_label(const gchar * text, GtkWidget * table, gint row)
+{
+    GtkWidget *label = gtk_label_new_with_mnemonic(text);
+
+    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
+
+    gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row + 1,
+                     GTK_FILL, GTK_FILL, 0, 0);
+
+    return label;
+}
+
+/* create_check:
+   creates a checkbox with a given label and places them in given array.
+*/
+GtkWidget *
+libbalsa_create_check(const gchar * text, GtkWidget * table, gint row,
+                      gboolean initval)
+{
+    GtkWidget *check_button;
+
+    check_button = gtk_check_button_new_with_mnemonic(text);
+
+    gtk_table_attach(GTK_TABLE(table), check_button, 0, 2, row, row + 1,
+                     GTK_FILL, 0, 0, 0);
+
+    if (initval)
+        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_button),
+                                     TRUE);
+
+    return check_button;
+}
+
+/* Create a text entry and add it to the table */
+GtkWidget *
+libbalsa_create_entry(GtkWidget * table, GCallback changed_func,
+                      gpointer data, gint row, const gchar * initval,
+                      GtkWidget * hotlabel)
+{
+    GtkWidget *entry;
+
+    entry = gtk_entry_new();
+
+    gtk_table_attach(GTK_TABLE(table), entry, 1, 2, row, row + 1,
+                     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
+
+    if (initval) {
+        gint zero = 0;
+
+        gtk_editable_insert_text(GTK_EDITABLE(entry), initval, -1, &zero);
+    }
+
+    gtk_label_set_mnemonic_widget(GTK_LABEL(hotlabel), entry);
+
+    /* Watch for changes... */
+    if (changed_func)
+        g_signal_connect(entry, "changed", changed_func, data);
+
+    return entry;
+}
+
+/* Create a GtkSizeGroup and add to it any GtkLabel packed in a GtkTable
+ * inside the chooser widget; size_group will be unreffed when the
+ * chooser widget is finalized. */
+static void
+lb_create_size_group_func(GtkWidget * widget, gpointer data)
+{
+    if (GTK_IS_LABEL(widget) && GTK_IS_TABLE(widget->parent))
+        gtk_size_group_add_widget(GTK_SIZE_GROUP(data), widget);
+    else if (GTK_IS_CONTAINER(widget))
+        gtk_container_foreach(GTK_CONTAINER(widget),
+                              lb_create_size_group_func, data);
+}
+
+GtkSizeGroup *
+libbalsa_create_size_group(GtkWidget * chooser)
+{
+    GtkSizeGroup *size_group;
+
+    size_group = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
+    g_object_weak_ref(G_OBJECT(chooser), (GWeakNotify) g_object_unref,
+                      size_group);
+    lb_create_size_group_func(chooser, size_group);
+
+    return size_group;
+}
