@@ -25,10 +25,7 @@
 
 #include <stdio.h>
 #include <gtk/gtk.h>
-
-#if GTK_CHECK_VERSION(2, 4, 0)
-# define NEW_CHARSET_WIDGET TRUE
-#endif
+#include <gmime/gmime.h>
 
 #if ENABLE_ESMTP
 #include <auth-client.h>
@@ -85,16 +82,24 @@ struct _LibBalsaCodesetInfo {
     const gchar *std;
     const gchar *win;
 };
+
+typedef void (*libbalsa_url_cb_t) (GtkTextBuffer *, GtkTextIter *,
+				   const gchar *, gpointer);
+typedef struct _LibBalsaUrlInsertInfo LibBalsaUrlInsertInfo;
+struct _LibBalsaUrlInsertInfo {
+    libbalsa_url_cb_t callback;
+    gpointer callback_data;
+    gchar *ml_url;
+    GString *ml_url_buffer;
+};
+
 extern LibBalsaCodesetInfo libbalsa_codeset_info[];
-#if NEW_CHARSET_WIDGET
 GtkWidget *libbalsa_charset_button_new(void);
-#endif                          /* NEW_CHARSET_WIDGET */
 LibBalsaTextAttribute libbalsa_text_attr_string(const gchar * string);
 LibBalsaTextAttribute libbalsa_text_attr_file(const gchar * filename);
 const gchar *libbalsa_file_get_charset(const gchar * filename);
 
 gchar *libbalsa_lookup_mime_type(const gchar * path);
-gchar *libbalsa_make_string_from_list(const GList *);
 
 size_t libbalsa_readfile(FILE * fp, char **buf);
 size_t libbalsa_readfile_nostat(FILE * fp, char **buf);
@@ -129,27 +134,67 @@ gboolean libbalsa_mktempdir(gchar ** name);
 LibBalsaCodeset libbalsa_set_fallback_codeset(LibBalsaCodeset codeset);
 gboolean libbalsa_utf8_sanitize(gchar ** text, gboolean fallback,
                                 gchar const **target);
-void libbalsa_insert_with_url(GtkTextBuffer * buffer,
-                              const char *chars,
-                              GtkTextTag * tag,
-                              void (*callback) (GtkTextBuffer *,
-                                                GtkTextIter *,
-                                                const gchar *,
-                                                gpointer),
-                              gpointer callback_data);
-void libbalsa_window_select_all(GtkWindow * window);
+gboolean libbalsa_utf8_strstr(const gchar *s1,const gchar *s2);
+gboolean libbalsa_insert_with_url(GtkTextBuffer * buffer,
+				  const char *chars,
+				  const char *all_chars,
+				  GtkTextTag * tag,
+				  LibBalsaUrlInsertInfo *url_info);
 void libbalsa_unwrap_selection(GtkTextBuffer * buffer, regex_t * rex);
 gboolean libbalsa_match_regex(const gchar * line, regex_t * rex,
 			      guint * count, guint * index);
 
-int libbalsa_safe_open (const char *path, int flags);
+int libbalsa_safe_open (const char *path, int flags, mode_t mode);
 int libbalsa_lock_file (const char *path, int fd, int excl, int dot, int timeout);
 int libbalsa_unlock_file (const char *path, int fd, int dot);
 int libbalsa_safe_rename (const char *src, const char *target);
-#ifdef HAVE_GLIB22
+#if GLIB_CHECK_VERSION(2, 2, 0)
 #define libbalsa_str_has_prefix(str, prefix) g_str_has_prefix((str), (prefix))
-#else				/* HAVE_GLIB22 */
+#else				/* GLIB_CHECK_VERSION(2, 2, 0) */
 gboolean libbalsa_str_has_prefix(const gchar * str, const gchar * prefix);
-#endif				/* HAVE_GLIB22 */
+#endif				/* GLIB_CHECK_VERSION(2, 2, 0) */
+
+gboolean libbalsa_ia_rfc2821_equal(const InternetAddress * a,
+				   const InternetAddress * b);
+
+/*
+ * Face and X-Face header support.
+ */
+gchar *libbalsa_get_header_from_path(const gchar * header,
+                                     const gchar * path, gsize * size,
+                                     GError ** err);
+GtkWidget *libbalsa_get_image_from_face_header(const gchar * content,
+                                               GError ** err);
+#if HAVE_COMPFACE
+GtkWidget *libbalsa_get_image_from_x_face_header(const gchar * content,
+                                                 GError ** err);
+#endif                          /* HAVE_COMPFACE */
+
+GQuark libbalsa_image_error_quark(void);
+#define LIBBALSA_IMAGE_ERROR libbalsa_image_error_quark()
+enum LibBalsaImageError {
+    LIBBALSA_IMAGE_ERROR_NO_DATA
+#if HAVE_COMPFACE
+        ,
+    LIBBALSA_IMAGE_ERROR_FORMAT,
+    LIBBALSA_IMAGE_ERROR_BUFFER,
+    LIBBALSA_IMAGE_ERROR_BAD_DATA
+#endif                          /* HAVE_COMPFACE */
+};
+
+#if HAVE_GTKSOURCEVIEW
+GtkWidget *libbalsa_source_view_new(gboolean highlight_phrases, GdkColor *q_colour);
+#endif                          /* HAVE_GTKSOURCEVIEW */
+
+GtkWidget *libbalsa_create_table(guint rows, guint columns);
+GtkWidget *libbalsa_create_label(const gchar * label, GtkWidget * table,
+                                 gint row);
+GtkWidget *libbalsa_create_entry(GtkWidget * table, GCallback func,
+                                 gpointer data, gint row,
+                                 const gchar * initval,
+                                 GtkWidget * hotlabel);
+GtkWidget *libbalsa_create_check(const gchar * label, GtkWidget * table,
+                                 gint row, gboolean initval);
+GtkSizeGroup *libbalsa_create_size_group(GtkWidget * chooser);
 
 #endif				/* __LIBBALSA_MISC_H__ */
