@@ -32,23 +32,35 @@
 #endif
 
 #include "address-book.h"
+#include "address-book-vcard.h"
 #include "address-book-extern.h"
 #include "address-book-ldif.h"
 #if ENABLE_LDAP
 #include "address-book-ldap.h"
-#endif
-#include "address-book-vcard.h"
+#endif /* ENABLE_LDAP */
+#if HAVE_SQLITE
 #include "address-book-gpe.h"
+#endif /* HAVE_SQLITE */
+#include "address-book-config.h"
+#include "libbalsa-conf.h"
+#include "libbalsa.h"
+#include "i18n.h"
 
 struct ABMainWindow {
-    GtkWidget *entry_list; /* GtkListView widget */
-    GtkWidget *edit_box, *apply_button, *remove_button, *cancel_button;
+    GtkWindow *window;
+    GtkWidget *entry_list; /* GtkTreeView widget */
+    GtkWidget *apply_button, *remove_button, *cancel_button;
     GtkWidget *edit_widget;
     GtkWidget *entries[NUM_FIELDS];
+    GtkRadioAction *first_radio_action;
 
     GList *address_book_list;
+    gchar *default_address_book_prefix;
+    LibBalsaAddressBook *default_address_book;
     LibBalsaAddressBook* address_book;
     LibBalsaAddress *displayed_address;
+    GtkActionGroup *action_group;
+    GtkUIManager *ui_manager;
 } contacts_app;
     
 
@@ -60,35 +72,28 @@ static gint bab_save_session(GnomeClient * client, gint phase,
                              gpointer client_data);
 static gint bab_kill_session(GnomeClient * client, gpointer client_data);
 
-static void ab_set_edit_widget(GtkWidget *w, gboolean can_remove);
+static void ab_set_edit_widget(LibBalsaAddress * address,
+                               gboolean can_remove);
+static void ab_clear_edit_widget(void);
 
-#define BALSA_CONFIG_PREFIX "balsa/"
 #define ADDRESS_BOOK_SECTION_PREFIX "address-book-"
-static void
-bab_config_init(void)
+static gboolean
+bab_config_init(const gchar * group, const gchar * value, gpointer data)
 {
     LibBalsaAddressBook *address_book;
-    void *iterator;
-    gchar *key, *val, *tmp;
-    int pref_len = strlen(ADDRESS_BOOK_SECTION_PREFIX);
 
-    iterator = gnome_config_init_iterator_sections(BALSA_CONFIG_PREFIX);
-    while ((iterator = gnome_config_iterator_next(iterator, &key, &val))) {
+    address_book = libbalsa_address_book_new_from_config(group);
+    if (address_book) {
+        contacts_app.address_book_list =
+            g_list_append(contacts_app.address_book_list, address_book);
 
-	if (strncmp(key, ADDRESS_BOOK_SECTION_PREFIX, pref_len) == 0) {
-	    tmp = g_strconcat(BALSA_CONFIG_PREFIX, key, "/", NULL);
-
-	    address_book = libbalsa_address_book_new_from_config(tmp);
-	    if (address_book)
-		contacts_app.address_book_list =
-		    g_list_append(contacts_app.address_book_list,
-				  address_book);
-	    g_free(tmp);
-	}
-	g_free(key);
-	g_free(val);
+        if (contacts_app.default_address_book_prefix
+            && strcmp(group,
+                      contacts_app.default_address_book_prefix) == 0)
+            contacts_app.default_address_book = address_book;
     }
 
+    return FALSE;
 }
 
 static void ab_warning(const char *fmt, ...);
@@ -169,76 +174,317 @@ bab_set_address_book(LibBalsaAddressBook *ab, GtkWidget* list,
     return TRUE;
 }
 
-#if GTK_CHECK_VERSION(2, 4, 0)
+static void
+bab_window_set_title(LibBalsaAddressBook * address_book)
+{
+    const gchar *type = "";
+    gchar *title;
+
+    if (LIBBALSA_IS_ADDRESS_BOOK_VCARD(address_book))
+        type = "vCard";
+    else if (LIBBALSA_IS_ADDRESS_BOOK_EXTERN(address_book))
+        type = "External query";
+    else if (LIBBALSA_IS_ADDRESS_BOOK_LDIF(address_book))
+        type = "LDIF";
+#if ENABLE_LDAP
+    else if (LIBBALSA_IS_ADDRESS_BOOK_LDAP(address_book))
+        type = "LDAP";
+#endif
+#if HAVE_SQLITE
+    else if (LIBBALSA_IS_ADDRESS_BOOK_GPE(address_book))
+        type = "GPE";
+#endif
+
+    title =
+        g_strconcat(type, _(" address book: "), address_book->name, NULL);
+    gtk_window_set_title(contacts_app.window, title);
+    g_free(title);
+}
+
 static void
 select_address_book_cb(GtkRadioAction * action, GtkRadioAction * current,
-                       gpointer callback_data)
+                       gpointer data)
 {
-    GList *l;
+    LibBalsaAddressBook *address_book;
 
     if (action != current)
         return;
-    l = g_list_nth(contacts_app.address_book_list,
-                   GPOINTER_TO_INT(callback_data));
-    if (!l)
+
+    address_book =
+        g_list_nth_data(contacts_app.address_book_list,
+                        gtk_radio_action_get_current_value(action));
+    if (!address_book)
         return;
-    ab_set_edit_widget(NULL, FALSE);
-    bab_set_address_book(LIBBALSA_ADDRESS_BOOK(l->data),
-                         contacts_app.entry_list, NULL);
-}
-#else /* GTK_CHECK_VERSION(2, 4, 0) */
-static void
-select_address_book_cb(gpointer callback_data, guint callback_action,
-                       GtkWidget *w)
-{
-    GList *l;
 
-    if(!GTK_CHECK_MENU_ITEM(w)->active) return;
-    l = g_list_nth(contacts_app.address_book_list,
-                          GPOINTER_TO_INT(callback_data));
-    if(!l) return;
-    ab_set_edit_widget(NULL, FALSE);
-    bab_set_address_book(LIBBALSA_ADDRESS_BOOK(l->data),
-                         contacts_app.entry_list, NULL);
+    ab_clear_edit_widget();
+    bab_set_address_book(address_book, contacts_app.entry_list, NULL);
+    bab_window_set_title(address_book);
 }
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 
 static void
-address_changed_cb(GtkWidget *w, gpointer data)
+address_changed_cb(struct ABMainWindow *aw)
 {
-    struct ABMainWindow * aw = (struct ABMainWindow*)data;
-    gtk_widget_set_sensitive(aw->apply_button,  TRUE);
+    gtk_widget_set_sensitive(aw->apply_button, TRUE);
     gtk_widget_set_sensitive(aw->cancel_button, TRUE);
 }
 
+/* File menu callback helpers */
+
+#define BAB_MERGE_ID "balsa-ab-merge-id-key"
 
 static void
-#if GTK_CHECK_VERSION(2, 4, 0)
-edit_new_person_cb(GtkAction * action, gpointer user_data)
-#else /* GTK_CHECK_VERSION(2, 4, 0) */
-edit_new_person_cb(gpointer callback_data, guint callback_action, GtkWidget *w)
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+add_address_book(LibBalsaAddressBook * address_book)
 {
-    GtkWidget *ew;
+    static guint pos;
+    gchar *label;
+    GtkRadioAction *radio_action;
+    gchar *accelerator;
+    guint merge_id;
+
+    label = g_strdup_printf("_%d:%s", ++pos, address_book->name);
+    radio_action = gtk_radio_action_new(address_book->name, label,
+                                        NULL, NULL, pos - 1);
+    g_free(label);
+
+    if (contacts_app.first_radio_action) {
+        GSList *group =
+            gtk_radio_action_get_group(contacts_app.first_radio_action);
+        gtk_radio_action_set_group(radio_action, group);
+        if (address_book == contacts_app.default_address_book)
+            contacts_app.first_radio_action = radio_action;
+    } else
+        contacts_app.first_radio_action = radio_action;
+
+    g_signal_connect(G_OBJECT(radio_action), "changed",
+                     G_CALLBACK(select_address_book_cb), NULL);
+
+    accelerator = pos <= 9 ? g_strdup_printf("<control>%d", pos) : NULL;
+    gtk_action_group_add_action_with_accel(contacts_app.action_group,
+                                           GTK_ACTION(radio_action),
+                                           accelerator);
+    g_free(accelerator);
+
+    merge_id = gtk_ui_manager_new_merge_id(contacts_app.ui_manager);
+    gtk_ui_manager_add_ui(contacts_app.ui_manager, merge_id,
+                          "/ui/MainMenu/FileMenu/",
+                          address_book->name, address_book->name,
+                          GTK_UI_MANAGER_AUTO, FALSE);
+    g_object_set_data(G_OBJECT(address_book), BAB_MERGE_ID,
+                      GUINT_TO_POINTER(merge_id));
+}
+
+static gboolean
+get_used_group(const gchar * key, const gchar * value, gpointer data)
+{
+    gint *max = data;
+    gint curr;
+
+    if (*value && (curr = atoi(value)) > *max)
+        *max = curr;
+
+    return FALSE;
+}
+
+static gchar *
+get_unused_group(const gchar * prefix)
+{
+    gint max = 0;
+
+    libbalsa_conf_foreach_group(prefix, get_used_group, &max);
+
+    return g_strdup_printf("%s%d", prefix, ++max);
+}
+
+#define BAB_RADIO_ACTION "bab-radio-action-key"
+#define BAB_PREFIX_LEN 3        /* strlen("_1:") */
+
+static void
+address_book_change(LibBalsaAddressBook * address_book, gboolean append)
+{
+    gchar *group;
+
+    if (append) {
+        contacts_app.address_book_list =
+            g_list_append(contacts_app.address_book_list, address_book);
+        add_address_book(address_book);
+    } else {
+        GtkRadioAction *radio_action =
+            g_object_get_data(G_OBJECT(address_book), BAB_RADIO_ACTION);
+        gchar *label;
+
+        g_object_get(G_OBJECT(radio_action), "label", &label, NULL);
+
+        if (strlen(label) <= BAB_PREFIX_LEN
+            || strcmp(label + BAB_PREFIX_LEN, address_book->name) != 0) {
+            gchar *new_label;
+
+            label[BAB_PREFIX_LEN] = 0;
+            new_label = g_strconcat(label, address_book->name, NULL);
+            g_object_set(G_OBJECT(radio_action), "label", new_label, NULL);
+            g_free(new_label);
+            bab_window_set_title(address_book);
+        }
+        g_free(label);
+    }
+
+    group = address_book->config_prefix ?
+        g_strdup(address_book->config_prefix) :
+        get_unused_group(ADDRESS_BOOK_SECTION_PREFIX);
+    libbalsa_address_book_save_config(address_book, group);
+    g_free(group);
+
+    libbalsa_conf_sync();
+}
+
+static void
+file_new_vcard_cb(GtkAction * action, gpointer user_data)
+{
+    balsa_address_book_config_new_from_type
+        (LIBBALSA_TYPE_ADDRESS_BOOK_VCARD, address_book_change,
+         contacts_app.window);
+}
+
+static void
+file_new_extern_cb(GtkAction * action, gpointer user_data)
+{
+    balsa_address_book_config_new_from_type
+        (LIBBALSA_TYPE_ADDRESS_BOOK_EXTERN, address_book_change,
+         contacts_app.window);
+}
+
+static void
+file_new_ldif_cb(GtkAction * action, gpointer user_data)
+{
+    balsa_address_book_config_new_from_type
+        (LIBBALSA_TYPE_ADDRESS_BOOK_LDIF, address_book_change,
+         contacts_app.window);
+}
+
+#if ENABLE_LDAP
+static void
+file_new_ldap_cb(GtkAction * action, gpointer user_data)
+{
+    balsa_address_book_config_new_from_type
+        (LIBBALSA_TYPE_ADDRESS_BOOK_LDAP, address_book_change,
+         contacts_app.window);
+}
+#endif /* ENABLE_LDAP */
+
+#if HAVE_SQLITE
+static void
+file_new_gpe_cb(GtkAction * action, gpointer user_data)
+{
+    balsa_address_book_config_new_from_type
+        (LIBBALSA_TYPE_ADDRESS_BOOK_GPE, address_book_change,
+         contacts_app.window);
+}
+#endif /* HAVE_SQLITE */
+
+static void
+file_properties_cb(GtkAction * action, gpointer user_data)
+{
+    LibBalsaAddressBook *address_book;
+    GtkAction *radio_action;
+
+    if (!(address_book = contacts_app.address_book))
+        return;
+
+    radio_action = gtk_action_group_get_action(contacts_app.action_group,
+                                               address_book->name);
+    g_object_set_data(G_OBJECT(address_book), BAB_RADIO_ACTION,
+                      radio_action);
+    balsa_address_book_config_new(address_book, address_book_change,
+                                  contacts_app.window);
+}
+
+static void
+file_delete_cb(GtkAction * action, gpointer user_data)
+{
+    LibBalsaAddressBook *address_book;
+    guint merge_id;
+    GtkAction *radio_action;
+    GList *list;
+
+    if (!(address_book = contacts_app.address_book)
+        || !g_list_next(contacts_app.address_book_list))
+        return;
+
+    if (address_book->config_prefix) {
+        libbalsa_conf_remove_group(address_book->config_prefix);
+        libbalsa_conf_private_remove_group(address_book->config_prefix);
+        libbalsa_conf_sync();
+    }
+
+    merge_id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(address_book),
+                                                  BAB_MERGE_ID));
+    gtk_ui_manager_remove_ui(contacts_app.ui_manager, merge_id);
+
+    radio_action = gtk_action_group_get_action(contacts_app.action_group,
+                                               address_book->name);
+    gtk_action_group_remove_action(contacts_app.action_group, radio_action);
+    g_object_unref(radio_action);
+    if (contacts_app.first_radio_action == (GtkRadioAction*) radio_action)
+        contacts_app.first_radio_action = NULL;
+
+    /* Leave a NULL item in the address book list, to avoid changing the
+     * positions of the other books. */
+    list = g_list_find(contacts_app.address_book_list, address_book);
+    list->data = NULL;
+    g_object_unref(address_book);
+
+    for (list = contacts_app.address_book_list; list; list = list->next)
+        if ((address_book = list->data) != NULL)
+            break;
+
+    if (!list)
+        return;
+
+    radio_action = gtk_action_group_get_action(contacts_app.action_group,
+                                               address_book->name);
+    if (!contacts_app.first_radio_action)
+        contacts_app.first_radio_action = (GtkRadioAction*) radio_action;
+    gtk_action_activate(radio_action);
+}
+
+static void
+edit_new_person_cb(GtkAction * action, gpointer user_data)
+{
+    GtkTreeSelection *selection;
+
     contacts_app.displayed_address = NULL;
-    ew = libbalsa_address_get_edit_widget(NULL, contacts_app.entries,
-                                          G_CALLBACK(address_changed_cb),
-                                          &contacts_app);
-    ab_set_edit_widget(ew, FALSE);
+    ab_set_edit_widget(NULL, FALSE);
     gtk_widget_set_sensitive(contacts_app.remove_button, FALSE);
+    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW
+                                            (contacts_app.entry_list));
+    gtk_tree_selection_unselect_all(selection);
+    gtk_widget_grab_focus(GTK_CONTAINER(contacts_app.edit_widget)->
+                          focus_child);
 }
 
-#if GTK_CHECK_VERSION(2, 4, 0)
 /* Normal items */
 static GtkActionEntry entries[] = {
     {"FileMenu", NULL, "_File"},
     {"EntryMenu", NULL, "_Entry"},
     {"HelpMenu", NULL, "_Help"},
-    {"New", GTK_STOCK_NEW, "_New", NULL, "New file", NULL},
-    {"Open", GTK_STOCK_OPEN, "_Open", NULL, "Open a file", NULL},
-    {"Save", GTK_STOCK_SAVE, "_Save", NULL, "Save file", NULL},
-    {"SaveAs", GTK_STOCK_SAVE_AS, "Save _As", "<shift><control>S",
-     "Save file as", NULL},
+    {"New", GTK_STOCK_NEW, "_New"},
+    {"NewVcard", NULL, N_("VCard Address Book (GnomeCard)"), NULL, NULL,
+     G_CALLBACK(file_new_vcard_cb)},
+    {"NewExtern", NULL, N_("External query (a program)"), NULL, NULL,
+     G_CALLBACK(file_new_extern_cb)},
+    {"NewLdif", NULL, N_("LDIF Address Book"), NULL, NULL,
+     G_CALLBACK(file_new_ldif_cb)},
+#if ENABLE_LDAP
+    {"NewLdap", NULL, N_("LDAP Address Book"), NULL, NULL,
+     G_CALLBACK(file_new_ldap_cb)},
+#endif /* ENABLE_LDAP */
+#if HAVE_SQLITE
+    {"NewGpe", NULL, N_("GPE Address Book"), NULL, NULL,
+     G_CALLBACK(file_new_gpe_cb)},
+#endif /* HAVE_SQLITE */
+    {"Properties", GTK_STOCK_PROPERTIES, "_Properties", NULL,
+     "Edit address book properties", G_CALLBACK(file_properties_cb)},
+    {"Delete", GTK_STOCK_DELETE, "_Delete", NULL,
+     "Delete address book", G_CALLBACK(file_delete_cb)},
     {"Quit", GTK_STOCK_QUIT, "_Quit", NULL, "Exit the program",
      gtk_main_quit},
     {"NewPerson", GTK_STOCK_NEW, "_New Person", "<shift><control>N",
@@ -258,10 +504,19 @@ static const char *ui_description =
 "<ui>"
 "  <menubar name='MainMenu'>"
 "    <menu action='FileMenu'>"
-"      <menuitem action='New'/>"
-"      <menuitem action='Open'/>"
-"      <menuitem action='Save'/>"
-"      <menuitem action='SaveAs'/>"
+"      <menu action='New'>"
+"        <menuitem action='NewVcard'/>"
+"        <menuitem action='NewExtern'/>"
+"        <menuitem action='NewLdif'/>"
+#if ENABLE_LDAP
+"        <menuitem action='NewLdap'/>"
+#endif /* ENABLE_LDAP */
+#if HAVE_SQLITE
+"        <menuitem action='NewGpe'/>"
+#endif /* HAVE_SQLITE */
+"      </menu>"
+"      <menuitem action='Properties'/>"
+"      <menuitem action='Delete'/>"
 "      <separator/>"
 "      <menuitem action='Quit'/>"
 "      <separator/>"
@@ -285,14 +540,13 @@ get_main_menu(GtkWidget * window, GtkWidget ** menubar,
     GtkAccelGroup *accel_group;
     GError *error;
     GList *ab;
-    int cnt;
-    GSList *group = NULL;
 
-    action_group = gtk_action_group_new("MenuActions");
+    contacts_app.action_group = action_group =
+        gtk_action_group_new("MenuActions");
     gtk_action_group_add_actions(action_group, entries,
                                  G_N_ELEMENTS(entries), window);
 
-    ui_manager = gtk_ui_manager_new();
+    contacts_app.ui_manager = ui_manager = gtk_ui_manager_new();
     gtk_ui_manager_insert_action_group(ui_manager, action_group, 0);
 
     accel_group = gtk_ui_manager_get_accel_group(ui_manager);
@@ -306,125 +560,30 @@ get_main_menu(GtkWidget * window, GtkWidget ** menubar,
         return;
     }
 
-    for (cnt = 1, ab = address_books; ab; ab = ab->next, cnt++) {
-        LibBalsaAddressBook *address_book;
-	gchar *label;
-        guint merge_id;
-        gchar *accelerator;
-        GtkRadioAction *radio_action;
-
-        address_book = LIBBALSA_ADDRESS_BOOK(ab->data);
-
-        label = g_strdup_printf("_%d:%s", cnt, address_book->name);
-        radio_action = gtk_radio_action_new(address_book->name, label,
-                                            NULL, NULL, cnt - 1);
-	g_free(label);
-        if (group)
-            gtk_radio_action_set_group(radio_action, group);
-        else {
-            group = gtk_radio_action_get_group(radio_action);
-	    gtk_action_activate(GTK_ACTION(radio_action));
-	}
-        g_signal_connect(G_OBJECT(radio_action), "changed",
-                         G_CALLBACK(select_address_book_cb),
-                         GINT_TO_POINTER(cnt - 1));
-
-        accelerator =
-            cnt <= 9 ? g_strdup_printf("<control>%d", cnt) : NULL;
-        gtk_action_group_add_action_with_accel(action_group,
-                                               GTK_ACTION(radio_action),
-                                               accelerator);
-        g_free(accelerator);
-
-        merge_id = gtk_ui_manager_new_merge_id(ui_manager);
-        gtk_ui_manager_add_ui(ui_manager, merge_id,
-                              "/ui/MainMenu/FileMenu/",
-                              address_book->name, address_book->name,
-                              GTK_UI_MANAGER_AUTO, FALSE);
-    }
+    for (ab = address_books; ab; ab = ab->next)
+        add_address_book(LIBBALSA_ADDRESS_BOOK(ab->data));
 
     if (menubar)
         /* Finally, return the actual menu bar created by the UIManager. */
         *menubar = gtk_ui_manager_get_widget(ui_manager, "/MainMenu");
 }
-#else /* GTK_CHECK_VERSION(2, 4, 0) */
-static GtkItemFactoryEntry menu_items[] = {
-  { "/_File",         NULL,         NULL, 0, "<Branch>" },
-  { "/File/_New",     "<control>N", NULL, 0, NULL },
-  { "/File/_Open",    "<control>O", (GtkItemFactoryCallback)NULL,
-0, NULL },
-  { "/File/_Save",    "<control>S", NULL, 0, NULL },
-  { "/File/Save _As", NULL,         NULL, 0, NULL },
-  { "/File/sep1",     NULL,         NULL, 0, "<Separator>" },
-  { "/File/_Quit",     "<control>Q", (GtkItemFactoryCallback)gtk_main_quit, 
-    0, NULL },
-  { "/File/sep2",     NULL,         NULL, 0, "<Separator>" },
-  { "/_Entry",         NULL,        NULL, 0, "<Branch>" },
-  { "/Entry/_New Person", NULL,   edit_new_person_cb, 0, NULL },
-  { "/Entry/_New Group",  NULL,   (GtkItemFactoryCallback)NULL, 0, NULL },
-  { "/_Help",          NULL,         NULL, 0, "<LastBranch>" },
-  { "/_Help/_About",   NULL,         NULL, 0, NULL },
-};
 
 static void
-get_main_menu(GtkWidget  *window, GtkWidget **menubar, GList* address_books)
+ab_set_edit_widget(LibBalsaAddress * address, gboolean can_remove)
 {
-    GtkItemFactory *item_factory;
-    GtkAccelGroup *accel_group;
-    gint nmenu_items = sizeof(menu_items)/sizeof(menu_items[0]);
-    GList *ab;
-    int cnt;
-    gchar *first_ab_path = NULL;
-
-    accel_group = gtk_accel_group_new();
-    item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>",
-                                       accel_group);
-    gtk_item_factory_create_items (item_factory, nmenu_items,
-                                   menu_items, window);
-    
-    for(cnt=1, ab= address_books; ab; ab = ab->next, cnt++) {
-        LibBalsaAddressBook *address_book = LIBBALSA_ADDRESS_BOOK(ab->data);
-        GtkItemFactoryEntry gife;
-        gife.path = g_strdup_printf("/File/_%d:%s", cnt,
-                                    address_book->name);
-        gife.accelerator = cnt<=9 
-            ? g_strdup_printf("<control>%d", cnt) : NULL;
-        gife.callback = select_address_book_cb;
-        gife.callback_action = 0;
-        gife.item_type = first_ab_path ? first_ab_path : "<RadioItem>";
-        gife.extra_data = NULL;
-        gtk_item_factory_create_item(item_factory, &gife, 
-                                     GINT_TO_POINTER(cnt-1), 1);
-        g_free(gife.accelerator);
-        g_free(gife.path);
-        if(!first_ab_path)
-            first_ab_path = g_strdup_printf("/File/1:%s",
-                                    address_book->name);
-    }
-    g_free(first_ab_path);
-    /* Attach the new accelerator group to the window. */
-    gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
-    
-    if (menubar)
-        /* Finally, return the actual menu bar created by the item factory. */
-        *menubar = gtk_item_factory_get_widget (item_factory, "<main>");
-}
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
-
-
-static void
-ab_set_edit_widget(GtkWidget *w, gboolean can_remove)
-{
-    if(contacts_app.edit_widget)
-        gtk_widget_destroy(contacts_app.edit_widget);
-    contacts_app.edit_widget = w;
-    if(w) {
-        gtk_box_pack_start(GTK_BOX(contacts_app.edit_box), w,
-                           FALSE, FALSE, 1);
-        gtk_widget_show_all(w);
-    }
-    gtk_widget_set_sensitive(contacts_app.apply_button,  FALSE);
+    libbalsa_address_set_edit_entries(address, contacts_app.entries);
+    gtk_widget_show_all(contacts_app.edit_widget);
+    gtk_widget_set_sensitive(contacts_app.apply_button, FALSE);
     gtk_widget_set_sensitive(contacts_app.remove_button, can_remove);
+    gtk_widget_set_sensitive(contacts_app.cancel_button, !address);
+}
+
+static void
+ab_clear_edit_widget(void)
+{
+    gtk_widget_hide(contacts_app.edit_widget);
+    gtk_widget_set_sensitive(contacts_app.apply_button,  FALSE);
+    gtk_widget_set_sensitive(contacts_app.remove_button, FALSE);
     gtk_widget_set_sensitive(contacts_app.cancel_button, FALSE);
 }
 
@@ -434,19 +593,17 @@ list_selection_changed_cb(GtkTreeSelection *selection, gpointer data)
     GtkTreeIter iter;
     GtkTreeModel *model;
     GValue gv = {0,};
-    GtkWidget *ew;
     LibBalsaAddress *address;
 
     if(!gtk_tree_selection_get_selected(selection, &model, &iter))
         return;
     gtk_tree_model_get_value(model, &iter, LIST_COLUMN_ADDRESS, &gv);
     address = LIBBALSA_ADDRESS(g_value_get_object(&gv));
-    if(address) {
-        ew = libbalsa_address_get_edit_widget(address, contacts_app.entries,
-                                              G_CALLBACK(address_changed_cb),
-                                              data);
-        ab_set_edit_widget(ew, TRUE);
-    } else ab_set_edit_widget(NULL, FALSE);
+    if (address) {
+        if (address != contacts_app.displayed_address)
+            ab_set_edit_widget(address, TRUE);
+    } else
+        ab_clear_edit_widget();
     g_value_unset(&gv);
     contacts_app.displayed_address = address;
 }
@@ -496,64 +653,96 @@ bab_window_list_new(gpointer cb_data)
 }
 
 static void
-apply_button_cb(GtkWidget *w, gpointer data)
+apply_button_cb(GtkWidget * w, gpointer data)
 {
-    LibBalsaAddress * newval =
-        libbalsa_address_new_from_edit_entries(contacts_app.entries);
-    LibBalsaABErr err = 
-        contacts_app.displayed_address 
-        ? libbalsa_address_book_modify_address(contacts_app.address_book,
-                                               contacts_app.displayed_address,
-                                               newval)
-        : libbalsa_address_book_add_address(contacts_app.address_book,
-                                            newval);
+    LibBalsaAddress *newval;
+    LibBalsaABErr err;
+
+    newval = libbalsa_address_new_from_edit_entries(contacts_app.entries);
+    if (!newval) {
+        /* Some error--no full_name? */
+        ab_warning(contacts_app.displayed_address ?
+                   "Cannot modify: %s\n" : "Cannot add: %s\n",
+                   "no displayed name");
+        gtk_widget_set_sensitive(contacts_app.apply_button,  FALSE);
+        return;
+    }
+
+    err = contacts_app.displayed_address ?
+        libbalsa_address_book_modify_address(contacts_app.address_book,
+                                             contacts_app.displayed_address,
+                                             newval) :
+        libbalsa_address_book_add_address(contacts_app.address_book,
+                                          newval);
+
     if(err == LBABERR_OK) {
+        GtkTreeSelection *selection;
+        GtkTreeModel *model;
+        GtkTreeIter iter;
+        GtkTreePath *path = NULL;
+
         gtk_widget_set_sensitive(contacts_app.apply_button,  FALSE);
         gtk_widget_set_sensitive(contacts_app.remove_button, TRUE);
         gtk_widget_set_sensitive(contacts_app.cancel_button, FALSE);
+
+        /* We need to reload the book; if we modified an existing
+         * address, we really need only to reload that one address, but
+         * we have no method for that. */
+        selection =
+            gtk_tree_view_get_selection(GTK_TREE_VIEW
+                                        (contacts_app.entry_list));
+        if (gtk_tree_selection_get_selected(selection, &model, &iter))
+            path = gtk_tree_model_get_path(model, &iter);
+        bab_set_address_book(contacts_app.address_book,
+                             contacts_app.entry_list, NULL);
+        if (path) {
+            gtk_tree_selection_select_path(selection, path);
+            gtk_tree_path_free(path);
+        }
+        if (!gtk_tree_selection_get_selected(selection, NULL, NULL))
+            ab_clear_edit_widget();
     } else 
-        ab_warning("Cannot add: %s\n",
+        ab_warning(contacts_app.displayed_address ?
+                   "Cannot modify: %s\n" : "Cannot add: %s\n",
                    libbalsa_address_book_strerror(contacts_app.address_book,
-                                                  err));                   
+                                                  err));
     g_object_unref(newval);
 }
 
 static void
 remove_button_cb(GtkWidget *w, gpointer data)
 {
-    LibBalsaABErr err = 
+    LibBalsaABErr err = LBABERR_OK;
+
+    if (contacts_app.displayed_address)
         libbalsa_address_book_remove_address(contacts_app.address_book,
                                              contacts_app.displayed_address);
     if(err == LBABERR_OK) {
         GtkTreeIter       iter;
         GtkTreeSelection *selection;
         GtkTreeView  *v = GTK_TREE_VIEW(contacts_app.entry_list);
-        GtkTreeModel *m = gtk_tree_view_get_model(v);
+        GtkTreeModel *m;
         selection       = gtk_tree_view_get_selection(v);
         if(gtk_tree_selection_get_selected(selection, &m, &iter))
             gtk_list_store_remove(GTK_LIST_STORE(m), &iter);
-        ab_set_edit_widget(NULL, FALSE);
+        ab_clear_edit_widget();
         contacts_app.displayed_address = NULL;
     } else 
         ab_warning("Cannot remove: %s\n",
                    libbalsa_address_book_strerror(contacts_app.address_book,
                                                   err));
 }
+
 static void
-cancel_button_cb(GtkWidget *w, gpointer data)
+cancel_button_cb(GtkWidget * w, gpointer data)
 {
-    struct ABMainWindow *abmw = (struct ABMainWindow*)data;
-    if(abmw->displayed_address) {
-        GtkWidget *ew =
-            libbalsa_address_get_edit_widget(abmw->displayed_address,
-                                             abmw->entries,
-                                             G_CALLBACK(address_changed_cb),
-                                             data);
-        ab_set_edit_widget(ew, TRUE);
-    } else ab_set_edit_widget(NULL, FALSE);
+    struct ABMainWindow *abmw = (struct ABMainWindow *) data;
+    if (abmw->displayed_address)
+        ab_set_edit_widget(abmw->displayed_address, TRUE);
+    else
+        ab_clear_edit_widget();
 }
 
-#define ELEMENTS(x) (sizeof(x)/sizeof((x)[0])) 
 static GtkWidget*
 bab_get_edit_button_box(struct ABMainWindow *abmw)
 {
@@ -628,12 +817,13 @@ bab_get_filter_box(void)
 static GtkWidget*
 bab_window_new()
 {
-    GtkWidget* menubar, *main_vbox, *cont_box, *vbox, *scroll;
+    GtkWidget* menubar = NULL, *main_vbox, *cont_box, *vbox, *scroll;
     GtkWidget *wnd = gnome_app_new("Contacts", "Contacts");
-    GList *first_ab;
+    GtkWidget *edit_box;
 
     get_main_menu(GTK_WIDGET(wnd), &menubar, contacts_app.address_book_list);
-    gnome_app_set_menus(GNOME_APP(wnd), GTK_MENU_BAR(menubar));
+    if (menubar)
+	gnome_app_set_menus(GNOME_APP(wnd), GTK_MENU_BAR(menubar));
     /* main vbox */
     main_vbox = gtk_vbox_new(FALSE, 1);
 
@@ -648,16 +838,21 @@ bab_window_new()
     gtk_widget_show(scroll);
     gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
 				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
-    gtk_box_pack_start(GTK_BOX(cont_box), scroll, TRUE,TRUE, 1);
+    gtk_box_pack_start(GTK_BOX(cont_box), scroll, TRUE, TRUE, 1);
     
     contacts_app.entry_list = bab_window_list_new(&contacts_app);
     gtk_container_add(GTK_CONTAINER(scroll), contacts_app.entry_list);
 
     vbox = gtk_vbox_new(FALSE, 1);
-    gtk_box_pack_start(GTK_BOX(cont_box), vbox, TRUE,TRUE, 1);
-    contacts_app.edit_box = gtk_vbox_new(FALSE, 1);
-    gtk_box_pack_start(GTK_BOX(vbox), contacts_app.edit_box,
-                       TRUE,TRUE, 1);
+    gtk_box_pack_start(GTK_BOX(cont_box), vbox, FALSE, FALSE, 1);
+    edit_box = gtk_vbox_new(FALSE, 1);
+    contacts_app.edit_widget = 
+        libbalsa_address_get_edit_widget(NULL, contacts_app.entries,
+                                         G_CALLBACK(address_changed_cb),
+                                         &contacts_app);
+    gtk_box_pack_start(GTK_BOX(edit_box), contacts_app.edit_widget,
+                       FALSE, FALSE, 1);
+    gtk_box_pack_start(GTK_BOX(vbox), edit_box, TRUE, TRUE, 1);
     gtk_box_pack_start(GTK_BOX(vbox),
                        bab_get_edit_button_box(&contacts_app),
                        FALSE, FALSE, 1);
@@ -668,10 +863,6 @@ bab_window_new()
     gtk_window_set_default_size(GTK_WINDOW(wnd), 500, 400);
     gnome_app_set_contents(GNOME_APP(wnd), main_vbox);
 
-    first_ab = g_list_first(contacts_app.address_book_list);
-    if(first_ab)
-        bab_set_address_book(LIBBALSA_ADDRESS_BOOK(first_ab->data),
-                             contacts_app.entry_list, NULL);
     return wnd;
 }
 
@@ -713,6 +904,12 @@ bab_init(void)
     memset(&contacts_app, 0, sizeof(contacts_app));
 }
 
+static void
+information_real(void)
+{
+    /* FIXME */
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -726,19 +923,21 @@ main(int argc, char *argv[])
     bindtextdomain(PACKAGE, GNOMELOCALEDIR);
     bind_textdomain_codeset(PACKAGE, "UTF-8");
     textdomain(PACKAGE);
-    /* FIXME: gnome_i18n_get_language seems to have gone away; 
-     * is this a reasonable replacement? */
-    setlocale(LC_CTYPE,
-              (const char *) gnome_i18n_get_language_list("LC_CTYPE")->data);
+    setlocale(LC_ALL, "");
 #endif
 
     /* FIXME: do we need to allow a non-GUI mode? */
     gtk_init_check(&argc, &argv);
     gnome_program_init(PACKAGE, VERSION, LIBGNOMEUI_MODULE, argc, argv,
+#ifndef GNOME_PARAM_GOPTION_CONTEXT
                        GNOME_PARAM_POPT_TABLE, NULL,
                        GNOME_PARAM_APP_PREFIX,  BALSA_STD_PREFIX,
                        GNOME_PARAM_APP_DATADIR, BALSA_STD_PREFIX "/share",
                        NULL);
+#else
+                       GNOME_PARAM_GOPTION_CONTEXT, NULL,
+                       GNOME_PARAM_NONE);
+#endif
 
 #ifdef GTKHTML_HAVE_GCONF
     if (!gconf_init(argc, argv, &gconf_error))
@@ -747,11 +946,18 @@ main(int argc, char *argv[])
 #endif
 
     bab_init();
+    libbalsa_init((LibBalsaInformationFunc) information_real);
 
     /* load address book data */
-    bab_config_init();
+    libbalsa_conf_push_group("Globals");
+    contacts_app.default_address_book_prefix =
+        libbalsa_conf_get_string("DefaultAddressBook");
+    libbalsa_conf_pop_group();
+    libbalsa_conf_foreach_group(ADDRESS_BOOK_SECTION_PREFIX,
+                                bab_config_init, NULL);
 
     ab_window = bab_window_new();
+    contacts_app.window = GTK_WINDOW(ab_window);
     g_signal_connect(G_OBJECT(ab_window), "destroy",
                      G_CALLBACK(bab_cleanup), NULL);
     g_signal_connect(G_OBJECT(ab_window), "delete-event",
@@ -765,6 +971,10 @@ main(int argc, char *argv[])
 		     G_CALLBACK(bab_kill_session), NULL);
 
     gtk_widget_show_all(ab_window);
+    gtk_widget_hide(contacts_app.edit_widget);
+
+    if (contacts_app.first_radio_action)
+        gtk_action_activate(GTK_ACTION(contacts_app.first_radio_action));
 
     gdk_threads_enter();
     gtk_main();
