@@ -1,6 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- * Copyright (C) 1997-2002 Stuart Parmenter and others,
+ * Copyright (C) 1997-2005 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -34,9 +34,12 @@
 
 #include "filter-funcs.h"
 #include "misc.h"
+#include "smtp-server.h"
 #include "balsa-app.h"
 #include "save-restore.h"
 
+#include "i18n.h"	/* Must come after balsa-app.h. */
+
 /* Global application structure */
 struct BalsaApplication balsa_app;
 
@@ -51,13 +54,6 @@ const gchar *pspell_suggest_modes[] = {
     "bad-spellers"
 };
 
-static void
-rememb_toggle_cb(GObject* button, LibBalsaServer *s)
-{
-    s->remember_passwd = 
-        gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
-}
-
 /* ask_password:
    asks the user for the password to the mailbox on given remote server.
 */
@@ -96,16 +92,21 @@ ask_password_real(LibBalsaServer * server, LibBalsaMailbox * mbox)
     gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), rememb);
     if(server->remember_passwd)
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rememb), TRUE);
-    g_signal_connect(G_OBJECT(rememb), "toggled", 
-                     G_CALLBACK(rememb_toggle_cb), server);
 
     gtk_widget_show_all(GTK_WIDGET(GTK_DIALOG(dialog)->vbox));
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
     gtk_widget_grab_focus (entry);
 
-    if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)
+    if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
+        unsigned old_rem = server->remember_passwd;
         passwd = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
+        server->remember_passwd = 
+            !!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(rememb));
+        libbalsa_server_set_password(server, passwd);
+        if(old_rem != server->remember_passwd)
+            libbalsa_server_config_changed(server);
+    }
     gtk_widget_destroy(dialog);
     return passwd;
 }
@@ -244,51 +245,11 @@ ask_password(LibBalsaServer *server, LibBalsaMailbox *mbox)
 static void
 authapi_exit (void)
 {
-    if (balsa_app.smtp_authctx != NULL)
-        auth_destroy_context (balsa_app.smtp_authctx);
+    g_slist_foreach(balsa_app.smtp_servers, (GFunc) g_object_unref, NULL);
+    g_slist_free(balsa_app.smtp_servers);
+    balsa_app.smtp_servers = NULL;
     auth_client_exit ();
 }
-
-
-/* Callback to get user/password info from SMTP server preferences.
-   This is adequate for simple username / password requests but does
-   not adequately cope with all SASL mechanisms.  */
-static int
-authinteract (auth_client_request_t request, char **result, int fields,
-              void *arg)
-{
-    int i;
-
-    for (i = 0; i < fields; i++) {
-	if (request[i].flags & AUTH_PASS)
-	    result[i] = balsa_app.smtp_passphrase;
-	else if (request[i].flags & AUTH_USER)
-	    result[i] = (balsa_app.smtp_user && *balsa_app.smtp_user)
-                ? balsa_app.smtp_user : NULL;
-
-    	/* Fail the AUTH exchange if something was requested
-    	   but not supplied. */
-    	if (result[i] == NULL)
-    	    return 0;
-    }
-    return 1;
-}
-
-#if HAVE_SMTP_TLS_CLIENT_CERTIFICATE
-static int
-tlsinteract (char *buf, int buflen, int rwflag, void *arg)
-{
-  char *pw;
-  int len;
-
-  pw = balsa_app.smtp_certificate_passphrase;
-  len = strlen (pw);
-  if (len + 1 > buflen)
-    return 0;
-  strcpy (buf, pw);
-  return len;
-}
-#endif
 #endif /* ESMTP */
 
 void
@@ -303,24 +264,12 @@ balsa_app_init(void)
     balsa_app.local_mail_directory = NULL;
 
 #if ENABLE_ESMTP
-    balsa_app.smtp_server = NULL;
-    balsa_app.smtp_user = NULL;
-    balsa_app.smtp_passphrase = NULL;
+    balsa_app.smtp_servers = NULL;
 
     /* Do what's needed at application level to allow libESMTP
        to use authentication.  */
     auth_client_init ();
     atexit (authapi_exit);
-    balsa_app.smtp_authctx = auth_create_context ();
-    auth_set_mechanism_flags (balsa_app.smtp_authctx, AUTH_PLUGIN_PLAIN, 0);
-    auth_set_interact_cb (balsa_app.smtp_authctx, authinteract, NULL);
-
-#if HAVE_SMTP_TLS_CLIENT_CERTIFICATE
-    /* Use our callback for X.509 certificate passwords.  If STARTTLS is
-       not in use or disabled in configure, the following is harmless. */
-    balsa_app.smtp_certificate_passphrase = NULL;
-    smtp_starttls_set_password_cb (tlsinteract, NULL);
-#endif
 #endif
 
     balsa_app.inbox = NULL;
@@ -362,6 +311,7 @@ balsa_app_init(void)
     balsa_app.show_notebook_tabs = FALSE;
     balsa_app.alternative_layout = FALSE;
     balsa_app.view_message_on_open = TRUE;
+    balsa_app.mw_action_after_move = NEXT_UNREAD;
 
     balsa_app.index_num_width = NUM_DEFAULT_WIDTH;
     balsa_app.index_status_width = STATUS_DEFAULT_WIDTH;
@@ -397,9 +347,6 @@ balsa_app_init(void)
     balsa_app.message_font = NULL;
     balsa_app.subject_font = NULL;
 
-    /*encoding */
-    balsa_app.encoding_style = 0;
-
     /* compose: shown headers */
     balsa_app.compose_headers = NULL;
 
@@ -427,6 +374,9 @@ balsa_app_init(void)
     balsa_app.filters=NULL;
 
     /* spell check */
+#if HAVE_GTKSPELL
+    balsa_app.spell_check_lang = NULL;
+#else                           /* HAVE_GTKSPELL */
     balsa_app.module = SPELL_CHECK_MODULE_ASPELL;
     balsa_app.suggestion_mode = SPELL_CHECK_SUGGEST_NORMAL;
     balsa_app.ignore_size = 0;
@@ -434,6 +384,7 @@ balsa_app_init(void)
 
     spell_check_modules_name = pspell_modules;
     spell_check_suggest_mode_name = pspell_suggest_modes;
+#endif                          /* HAVE_GTKSPELL */
 
     /* Information messages */
     balsa_app.information_message = 0;
@@ -474,13 +425,18 @@ balsa_app_destroy(void)
     config_save();
 
     g_list_foreach(balsa_app.address_book_list, (GFunc)g_object_unref, NULL);
-    g_slist_foreach(balsa_app.filters,          (GFunc)libbalsa_filter_free, 
+    g_list_free(balsa_app.address_book_list);
+    balsa_app.address_book_list = NULL;
+
+    /* close all mailboxes */
+    gtk_widget_destroy(balsa_app.notebook);
+
+    /* now free filters */
+    g_slist_foreach(balsa_app.filters, (GFunc)libbalsa_filter_free, 
 		    GINT_TO_POINTER(TRUE));
     g_slist_free(balsa_app.filters);
     balsa_app.filters = NULL;
 
-    /* close all mailboxes */
-    gtk_widget_destroy(balsa_app.notebook);
     /* g_slist_free(opt_attach_list); */
     g_object_unref(balsa_app.colormap);
     if(balsa_app.debug) g_print("balsa_app: Finished cleaning up.\n");
@@ -572,8 +528,14 @@ open_mailboxes_idle_cb(gchar ** urls)
 	    /* Do not try to open it next time. */
 	    LibBalsaMailboxView *view =
 		g_hash_table_lookup(libbalsa_mailbox_view_table, *tmp);
-	    view->open = FALSE;
-	    view->in_sync = FALSE;
+            /* The mailbox may have bee requested to be open because
+               its stored view might say so or the user requested it
+               from the command line - in which case, view may or may
+               not be present. We will be careful here. */
+            if(view) {
+                view->open = FALSE;
+                view->in_sync = FALSE;
+            }
             balsa_information(LIBBALSA_INFORMATION_WARNING,
                               _("Couldn't open mailbox \"%s\""), *tmp);
 	}
@@ -592,12 +554,6 @@ GtkWidget *
 balsa_stock_button_with_label(const char *icon, const char *text)
 {
     GtkWidget *button;
-#if BALSA_MAJOR < 2
-    GtkWidget *pixmap;
-
-    pixmap = gnome_stock_new_with_icon(icon);
-    button = gnome_pixmap_button(pixmap, label);
-#else
     GtkWidget *pixmap = gtk_image_new_from_stock(icon, GTK_ICON_SIZE_BUTTON);
     GtkWidget *align = gtk_alignment_new(0.5, 0.5, 0, 0);
     GtkWidget *hbox = gtk_hbox_new(FALSE, 0);
@@ -614,7 +570,6 @@ balsa_stock_button_with_label(const char *icon, const char *text)
     }
 
     gtk_widget_show_all(button);
-#endif                          /* BALSA_MAJOR < 2 */
     return button;
 }
 
@@ -655,18 +610,16 @@ BalsaMailboxNode *
 balsa_find_mailbox(LibBalsaMailbox * mailbox)
 {
     BalsaFind bf;
-    gboolean is_sub_thread = libbalsa_am_i_subthread();
 
-    if (is_sub_thread)
-	gdk_threads_enter();
+    gdk_threads_enter();
 
     bf.data = mailbox;
     bf.mbnode = NULL;
-    gtk_tree_model_foreach(GTK_TREE_MODEL(balsa_app.mblist_tree_store),
-			   find_mailbox, &bf);
+    if (balsa_app.mblist_tree_store)
+        gtk_tree_model_foreach(GTK_TREE_MODEL(balsa_app.mblist_tree_store),
+                               find_mailbox, &bf);
 
-    if (is_sub_thread)
-	gdk_threads_leave();
+    gdk_threads_leave();
 
     return bf.mbnode;
 }
@@ -789,9 +742,11 @@ balsa_find_iter_by_data_func(GtkTreeModel * model, GtkTreePath * path,
 			       GtkTreeIter * iter, gpointer user_data)
 {
     struct balsa_find_iter_by_data_info *bf = user_data;
-    BalsaMailboxNode *mbnode;
+    BalsaMailboxNode *mbnode = NULL;
 
     gtk_tree_model_get(model, iter, 0, &mbnode, -1);
+    if(!mbnode)
+        return FALSE;
     if (mbnode == bf->data || mbnode->mailbox == bf->data) {
 	*bf->iter = *iter;
 	bf->found = TRUE;
@@ -880,76 +835,6 @@ balsa_remove_children_mailbox_nodes(BalsaMailboxNode * mbnode)
     g_slist_free(specials);
 }
 
-/* create_label:
-   Create a label and add it to a table in the first column of given row,
-   setting the keyval to found accelerator value, that can be later used 
-   in create_entry.
-*/
-GtkWidget *
-create_label(const gchar * label, GtkWidget * table, gint row)
-{
-    GtkWidget *w = gtk_label_new_with_mnemonic(label);
-
-    gtk_misc_set_alignment(GTK_MISC(w), 1.0, 0.5);
-    gtk_table_attach(GTK_TABLE(table), w, 0, 1, row, row + 1,
-		     GTK_FILL, GTK_FILL, 5, 5);
-    gtk_widget_show(w);
-    return w;
-}
-
-/* create_check:
-   creates a checkbox with a given label and places them in given array.
-*/
-GtkWidget *
-create_check(GtkDialog *mcw, const gchar *label, GtkWidget *table, gint row,
-             gboolean initval)
-{
-    GtkWidget *cb, *l;
-    
-    cb = gtk_check_button_new();
-
-    l = gtk_label_new_with_mnemonic(label);
-    gtk_label_set_mnemonic_widget(GTK_LABEL(l), cb);
-    gtk_misc_set_alignment(GTK_MISC(l), 0.0, 0.5);
-    gtk_widget_show(l);
-
-    gtk_container_add(GTK_CONTAINER(cb), l);
-
-    gtk_table_attach(GTK_TABLE(table), cb, 1, 2, row, row+1,
-		     GTK_FILL, GTK_FILL, 5, 5);
-
-    if(initval) 	
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb), TRUE);
-
-    return cb;
-}
-
-/* Create a text entry and add it to the table */
-GtkWidget *
-create_entry(GtkDialog *mcw, GtkWidget * table, 
-	     GtkSignalFunc changed_func, gpointer data, gint row, 
-	     const gchar * initval, GtkWidget* hotlabel)
-{
-    GtkWidget *entry = gtk_entry_new();
-    gtk_table_attach(GTK_TABLE(table), entry, 1, 2, row, row + 1,
-		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 5);
-    if (initval) {
-        gint zero = 0;
-        
-        gtk_editable_insert_text(GTK_EDITABLE(entry), initval, -1, &zero);
-    }
-
-    gtk_label_set_mnemonic_widget(GTK_LABEL(hotlabel), entry);
-
-    /* Watch for changes... */
-    if(changed_func)
-	g_signal_connect(G_OBJECT(entry), "changed", 
-			 G_CALLBACK(changed_func), data);
-
-    gtk_widget_show(entry);
-    return entry;
-}
-
 /* balsa_find_index_by_mailbox:
    returns BalsaIndex displaying passed mailbox, or NULL, if mailbox is 
    not displayed.
@@ -967,7 +852,7 @@ balsa_find_index_by_mailbox(LibBalsaMailbox * mailbox)
 	  gtk_notebook_get_nth_page(GTK_NOTEBOOK(balsa_app.notebook), i));
 	 i++) {
         index = gtk_bin_get_child(GTK_BIN(page));
-	if (BALSA_INDEX(index)->mailbox_node != NULL
+	if (index && BALSA_INDEX(index)->mailbox_node
             && BALSA_INDEX(index)->mailbox_node->mailbox == mailbox)
 	    return BALSA_INDEX(index);
     }
