@@ -41,18 +41,27 @@ typedef struct _BalsaMessageClass BalsaMessageClass;
 
 typedef struct _BalsaPartInfo BalsaPartInfo;
 
+typedef struct _BalsaMimeWidget BalsaMimeWidget;
+
+typedef enum {
+    BALSA_MESSAGE_FOCUS_STATE_NO,
+    BALSA_MESSAGE_FOCUS_STATE_YES,
+    BALSA_MESSAGE_FOCUS_STATE_HOLD
+} BalsaMessageFocusState;
+
 struct _BalsaMessage {
 	GtkNotebook parent;
 
-	/* Widget to hold headers */
-	GtkWidget *header_container;
+        /* Top-level MIME widget */
+        BalsaMimeWidget *bm_widget;
+
+	/* header-related information */
 	ShownHeaders shown_headers;
 	gboolean show_all_headers;
 
 	/* Widgets to hold content */
         GtkWidget *cont_viewport;
-	GtkWidget *content;
-	gboolean content_has_focus;
+	GtkWidget *scroll;
 
         /* Widget to hold structure tree */
         GtkWidget *treeview;
@@ -62,12 +71,14 @@ struct _BalsaMessage {
     
 	gboolean wrap_text;
 
-	BalsaPartInfo *current_part;
+        BalsaPartInfo *current_part;
         GtkWidget *parts_popup;
 
 	LibBalsaMessage *message;
 
 	gboolean close_with_msg;
+
+        BalsaMessageFocusState focus_state;
 };
 
 struct _BalsaMessageClass {
@@ -85,7 +96,9 @@ gboolean balsa_message_set(BalsaMessage * bmessage,
 			   LibBalsaMessage * message);
 
 void balsa_message_next_part(BalsaMessage * bmessage);
+gboolean balsa_message_has_next_part(BalsaMessage * bmessage);
 void balsa_message_previous_part(BalsaMessage * bmessage);
+gboolean balsa_message_has_previous_part(BalsaMessage * bmessage);
 void balsa_message_save_current_part(BalsaMessage * bmessage);
 
 void balsa_message_set_displayed_headers(BalsaMessage * bmessage,
@@ -94,14 +107,22 @@ void balsa_message_set_wrap(BalsaMessage * bmessage, gboolean wrap);
 
 gboolean balsa_message_can_select(BalsaMessage * bmessage);
 gboolean balsa_message_grab_focus(BalsaMessage * bmessage);
+gchar * balsa_message_sender_to_gchar(InternetAddressList * list, gint which);
+GtkWidget *balsa_message_current_part_widget(BalsaMessage * bmessage);
 
-void reflow_string(gchar * str, gint mode, gint * cur_pos, int width);
 #ifdef HAVE_GTKHTML
 #define BALSA_MESSAGE_ZOOM_KEY "balsa-message-zoom"
 gboolean balsa_message_can_zoom(BalsaMessage * bm);
 void balsa_message_zoom(BalsaMessage * bm, gint in_out);
 #endif				/* HAVE_GTKHTML */
 
+#ifdef HAVE_GPGME
+void balsa_message_perform_crypto(LibBalsaMessage * message,
+				  LibBalsaChkCryptoMode chk_mode,
+				  gboolean no_mp_signed,
+				  guint max_ref);
+#endif
+
 #ifdef __cplusplus
 }
 #endif				/* __cplusplus */
