@@ -39,6 +39,8 @@
 #include <locale.h>
 #endif
 
+#include "i18n.h"
+
 #include "address-entry.h"
 
 #include "balsa-app.h"
@@ -78,9 +80,6 @@ GIOChannel *mail_thread_msg_receive;
 GIOChannel *send_thread_msg_send;
 GIOChannel *send_thread_msg_receive;
 
-/* Semaphore to prevent dual use of appbar progressbar */
-int updating_progressbar;
-
 static void threads_init(void);
 static void threads_destroy(void);
 #endif				/* BALSA_USE_THREADS */
@@ -148,12 +147,6 @@ balsa_handle_automation_options() {
            printf("Unread: %ld Unsent: %ld\n", (long)unread, (long)unsent);
        }
 
-       if (cmd_get_stats) {
-           CORBA_long unread = 0, unsent = 0;
-	   GNOME_Balsa_Application_getStats (app, &unread, &unsent, &ev);
-           printf("Unread: %ld Unsent: %ld\n", (long)unread, (long)unsent);
-       }
-
        if (cmd_line_open_mailboxes)
 	   GNOME_Balsa_Application_openMailbox (app,
 					       cmd_line_open_mailboxes,
@@ -185,12 +178,13 @@ balsa_handle_automation_options() {
 
 	   GNOME_Balsa_Composer_sendMessage(server,
 					    "",
-					    opt_compose_email,
+					    opt_compose_email ? opt_compose_email : "",
 					    "",
 					    "",
 					    attachs,
 					    0, 
 					    &ev );
+           CORBA_exception_free( &ev );
        }
 
        exit(0);
@@ -207,9 +201,10 @@ balsa_handle_automation_options() {
 static void
 balsa_init(int argc, char **argv)
 {
-    poptContext context;
-    int opt;
     static char *attachment = NULL;
+#ifndef GNOME_PARAM_GOPTION_CONTEXT
+    int opt;
+    poptContext context;
     static struct poptOption options[] = {
 
 	{"checkmail", 'c', POPT_ARG_NONE,
@@ -257,7 +252,61 @@ balsa_init(int argc, char **argv)
                        GNOME_PARAM_APP_DATADIR, BALSA_STD_PREFIX "/share",
 		       GNOME_PARAM_HUMAN_READABLE_NAME, _("The Balsa E-Mail Client"),
                        NULL);
+#else /* USE GOption interface */
+    static gchar **remaining_args = NULL;
+    static GOptionEntry option_entries[] = {
+	{"checkmail", 'c', 0, G_OPTION_ARG_NONE,
+	 &(cmd_check_mail_on_startup),
+	 N_("Get new mail on startup"), NULL},
+	{"compose", 'm', 0, G_OPTION_ARG_STRING, &(opt_compose_email),
+	 N_("Compose a new email to EMAIL@ADDRESS"), "EMAIL@ADDRESS"},
+	{"attach", 'a', 0, G_OPTION_ARG_STRING, &(attachment),
+	 N_("Attach file at PATH"), "PATH"},
+	{"open-mailbox", 'o', 0, G_OPTION_ARG_STRING,
+         &(cmd_line_open_mailboxes),
+	 N_("Opens MAILBOXNAME"), N_("MAILBOXNAME")},
+	{"open-unread-mailbox", 'u', 0, G_OPTION_ARG_NONE,
+	 &(cmd_open_unread_mailbox),
+	 N_("Opens first unread mailbox"), NULL},
+	{"open-inbox", 'i', 0, G_OPTION_ARG_NONE,
+	 &(cmd_open_inbox),
+	 N_("Opens default Inbox on startup"), NULL},
+	{"get-stats", 's', 0, G_OPTION_ARG_NONE,
+	 &(cmd_get_stats),
+	 N_("Prints number unread and unsent messages"), NULL},
+	{"debug-pop", 'd', 0, G_OPTION_ARG_NONE, &PopDebug,
+	 N_("Debug POP3 connection"), NULL},
+	{"debug-imap", 'D', 0, G_OPTION_ARG_NONE, &ImapDebug,
+	 N_("Debug IMAP connection"), NULL},
+        /* last but not least a special option that collects filenames */
+        { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY,
+          &remaining_args,
+          "Special option that collects any remaining arguments for us" },
+        { NULL }
+    };
+    GOptionContext *option_context = g_option_context_new("balsa");
+    GnomeProgram *my_app;
+    g_option_context_add_main_entries(option_context, option_entries, NULL);
 
+    my_app = gnome_program_init(PACKAGE, VERSION,
+                                LIBGNOMEUI_MODULE, argc, argv,
+                                GNOME_PARAM_GOPTION_CONTEXT, option_context,
+                                GNOME_PARAM_APP_DATADIR,
+                                BALSA_STD_PREFIX "/share",
+                                GNOME_PARAM_NONE);
+
+    if (remaining_args != NULL) {
+        gint i, num_args;
+        
+        num_args = g_strv_length (remaining_args);
+        for (i = 0; i < num_args; ++i) {
+            /* process remaining_args[i] here */
+            /* we do nothing for now */
+        }
+        g_strfreev (remaining_args);
+        remaining_args = NULL;
+    }
+#endif /* OPTION HANDLING */
     balsa_handle_automation_options();  
     
 }
@@ -323,24 +372,7 @@ mailboxes_init(gboolean check_only)
     }
 }
 
-
 #ifdef BALSA_USE_THREADS
-#if GTK_CHECK_VERSION(2, 4, 0)
-/* Recursive mutex for gdk_threads_{enter,leave}. */
-static pthread_mutex_t balsa_threads_mutex;
-
-static void
-balsa_threads_enter(void)
-{
-    pthread_mutex_lock(&balsa_threads_mutex);
-}
-
-static void
-balsa_threads_leave(void)
-{
-    pthread_mutex_unlock(&balsa_threads_mutex);
-}
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 
 static void
 threads_init(void)
@@ -353,17 +385,13 @@ threads_init(void)
         g_warning("pthread_mutexattr_init failed with %d\n", status);
     pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
     pthread_mutex_init(&mailbox_lock, &attr);
-#if GTK_CHECK_VERSION(2, 4, 0)
-    pthread_mutex_init(&balsa_threads_mutex, &attr);
-    gdk_threads_set_lock_functions(G_CALLBACK(balsa_threads_enter),
-                                   G_CALLBACK(balsa_threads_leave));
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+
+    libbalsa_threads_init();
     gdk_threads_init();
 
     pthread_mutexattr_destroy(&attr);
     pthread_mutex_init(&send_messages_lock, NULL);
     checking_mail = 0;
-    updating_progressbar = 0;
     if (pipe(mail_thread_pipes) < 0) {
 	g_log("BALSA Init", G_LOG_LEVEL_DEBUG,
 	      "Error opening pipes.\n");
@@ -390,9 +418,7 @@ threads_destroy(void)
 {
     pthread_mutex_destroy(&mailbox_lock);
     pthread_mutex_destroy(&send_messages_lock);
-#if GTK_CHECK_VERSION(2, 4, 0)
-    pthread_mutex_destroy(&balsa_threads_mutex);
-#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+    libbalsa_threads_destroy();
 }
 
 #endif				/* BALSA_USE_THREADS */
@@ -406,13 +432,12 @@ initial_open_unread_mailboxes()
 {
     GList *i, *gl;
     gdk_threads_enter();
-    gl = balsa_mblist_find_all_unread_mboxes();
+    gl = balsa_mblist_find_all_unread_mboxes(NULL);
 
     if (gl) {
         for (i = g_list_first(gl); i; i = g_list_next(i)) {
             printf("opening %s..\n", (LIBBALSA_MAILBOX(i->data))->name);
-            if(0)
-                balsa_mblist_open_mailbox(LIBBALSA_MAILBOX(i->data));
+            balsa_mblist_open_mailbox(LIBBALSA_MAILBOX(i->data));
         }
         g_list_free(gl);
     }
@@ -510,19 +535,14 @@ scan_mailboxes_idle_cb()
  * principle grow indefinetely. */
 static gboolean
 mbnode_expunge_func(GtkTreeModel *model, GtkTreePath *path,
-                    GtkTreeIter *iter)
+                    GtkTreeIter *iter, GSList ** list)
 {
     BalsaMailboxNode *mbnode;
 
+    gtk_tree_model_get(model, iter, 0, &mbnode, -1);
     g_return_val_if_fail(mbnode, FALSE);
 
-    gtk_tree_model_get(model, iter, 0, &mbnode, -1);
-    if (mbnode->mailbox && libbalsa_mailbox_is_open(mbnode->mailbox)) {
-        time_t tm = time(NULL);
-        if(tm-mbnode->last_use > balsa_app.expunge_timeout)
-            libbalsa_mailbox_sync_storage(mbnode->mailbox, TRUE);
-    }
-    g_object_unref(mbnode);
+    *list = g_slist_prepend(*list, mbnode);
 
     return FALSE;
 }
@@ -530,24 +550,81 @@ mbnode_expunge_func(GtkTreeModel *model, GtkTreePath *path,
 static gboolean
 periodic_expunge_cb(void)
 {
+    GSList *list = NULL, *l;
+
 #if !defined(ENABLE_TOUCH_UI)
     /* should we enforce expunging now and then? Perhaps not... */
     if(!balsa_app.expunge_auto) return TRUE;
 #endif
 
+    gdk_threads_enter();
     libbalsa_information(LIBBALSA_INFORMATION_MESSAGE,
                          _("Compressing mail folders..."));
-    /* make sure it is shown before we continue */
-    gdk_threads_enter();
     gtk_tree_model_foreach(GTK_TREE_MODEL(balsa_app.mblist_tree_store),
 			   (GtkTreeModelForeachFunc)mbnode_expunge_func,
-			   NULL);
+			   &list);
     gdk_threads_leave();
+
+    for (l = list; l; l = l->next) {
+        BalsaMailboxNode *mbnode = l->data;
+        if (mbnode->mailbox && libbalsa_mailbox_is_open(mbnode->mailbox)) {
+            time_t tm = time(NULL);
+            if (tm-mbnode->last_use > balsa_app.expunge_timeout)
+                libbalsa_mailbox_sync_storage(mbnode->mailbox, TRUE);
+        }
+        g_object_unref(mbnode);
+    }
+    g_slist_free(list);
+
     /* purge imap cache? leave 15MB */
     libbalsa_imap_purge_temp_dir(15*1024*1024);
+
     return TRUE; /* do it later as well */
 }
 
+/*
+ * Wrappers for libbalsa access to the progress bar.
+ */
+
+/*
+ * Initialize the progress bar and set text.
+ */
+static gdouble prev_fraction;
+static void
+balsa_progress_set_text(LibBalsaProgress * progress, const gchar * text,
+                        guint total)
+{
+    gboolean rc = FALSE;
+
+    gdk_threads_enter();
+    if (!text || total >= LIBBALSA_PROGRESS_MIN_COUNT)
+        rc = balsa_window_setup_progress(balsa_app.main_window, text);
+    prev_fraction = 0;
+    gdk_threads_leave();
+
+    *progress = (text && rc) ?
+        LIBBALSA_PROGRESS_YES : LIBBALSA_PROGRESS_NO;
+}
+
+/* 
+ * Set the fraction in the progress bar.
+ */
+
+static void
+balsa_progress_set_fraction(LibBalsaProgress * progress, gdouble fraction)
+{
+    if (*progress == LIBBALSA_PROGRESS_NO
+        || (fraction != 0 && fraction != 1
+            && fraction < prev_fraction + LIBBALSA_PROGRESS_MIN_UPDATE))
+        return;
+    prev_fraction = fraction;
+
+    gdk_threads_enter();
+    balsa_window_increment_progress(balsa_app.main_window, fraction,
+                                    !libbalsa_am_i_subthread());
+    gdk_threads_leave();
+}
+
 /* -------------------------- main --------------------------------- */
 int
 main(int argc, char *argv[])
@@ -564,10 +641,7 @@ main(int argc, char *argv[])
     bindtextdomain(PACKAGE, GNOMELOCALEDIR);
     bind_textdomain_codeset(PACKAGE, "UTF-8");
     textdomain(PACKAGE);
-    /* FIXME: gnome_i18n_get_language seems to have gone away; 
-     * is this a reasonable replacement? */
-    setlocale(LC_CTYPE,
-              (const char *) gnome_i18n_get_language_list("LC_CTYPE")->data);
+    setlocale(LC_ALL, "");
 #endif
 
 
@@ -601,13 +675,14 @@ main(int argc, char *argv[])
     libbalsa_init((LibBalsaInformationFunc) balsa_information_real);
     libbalsa_filters_set_url_mapper(balsa_find_mailbox_by_url);
     libbalsa_filters_set_filter_list(&balsa_app.filters);
+
+    libbalsa_progress_set_text     = balsa_progress_set_text;
+    libbalsa_progress_set_fraction = balsa_progress_set_fraction;
     
     /* checking for valid config files */
     config_init(cmd_get_stats);
-#if NEW_ADDRESS_ENTRY_WIDGET
     libbalsa_address_entry_set_address_book_list(balsa_app.
                                                  address_book_list);
-#endif /* NEW_ADDRESS_ENTRY_WIDGET */
 
     libbalsa_mailbox_view_table =
 	g_hash_table_new_full(g_str_hash, g_str_equal,
@@ -617,7 +692,7 @@ main(int argc, char *argv[])
 
     default_icon = balsa_pixmap_finder("balsa_icon.png");
     if(default_icon) { /* may be NULL for developer installations */
-        gnome_window_icon_set_default_from_file(default_icon);
+        gtk_window_set_default_icon_from_file(default_icon, NULL);
         g_free(default_icon);
     }
 
